! 
! Module: ascregion_mod - writes ASC compatible FITS region files
!
!
! Developer: Tim Harries (tjh@mssl.ucl.ac.uk)
!
! History:
!
!    Created: 30 July 1998
!
!
module ascregion_mod

use dal
use errorhandling
implicit none

integer, parameter :: MAX_POLY = 50
integer, parameter :: SHAPE_LEN = 20


contains

!--------------------------------------------------------------------

  type(DataSetT) function createRegion(filename) result(regDataSet)

    character(len=*), intent(in) :: filename
    type(TableT) :: tableSet
    type(ColumnT) :: col
    integer :: dims(0:0)
    integer :: cdims(0:0)
    dims(0) = MAX_POLY
    cdims(0) = SHAPE_LEN

    regDataSet = dataSet(filename, create)
    tableSet = addTable(regDataSet, "REGION", 0)
    call setAttribute(tableSet, "HDUCLAS1","REGION")
    call setAttribute(tableSet, "MFORM1","X,Y")
    col = addColumn(tableSet, "SHAPE", String, " ", cdims)
    call release(col)

    col = addColumn(tableSet, "X", Real32, " ", dims)
    call release(col)

    col = addColumn(tableSet, "Y", Real32, " ", dims)
    call release(col)

    col = addColumn(tableSet, "R", Real32, " ",  dims)
    call release(col)

    col = addColumn(tableSet, "ROTANG", Real32, " ",dims)
    call release(col)

    col = addColumn(tableSet, "COMPONENT", Integer32)
    call release(col)

    call release(tableSet)


  end function createRegion

!--------------------------------------------------------------------

  subroutine addPolygon(regDataSet, x, y, nPts, iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x(*)
    real, intent(in) :: y(*)
    integer, intent(in) :: nPts
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    integer :: i
    type(TableT) :: tableSet
    type(ColumnT) :: col
    real(kind=single), pointer :: vals(:,:)
    integer :: nRows
    character(len=SHAPE_LEN) :: shape

    shape = "POLYGON"

    if ( (nPts+1) > MAX_POLY) call fatal("bigpoly", &
                     "Polygon has too many nodes for region file")

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)

    col = column(tableSet, "X", modify)
    vals => real32VectorData(col)
    vals(:,nRows) =0.
    do i = 1, nPts
      vals(i-1, nRows) = x(i)
    enddo
    vals(nPts, nRows) = x(1)
    call release(col)

    col = column(tableSet, "Y", modify)
    vals => real32VectorData(col)
    vals(:,nRows) =0.
    do i = 1, nPts
      vals(i-1,nRows) = y(i)
    enddo
    vals(nPts,nRows) = y(1)
    call release(col)

    call setComponent(tableSet, nRows, iComponent)
       
    call release(tableSet)
   
  end subroutine addPolygon

!--------------------------------------------------------------------

  subroutine addCircle(regDataSet, x, y, r, iComponent, usethis)
    type(DataSetT) :: regDataSet
    real, intent(in) :: x
    real, intent(in) :: y
    real, intent(in) :: r
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    type(TableT) :: tableSet
    integer :: nRows
    character(len=SHAPE_LEN) :: shape

    shape="CIRCLE"

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)

    call setXY(tableSet, nRows, x, y)

    call setSingle(tableSet, nRows, "R", r)

    call setComponent(tableSet, nRows, iComponent)

  end subroutine addCircle

!--------------------------------------------------------------------

  subroutine addAnnulus(regDataSet, x, y, r1, r2, iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x
    real, intent(in) :: y
    real, intent(in) :: r1, r2
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    real r(2)
    type(TableT) :: tableSet
    integer :: nRows
    character(len=SHAPE_LEN) :: shape

    shape = "ANNULUS"

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)

    call setXY(tableSet, nRows, x, y)

    r(1) = r1
    r(2) = r2

    call setMany(tableSet, nRows, "R", r, 2)

    call setComponent(tableSet, nRows,iComponent)
    
  end subroutine addAnnulus

!--------------------------------------------------------------------

  subroutine addPoint(regDataSet, x, y, iComponent, usethis)
    type(DataSetT) :: regDataSet
    real, intent(in) :: x
    real, intent(in) :: y
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    type(TableT) :: tableSet
    integer :: nRows
    character(len=SHAPE_LEN) :: shape

    shape = "POINT"

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)

    call setXY(tableSet, nRows, x , y)

    call setComponent(tableSet, nRows,iComponent)
    
  end subroutine addPoint

!--------------------------------------------------------------------

  subroutine addEllipse(regDataSet, x, y, rMaj, rMin, rotAng, &
                        iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x
    real, intent(in) :: y
    real, intent(in) :: rMaj, rMin, rotAng
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    real :: r(2)
    type(TableT) :: tableSet
    integer :: nRows
    character(len=SHAPE_LEN) :: shape

    shape = "ELLIPSE"

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)


    call setXY(tableSet, nRows, x, y)


    r(1) = rMaj
    r(2) = rMin
    call setMany(tableSet, nRows, "R", r , 2)

    call setSingle(tableSet, nRows, "ROTANG", rotAng)

    call setComponent(tableSet, nRows,iComponent)
    
  end subroutine addEllipse

!--------------------------------------------------------------------

  subroutine addElliptAnnulus(regDataSet, x, y, rMajIn, rMinIn, &
    rMajOut, rMinOut, rotAngin, rotAngOut, iComponent, usethis)

    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x
    real, intent(in) :: y
    real, intent(in) :: rMajIn, rMinIn, rotAngIn
    real, intent(in) :: rMajOut, rMinOut, rotAngOut
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    real :: r(4)
    type(TableT) :: tableSet
    integer :: nRows
    character(len=SHAPE_LEN) :: shape

    shape = "ELLIPTANNULUS"

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)

    call setXY(tableSet, nRows, x, y)

    r(1) = rMajIn
    r(2) = rMinIn
    r(3) = rMajOut
    r(4) = rMinOut

    call setMany(tableSet, nRows, "R", r , 4)


    r(1) = rotAngIn
    r(2) = rotAngOut

    call setMany(tableSet, nRows, "ROTANG", r, 2)

    call setComponent(tableSet, nRows,iComponent)
    
  end subroutine addElliptAnnulus

!--------------------------------------------------------------------

  subroutine addBox(regDataSet, x, y, xsize, ysize, iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x
    real, intent(in) :: y
    real, intent(in) :: xSize, ySize
    logical, intent(in) :: usethis
    integer, intent(in) :: iComponent
    real :: r(2)
    type(TableT) :: tableSet
    integer :: nRows
    character(len=SHAPE_LEN) :: shape

    shape = "BOX"

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)

    call setXY(tableSet, nRows, x, y)

    r(1) = xSize
    r(2) = ySize
    call setMany(tableSet, nRows, "R", r , 2)

    call setComponent(tableSet, nRows,iComponent)
    
  end subroutine addBox

!--------------------------------------------------------------------

  subroutine addRotbox(regDataSet, x, y, xsize, ysize, rotAng, &
                       iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x
    real, intent(in) :: y
    real, intent(in) :: xSize, ySize
    real, intent(in) :: rotAng
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    type(TableT) :: tableSet
    integer :: nRows
    real :: r(2)
    character(len=SHAPE_LEN) :: shape

    shape = "ROTBOX"

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)

    call setXY(tableSet, nRows, x, y)

    r(1) = xSize
    r(2) = ySize
    call setMany(tableSet, nRows, "R", r , 2)

    call setSingle(tableSet, nRows, "ROTANG", rotAng)

    call setComponent(tableSet, nRows,iComponent)
    
  end subroutine addRotbox

!--------------------------------------------------------------------

  subroutine addRectangle(regDataSet, x1, x2, y1, y2, iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x1, x2
    real, intent(in) :: y1, y2
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    type(TableT) :: tableSet
    real :: r(2)
    integer :: nRows
    character(len=SHAPE_LEN) :: shape

    shape = "RECTANGLE"

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)

    r(1) = x1
    r(2) = x2
    call setMany(tableSet, nRows, "X", r , 2)

    r(1) = y1
    r(2) = y2
    call setMany(tableSet, nRows, "Y", r , 2)

    call setComponent(tableSet, nRows,iComponent)
    
  end subroutine addRectangle

!--------------------------------------------------------------------

  subroutine addRotrectangle(regDataSet, x1, x2, y1, y2, rotAng, &
                             iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x1, x2
    real, intent(in) :: y1, y2
    real, intent(in) :: rotAng
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    real :: r(2)
    type(TableT) :: tableSet
    integer :: nRows
    character(len=SHAPE_LEN) :: shape

    shape = "ROTRECTANGLE"

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)

    r(1) = x1
    r(2) = x2
    call setMany(tableSet, nRows, "X", r , 2)

    r(1) = y1
    r(2) = y2
    call setMany(tableSet, nRows, "Y", r , 2)

    call setSingle(tableSet, nRows, "ROTANG", rotAng)

    call setComponent(tableSet, nRows,iComponent)
    
  end subroutine addRotrectangle

!--------------------------------------------------------------------

  subroutine addPie(regDataSet, x, y, Ang1, Ang2, iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x, y, ang1, ang2
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis

    call addPieSector("PIE", regDataSet, x, y, Ang1, Ang2, &
                      iComponent, usethis)
  end subroutine addPie

!--------------------------------------------------------------------

  subroutine addSector(regDataSet, x, y, Ang1, Ang2, iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x, y, ang1, ang2
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis

    call addPieSector("SECTOR", regDataSet, x, y, Ang1, Ang2, &
                      iComponent, usethis)
  end subroutine addSector

!--------------------------------------------------------------------

  subroutine addPieSector(shape, regDataSet, x, y, Ang1, Ang2, &
                          iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x, y
    real, intent(in) :: Ang1, Ang2
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    type(TableT) :: tableSet
    real :: r(2)
    integer :: nRows
    character(len=*) :: shape

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)


    call setXY(tableSet, nRows, x, y)

    r(1) = Ang1
    r(2) = Ang2

    call setMany(tableSet, nRows, "ROTANG", r , 2)

    call setComponent(tableSet, nRows,iComponent)
    
  end subroutine addPieSector

!--------------------------------------------------------------------

  subroutine addDiamond(regDataSet, x, y, xsize, ysize, iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x, y, xsize, ysize
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis

    call addDiamondRhombus("DIAMOND", regDataSet, x, y, xsize, ysize, &
                           iComponent, usethis)
  end subroutine addDiamond

!--------------------------------------------------------------------


  subroutine addRhombus(regDataSet, x, y, xsize, ysize, iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x, y, xsize, ysize
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis

    call addDiamondRhombus("RHOMBUS", regDataSet, x, y, xsize, ysize,  &
                           iComponent, usethis)
  end subroutine addRhombus

!--------------------------------------------------------------------


  subroutine addDiamondRhombus(shape, regDataSet, x, y, xsize, ysize, &
                               iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x, y
    real, intent(in) :: xsize, ysize
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    type(TableT) :: tableSet
    real :: r(2)
    integer :: nRows
    character(len=*) :: shape

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)


    call setXY(tableSet, nRows, x, y)

    r(1) = xsize
    r(2) = ysize

    call setMany(tableSet, nRows, "R", r , 2)

    call setComponent(tableSet, nRows,iComponent)
    
  end subroutine addDiamondRhombus

!--------------------------------------------------------------------

  subroutine addRotdiamond(regDataSet, x, y, xsize, ysize, rotang, &
                           iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x, y, xsize, ysize
    real, intent(in) :: rotang
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis

    call addRotDiamondRhombus("ROTDIAMOND", regDataSet, x, y, &
                              xsize, ysize, rotang, iComponent, usethis)
  end subroutine addRotdiamond

!--------------------------------------------------------------------


  subroutine addRotrhombus(regDataSet, x, y, xsize, ysize, rotang, &
                           iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x, y, xsize, ysize
    real, intent(in) :: rotang
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis

    call addRotDiamondRhombus("ROTRHOMBUS", regDataSet, x, y, &
                              xsize, ysize, rotang, iComponent, usethis)
  end subroutine addRotrhombus

!--------------------------------------------------------------------

  subroutine addRotDiamondRhombus(shape, regDataSet, x, y, xsize, ysize, &
                                  rotAng, iComponent, usethis)
    type(DataSetT), intent(in) :: regDataSet
    real, intent(in) :: x, y
    real, intent(in) :: xsize, ysize
    real, intent(in) :: rotAng
    integer, intent(in) :: iComponent
    logical, intent(in) :: usethis
    real :: r(2)
    type(TableT) :: tableSet
    integer :: nRows
    character(len=*) :: shape

    call getTable(regDataSet, tableSet, nRows)

    call setShape(tableSet, shape, nRows, usethis)


    call setXY(tableSet, nRows, x, y)

    r(1) = xsize
    r(2) = ysize

    call setMany(tableSet, nRows, "R", r , 2)

    call setSingle(tableSet, nRows, "ROTANG", rotAng)

    call setComponent(tableSet, nRows,iComponent)
    
  end subroutine addRotDiamondRhombus

!--------------------------------------------------------------------

  subroutine closeRegion(regDataSet)
    type(DataSetT), intent(inout) :: regDataSet
    call release(regDataSet)
  end subroutine closeRegion

!--------------------------------------------------------------------

  subroutine getTable(regDataSet, tableSet, nRows)
    type(DataSetT), intent(in) :: regDataSet
    type(TableT), intent(out) :: tableSet
    integer, intent(out) :: nRows

    tableSet = table(regDataSet, "REGION")
    nRows = numberOfRows(tableSet)
    call insertRows(tableSet, nRows, 1)
  end subroutine getTable
  
!--------------------------------------------------------------------

  subroutine setShape(tableSet, shape, nRow, usethis)

    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: shape
    integer, intent(in) :: nRow
    logical, intent(in) :: usethis
    type(ColumnT) :: col
    character(len=SHAPE_LEN) :: cshape

    cshape = " "
    if (.not.usethis) then
       cshape = "!"//trim(shape)
       else
       cshape = shape
    endif
    
    col = column(tableSet, "SHAPE", modify)
    call setStringCell(col, nRow, cshape)
    call release(col)

  end subroutine setShape

!--------------------------------------------------------------------

  subroutine setComponent(tableSet, nRow, iComponent)
    type(TableT), intent(in) :: tableSet
    integer, intent(in) :: nRow
    integer, intent(in) :: iComponent
    integer(kind=int32), pointer :: ivals(:)
    type(ColumnT) :: col

    col = column(tableSet, "COMPONENT", modify)
    ivals => int32Data(col)
    ivals(nRow) = iComponent
    call release(col)
  end subroutine setComponent

!--------------------------------------------------------------------

  subroutine setXY(tableSet, nRows, x, y)

    type(TableT), intent(in) :: tableSet
    integer, intent(in) :: nRows
    real, intent(in) :: x, y
    type(ColumnT) :: col
    real(kind=Single), pointer :: vals(:,:)

    col = column(tableSet, "X", modify)
    vals => real32VectorData(col)
    vals(:,nRows) =0.
    vals(0, nRows) = x
    call release(col)

    col = column(tableSet, "Y", modify)
    vals => real32VectorData(col)
    vals(:,nRows) =0.
    vals(0, nRows) = y
    call release(col)

  end subroutine setXY

!--------------------------------------------------------------------

  subroutine setMany(tableSet, nRows, colName, r, nPts)
    type(TableT), intent(in) :: tableSet
    integer, intent(in) :: nRows
    character(len=*), intent(in) :: colName
    real, intent(in) :: r(*)
    integer, intent(in) :: nPts
    type(ColumnT) :: col
    integer :: i
    real(kind=Single), pointer :: vals(:,:)

    col = column(tableSet, colName, modify)
    vals => real32VectorData(col)
    vals(:,nRows) =0.
    do i = 1, nPts
      vals(i-1,nRows) = r(i)
    enddo
    call release(col)
  end subroutine setMany

!--------------------------------------------------------------------

  subroutine setSingle(tableSet, nRows, colName, r)
    type(TableT), intent(in) :: tableSet
    integer, intent(in) :: nRows
    character(len=*), intent(in) :: colName
    real, intent(in) :: r
    type(ColumnT) :: col
    real(kind=Single), pointer :: vals(:,:)

    col = column(tableSet, colName, modify)
    vals => real32VectorData(col)
    vals(:,nRows) =0.
    vals(0,nRows) = r
    call release(col)
  end subroutine setSingle

!--------------------------------------------------------------------


end module ascregion_mod

