module frametime_mod

   use cal
   use dal
   use errorhandling

   implicit none

   public

contains
!**************************************
! Subroutine fetFrameTime
!**************************************
subroutine getFrameTime(wdxDataSetH, frametime, deadfrac, binBpe) 

   type(DataSetT), intent(in) :: wdxDataSetH                         ! Priority Window Data dataset handle.

   type(OMframeParameters) :: frameParameters                        ! Om frame parameters

   real(double) :: frameTime                                    ! Frame time.
   real(double) :: deadFrac                                     ! Dead time in units of fraction of ccd frame time.
   logical, intent(in) :: binBpe
   type(TableT) :: wdxTableH                                         ! Priority Window Data table dataset handle.
   type(ColumnT), dimension(5) :: wdxColumnH                         ! Priority Window Data column handle.

   integer(int32), dimension(:), pointer :: nDetectorWindowsP   ! Number of detector windows pointer.
   integer(int32), dimension(:, :), pointer :: dWX0P            ! Detector window X0 pointer.
   integer(int32), dimension(:, :), pointer :: dWY0P            ! Detector window Y0 pointer.
   integer(int32), dimension(:, :), pointer :: dWDXP            ! Detector window DX pointer.
   integer(int32), dimension(:, :), pointer :: dWDYP            ! Detector window DY pointer.

   integer(int32) :: nDetectorWindows                           ! Number of detector windows.
   integer(int32), dimension(16) :: dWX0                        ! Detector window X0.
   integer(int32), dimension(16) :: dWY0                        ! Detector window Y0.
   integer(int32), dimension(16) :: dWDX                        ! Detector window DX.
   integer(int32), dimension(16) :: dWDY                        ! Detector window DY.

   integer(int32) :: i                                          ! Count.
   integer(int32) :: j                                          ! Count.
   integer(int32) :: constant                                          

   type detectorWindow
      integer :: x0
      integer :: y0
      integer :: dx
      integer :: dy
   end type detectorWindow

   type(detectorWindow), dimension(16) :: window
  

   real(double) :: NVPI                                      ! Number of vertical pixels in image.
   real(double) :: NVPS                                      ! Number of vertical pixels in store.
   real(double) :: NHP                                       ! Number of horizontal pixels. 
   real(double) :: VTRANS                                    ! Vertical transfer time.
   real(double) :: HTRANS                                    ! Horizontal transfer time.
   real(double) :: V2H                                       ! Vertical 2 horizontal switch time.
   real(double) :: H2V                                       ! Horizontal 2 vertical switch time.
   real(double) :: POFFSET                                   ! Prime CCD offset (52, 27).
   real(double) :: ROFFSET                                   ! Redundant CCD offset (64, 15).

!   integer(kind=int32), parameter :: NVPI = 290                      ! Number of vertical pixels in image.
!   integer(kind=int32), parameter :: NVPS = 288                      ! Number of vertical pixels in store.
!   integer(kind=int32), parameter :: NHP = 407                       ! Number of horizontal pixels. 
!   integer(kind=int32), parameter :: VTRANS = 600                    ! Vertical transfer time.
!   integer(kind=int32), parameter :: HTRANS = 100                    ! Horizontal transfer time.
!   integer(kind=int32), parameter :: V2H = 200                       ! Vertical 2 horizontal switch time.
!   integer(kind=int32), parameter :: H2V = 100                       ! Horizontal 2 vertical switch time.
!   integer(kind=int32), parameter :: POFFSET = 27                    ! Prime CCD offset (52, 27).
!   integer(kind=int32), parameter :: ROFFSET = 15                    ! Redundant CCD offset (64, 15).

   real(double) :: FLUSH                                        ! Flush sequence.
   real(double) :: READDUMP                                     ! Readout and dump sequence.
   real(double) :: READOUT                                      ! Readout sequence.

   real(double), dimension(-3:256) :: windowReadoutTime         ! Window readout time array.

 

! Get frame parameters

   call CAL_setState(OM)
   call CAL_omGetFrameParameters(frameParameters)

   NVPI = frameParameters%nVerticalPixelsImage
   NVPS = frameParameters%nVerticalPixels
   NHP = frameParameters%nHorizontalPixels
   VTRANS = frameParameters%verticalTransferTime
   HTRANS = frameParameters%horizontalTransfertime
   V2H = frameParameters%verticalHorizontalSwitchTime
   H2V = frameParameters%horizontalVerticalSwitchTime
   POFFSET = frameParameters%redundantCcdOffset
   ROFFSET = frameParameters%redundantCcdOffset
 
!   write(*, *) "nvp1=", nvpi
!   write(*, *) "nvps=", nvps
!   write(*, *) "nhp=", nhp
!   write(*, *) "vtrans=", vtrans
!   write(*, *) "htrans=", htrans
!   write(*, *) "v2h=", v2h
!   write(*, *) "poffset=", poffset
!   write(*, *) "roffset=", roffset




!   NVPI = 290
!   POFFSET = 15
!   NHP = 407
!   HTRANS = 1.d-7
!   V2H = 2.d-7
!   H2V = 1.d-7

   FLUSH = V2H+(NHP*HTRANS)+H2V    
   READDUMP = VTRANS+FLUSH
   READOUT = READDUMP

!   write(*,*)' flush,rddump,rdout= ',flush,readdump,readout

! Read in OM ODF Priority Window Data Auxiliary file.

   if (hasBlock(wdxDataSetH, "OMWDX1")) then

! Get a handle on the window data binary table extension.

      wdxTableH = table(wdxDataSetH, "OMWDX1")

! Extract the number of detector windows and the window data.

      wdxColumnH(1) = column(wdxTableH, "DETWIND", read)
      nDetectorWindowsP => int32Data(wdxColumnH(1))
      nDetectorWindows = nDetectorWindowsP(0)

      wdxColumnH(2) = column(wdxTableH, "DWX0", read)
      dWX0P => int32VectorData(wdxColumnH(2))
      dWX0 = dWX0P(:, 0)

      wdxColumnH(3) = column(wdxTableH, "DWY0", read)
      dWY0P => int32VectorData(wdxColumnH(3))
      dWY0 = dWY0P(:, 0)

      wdxColumnH(4) = column(wdxTableH, "DWXSIZE", read)
      dWDXP => int32VectorData(wdxColumnH(4))
      dWDX = dWDXP(:, 0)

      wdxColumnH(5) = column(wdxTableH, "DWYSIZE", read)
      dWDYP => int32VectorData(wdxColumnH(5))
      dWDY = dWDYP(:, 0)


! Release handles on data.

      call release(wdxColumnH(1))
      call release(wdxColumnH(2))
      call release(wdxColumnH(3))
      call release(wdxColumnH(4))
      call release(wdxColumnH(5))
      call release(wdxTableH)
!      call release(wdxDataSetH)
   else

! Incompatible file.

!      write (errstr, *) "Could not find any compatibile blocks in set ", wdxSet
      call fatal("badFileType")
   end if


   !*******************************************
   ! 26.9.2001 constant determined by binning
   ! BINBPE present, if parameter present
   !*******************************************     
   if (binBpe) then
       constant = 4
   else
       constant = 8 ! Original value
   end if  



   do i = 1, nDetectorWindows
      window(i)%x0 = dWX0(i) / constant
      window(i)%y0 = dWY0(i) / constant 
      window(i)%dx = dWDX(i) / constant 
      window(i)%dy = dWDY(i) / constant
      
   end do




! Calculate frametime.

! Move image to store.

   frameTime = NVPI * VTRANS

! Flush readout.

   frameTime = frameTime + FLUSH

! Move to active area. (Note that we are using the prime CCD,
! should the redundant CCD be used, use ROFFSET).

   frameTime = frameTime + dble((POFFSET - 2.0D0) * VTRANS)


! Move through windows.
   
   

   windowReadoutTime = 0
   do i = 1, nDetectorWindows
      
      if ((window(i)%dx > 0) .and. (window(i)%dy > 0)) then 
          if  ((window(i)%y0 ) .gt. 256) then
              call fatal("GetFrameTime", "window%y0 > 256- can't continue")
          
          elseif (window(i)%dy .gt. 256) then
              call fatal("GetFrameTime", "window%dy > 256- can't continue")
      
                                                                                                                        
          elseif ((window(i)%y0 + window(i)%dy) .gt. 256) then
              call fatal("GetFrameTime", "window%y0+window%dy > 256- can't continue")
          end if
          windowReadoutTime(window(i)%y0-1 : window(i)%y0 + window(i)%dy) = READOUT
      end if
   end do


! Check for vertical transfers sequences and readout and dump sequences.

   i = -3
   do while (i <= 256)

      if (windowReadoutTime(i) == 0) then
         j = i
         do while (j <= 256)
          
            if (windowReadoutTime(j) /= 0) then
               if ((j - i) > 2) then
                  windowReadoutTime(i : j - 3) = VTRANS
                  windowReadoutTime(j - 2 : j - 1) = READDUMP
               else
                  windowReadoutTime(i : j - 1) = READDUMP

               end if
               exit
            end if
            j = j + 1
         end do
      end if
      i = i + 1
   end do

   frameTime = frameTime + dble(sum(windowReadoutTime))
   deadFrac = VTRANS * NVPI / frameTime 

   !*******************************
   ! Convert from s to ms.
   !*******************************
   frameTime = frameTime * 1.0d3

 end subroutine getFrameTime

end module frametime_mod








