!******************************************************
! August 2000
! Module to correct the astrometry for misalignment
! of the ra-dec axes of the XMM and usno catalogue.
! Will also determine the rotation between the 2 axes.
! Based on the original code in omatt_mod.90, but ammended
! so that it works.
!******************************************************min
module omastromcorrect_mod
    use param
    use dal
    use errorHandling
    use sla
    use sasdal
    use omallocation_mod    
    use omutility_mod  
    use omconstants_mod
    use omnumerics_mod
    use omgraph_mod
public


 
    !**************************************************************
    ! A structure for use in matching the sources to the catalogue
    !**************************************************************
    Type SHIFT
        real(single) :: x
        real(single) :: y
        real(single) :: radius
    end Type SHIFT

    !*********************************************
    ! A structure for reading the catalogue data   
    !*********************************************
    type CatalogueData
        real(double) :: RA = 0.0_double
        real(double) :: dec = 0.0_double
        real(single) :: error = 0.0
        real(single) :: magnitude = 0.0
    end type CatalogueData



    !**************************************************
    ! A structure for storing the fitted parameters
    !*************************************************
    type CorrectionData
        real(single) :: xShift = 0.0, sigXShift = 0.0
        real(single) :: yShift = 0.0, sigYShift = 0.0
        real(single) :: rotation = 0.0, sigRotation = 0.0
        real(single) :: xScale = 0.0, sigXScale = 0.0
        real(single) :: yScale = 0.0, sigYScale = 0.0
        real(single) :: magOffset = 0.0
        real(single) :: percentageMatched = 0.0
        real(double) :: maxLikelihood = 0.0_double
        real(double) :: rmsResidual = 0.0_double
        integer(int32) :: noMatched = 0
        character(4) :: filter
        character(13) :: expId

    end type CorrectionData   

    real(double) :: radiansToArcSeconds  

    integer(int32), private :: minNoStars = -1
    logical, private :: correctPositions = .true.
    real(single), private :: tolerance = 3.0  ! Tolerance for catalog star search (arcsec) was 4.

    integer(int8), private :: xCorrRange = 30 ! was 20 

contains


!*************************************************************************
! Sets the minimum number of stars that can be used in the astrometry fit
!*************************************************************************
subroutine setXCorrRange( range )
    integer(int8), intent(in) :: range
    
    xCorrRange = range

end subroutine setXCorrRange

!*************************************************************************
! Sets the tolerance (arc secs)
!*************************************************************************
subroutine setTolerance( tol )
    real(single), intent(in) :: tol
    
    tolerance = tol

end subroutine setTolerance 

!*************************************************************************
! Sets the minimum number of stars that can be used in the astrometry fit
!*************************************************************************
subroutine setMinNoSources( nStars )
    integer(int32), intent(in) :: nStars
    
    minNoStars = nStars

end subroutine setMinNoSources 

!***************************************************************
! Sets logical variable correctPositions
! If true then source-positions will be corrected
!**************************************************************
subroutine setCorrectPositions( correct )
    logical, intent(in) :: correct
    
    correctPositions = correct

end subroutine setCorrectPositions



!***************************************************************
! Astrometry Correction to star positions using a star catalogue
! NB ras and decs are in radians
! NB ras and decs should be in radians, as should  raPnt, decPnt
!***************************************************************
subroutine CorrectAstrometry(noStars, raSource, decSource, &
        posErr, sourceMagnitude, raPnt, decPnt, fileName, &
        status, corrData, catDataOut)     
    integer(int32), intent(in) :: noStars                   ! The number of stars
   
    real(double) :: raSource(:)
    real(double) :: decSource(:) 
    real(single) :: posErr(:) 
    real(single) :: sourceMagnitude(:) 
    real(double), intent(in) :: raPnt, decPnt
    character(*), intent(in) :: fileName             ! The catalogue file name     
    logical, intent(out) :: status
    type(CorrectionData), intent(out) :: corrData
    type(CatalogueData), optional, pointer :: catDataOut(:)
    real(double)    :: ep2     ! equinox
    type(DataSetT) :: dataSetH
    real(double), pointer, dimension(: ,  : ) :: xye => null(), xym => null()      
    real(double) :: raSourceMin, raSourceMax
    real(double) :: decSourceMin, decSourceMax 
    real(double) :: raCatMin, raCatMax
    real(double) :: decCatMin, decCatMax 
    real(double) :: raMin, decMin
    integer(int32)  :: nCat                                ! number of catalogue stars
    real(double), pointer :: xiSource(:) => null(), etaSource(:) => null()
    real(double), pointer :: xiCat(:) => null(), etaCat(:) =>null()
    real(double), pointer ::  separation(:) => null()
  
    real(single) :: mag, magl, magu
    character(30) :: expID, object
    character(30) :: minNoStarstext
    
    integer :: i, j, k, l, nmag
  
    integer                     :: istat          ! status for memory allocation
    
    logical, pointer  :: useCat(:) => null()
    integer :: nCatInWindow = 0
    character(250)   :: catFileName
  
    
    integer(int32), pointer :: sourceIndex(:) => null()
    integer(int32), pointer :: tempSourceIndex(:) => null()

    integer(int32), pointer :: catIndex(:) => null()
    integer(int32), pointer :: tempCatIndex(:) => null()
 
    real(double), dimension(1 : 6) :: coeffs   
 
    logical :: okay

    real(single), dimension(1 : 4) :: maxMagDifference = (/0.5, 0.5, 0.5, 0.5/)  ! last was 999 
 
    integer, dimension(1 : 4)           :: nSigma = (/3, 3, 3, 3/)

    integer :: nStars
    integer :: count1
    integer :: maxNoMatched = 0
    real(single), dimension(1 : 4) ::  minMagLimit = (/12.0, 13.0, 13.0, 13.0/)
    real(single), dimension(1 : 4) ::  maxMagLimit = (/15.0, 18.0, 21.0, 25.0/)
    real(single), dimension(1 : 17) ::  magOffset = (/0.0, -0.5, 0.5, -1.0, 1.0, -1.5, 1.5, &
                                                -2.0, 2.0, -2.5, 2.5, -3.0, 3.0, -3.5, &
                                                3.5, -4.0, 4.0/)


    character(200) :: solutionString
    integer(int32), dimension(1 : 4) :: noIntervals = (/1, 2, 4, 4/)
    real(single) :: medianPositionDifference
    real(single) :: medianMagDifference, minMedianMagDifference
    integer :: iStart, iFinish, jStart, jFinish
    type(CatalogueData), pointer :: catData(:)
    real(single) :: minMag, maxMag, distance, maxDistance
    real(single) :: minRmsResidual, minXShift, minYshift
    real(double) :: rmsResidual = 0.0_double
!    real(double), pointer :: weight(:) => null()






   radiansToArcSeconds = radiansToDegrees * 3600.0_double

    catFileName = trim(stringParameter( "catfile" ))

    !********************************************************
    ! Get the catalogue data from the fits file (if present)
    !********************************************************
    if ( .not. setExists( trim(catFileName) )) then
        call warning("Star catalogue not found- no astrometry correction done", trim(catFileName) ) 
        return       ! File doesn't exist
    end if



    !*********************************************
    ! Don't use stars with large position errors 
    ! (The error is in arcsecs)
    !*********************************************    
    nStars = 0

    do i = 1, noStars
        if (posErr(i) .le. tolerance) then
            nStars = nStars + 1
        end if 
    end do







!    weight => allocateSingle(1, noStars, "weight")

!    weight = 1.0 / (max(1.0, poserr * poserr))
    !******************************
    ! Get the catalogue data
    !******************************

    dataSetH = DataSet(fileName, read)
  
    ep2 = real32Attribute(dataSetH, "EQUINOX")

 
    catData =>  GetCatalogueData(catFileName, ep2, nCat)
      
 
    if (nCat .eq. 0) then 
        call warning("noCat", "Catalogue has not been used to correct positions")
        status = .false.
        if (associated(catData)) then
            deallocate(catData)
            nullify(catData)
        end if 

        return
    end if


    if (present(catDataOut)) then

!        allocate(catDataOut(1 : nCat))
!        catDataOut = catData(1 : nCat)
    end if
   
    xiSource => AllocateDouble(1, noStars, "xiSource")

    etaSource => AllocateDouble(1, noStars, "etaSource")
    
    !*********************************************************************
    ! Check for an overlap of the catalogue with the source positions
    !********************************************************************

    raSourceMin = minval(raSource) - (real(xcorrRange) / 3600.0) * radiansToDegrees
    raMin = raSourceMin
    raSourceMax = maxval(raSource) + (real(xcorrRange) / 3600.0) * radiansToDegrees
    
    decSourceMin = minval(decSource) - (real(xcorrRange) / 3600.0) * radiansToDegrees
    decMin = decSourceMin
    decSourceMax = maxval(decSource) + (real(xcorrRange) / 3600.0) * radiansToDegrees
   
   

    
    expID = trim(stringAttribute(dataSetH, "EXP_ID"))
    corrData%expId = stringAttribute(dataSetH, "EXP_ID")

    object = trim(stringAttribute(dataSetH, "OBJECT")) 
    corrData%filter = stringAttribute(dataSetH, "FILTER") 
    

    call release(dataSetH)

  

    raCatMin = 1000.0
    raCatMax = -1000.0

    decCatMin = 10000.0
    decCatMax = -10000.0
    do i = 1, nCat

        raCatMin = min(raCatMin, catData(i)%ra)
        raCatMax = max(raCatMax, catData(i)%ra)


        decCatMin = min(decCatMin, catData(i)%dec)
        decCatMax = max(decCatMax, catData(i)%dec)



    end do

        
    if (raCatMin .gt. raSourceMax .or. decCatMin .gt. decSourceMax .or. &
        raCatMax .lt. raSourceMin .or. decCatMax .lt. decSourceMin) then 
        
      
        call warning("Cannot do astrometry correction", "No overlap between sources and catalogue stars")
        write(errstr, '(a, f8.4, a, f8.4)') "ra source min=", raSourceMin,  " ra source max=", raSourceMax      
        call message(AppMsg, Verbose)

        write(errstr, '(a, f8.4, a, f8.4)') "ra cat min=", raCatMin,  " ra cat max=", raCatMax      
        call message(AppMsg, Verbose)

        write(errstr, '(a, f8.4, a, f8.4)') "dec source min=", decSourceMin,  " dec source max=", decSourceMax      
        call message(AppMsg, Verbose)

        write(errstr, '(a, f8.4, a, f8.4)') "dec cat min=", decCatMin,  " dec cat max=", decCatMax      
        call message(AppMsg, Verbose)

        status = .false.

        if (associated(catData)) then
            deallocate(catData)
            nullify(catData)

        end if 
       
        return
    end if
  
  


    !*****************************************************
    ! Create a new list of  the catalog stars that only 
    ! includes those that fall in the osw
    !*****************************************************

    nCatinWindow = 0
   
    minMag = 10.0
    maxMag = 21.0
 



    maxDistance = (real(2 * xCorrRange) + tolerance + 10.0) / cos(decPnt)

    do i = 1, nCat
        if (catData(i)%magnitude .lt. minMag .or. catData(i)%magnitude .gt. maxMag) cycle
       
        do j = 1, noStars

            distance = ArcLength(raSource(j), decSource(j), & 
                        catData(i)%RA, catData(i)%dec)
   
            if (distance .le. maxDistance) then
                nCatInWindow = nCatinWindow + 1
                catData(nCatInWindow)%RA = catData(i)%RA
                catData(nCatInWindow)%dec = catData(i)%dec
                catData(nCatInWindow)%magnitude = catData(i)%magnitude
                exit
            end if 

        end do
    end do





    if ( minNostars .lt. 0) then
        
        minNoStars = max( 4, min(10, min( ncatInWindow / 10, nStars / 10 )) )

    end if 

   
    write(minNoStarstext, '(a, i4)') "Need at least ", minNoStars
    minNostarsText = trim(minNostarsText)
 


    if (nStars .lt. minNoStars) then

        write(errstr, '(a, i1, a)') "Too few sources for astrometry (", nStars, ")"   
        call warning(trim(errstr), trim(minNoStarsText))
        status = .false.
        return
    end if











  
    if (nCatInWindow .lt. minNoStars) then 


        call warning("Too few guide stars for astrometry correction", "Catalogue has not been used to correct positions")
        write(errstr, '(a, i5, a, i5)') "Number of stars in catalogue=", nCat, &
                     " number of catalog stars in window=",  nCatInWindow
        call message(AppMsg, Verbose)

        status = .false.


        if (associated(catData)) then
            deallocate(catData)
            nullify(catData)
        end if 

        return


    end if

    nCat = nCatInWindow 

    xiCat => AllocateDouble(1, nCat, "xiCat")
    etaCat => AllocateDouble(1, nCat, "etaCat")

 
    do i = 1, nCat

        !*********************************  
        ! Compute tangent plane coords
        ! xi_gs(i) & eta_gs(i)
        !*********************************
        call sla_ds2tp(catData(i)%RA, catData(i)%dec, raPnt, &
                  decPnt, xiCat(i), etaCat(i), j)

        !*********************************
        ! Check to see okay
        !********************************
        call SlaErrorCheck(j)
 
    end do


    allocate(useCat(1 : nCat), stat = istat)
    call CheckAllocation(istat, "useGs")

    useCat = .true.

 
    write(errstr, '(a, i6, a, i6)') "No of stars=", noStars,  " no of catalogue stars=", nCat      
    call message(AppMsg, Verbose)

  
    separation => AllocateDouble(1, nStars, "separation")
    separation = 1.0_double

    xye => Allocate2DDouble(1, 2, 1, nCat, "xye")
    xym => Allocate2DDouble(1, 2, 1, nCat, "xym")
 
  
    !******************************************  
    ! Compute the tangent plane coordinates
    ! of each source using the RAS and decs and
    ! approximate camera pointing direction.
    !******************************************
      
     
      
    do i = 1, noStars

        call sla_ds2tp(raSource(i), decSource(i), raPnt, &
                  decPnt, xiSource(i), etaSource(i), j)

        !*********************************
        ! Check to see okay
        !********************************
        call SlaErrorCheck(j)
 
    end do
  
    count1 = 0

    do i = 1, noStars
        if (sourceMagnitude(i) .gt. 0.0) count1 = count1 + 1
    end do

    write(errstr, '(a, f6.2, a, f6.2)') "Minimum source magnitude=", minval(sourceMagnitude, sourceMagnitude .gt. 0.0),  &
                    " maximum=", maxval(sourceMagnitude)

    call message(AppMsg, Verbose)

    write(errstr, '(a, f6.2, a, f6.2)') "Minimum catalogue magnitude=", minval(catdata%magnitude, catdata%magnitude .gt. 0.0),  &
                    " maximum=", maxval(catdata%magnitude, catdata%magnitude  .lt. 50.0)

    call message(AppMsg, Verbose)


    if (real(count1) / real(noStars) .lt. 0.5) then
    
        maxMagDifference = 9999.0
        maxMagLimit = 25.0
    else
 
        maxMagLimit(1) = 14.5
        count1 = 0    
        do while (count1 .lt. 20)
            count1 = 0
            maxmagLimit(1) = maxMagLimit(1) + 0.5
            if (maxmagLimit(1) .gt.  maxMagLimit(3)) exit
            do i = 1, noStars
                if (sourceMagnitude(i) .gt. 0.0 .and. sourceMagnitude(i) .le. &
                   maxMagLimit(1) .and. sourceMagnitude(i) .ge. minMagLimit(1)) count1 = count1 + 1
            end do
    
        end do
        if (maxMagLimit(1) .ge. maxMagLimit(2)) then    
            maxMagLimit(2) = maxMagLimit(3)

        end if
    end if
      


    !******************************************
    ! Find the x and y offsets that give the 
    ! best match between the sources and
    ! catalogue objects 
    !******************************************
    sourceIndex => AllocateInt32(1, noStars, "sourceIndex")
    
    catIndex => AllocateInt32(1, noStars, "catIndex")
 


    iStart = -xCorrRange
    iFinish = xCorrRange
    jStart = iStart
    jFinish = iFinish

    !**********************************************************
    ! Find the best offsets using smaller and smaller grid
    ! searches
    !*********************************************************
    

    corrData%noMatched = 0



    iStart = -xCorrRange
    iFinish = xCorrRange
    jStart = iStart
    jFinish = iFinish



    j = 1

    do i = 1, 4
        
 
        write(errstr, '(a, i1, a, f6.1)') "Iteration ", i, " mag limit=", maxmagLimit(i)                
        call message(AppMsg, Verbose)

   
        sourceIndex = 0
        catIndex = 0   
        
        if (i .eq. 1) then
 
            tempSourceIndex => AllocateInt32(1, noStars, "tempsourceIndex")    
            tempCatIndex => AllocateInt32(1, noStars, "tempcatIndex")
            magl = 12.5            
            do l = 1, 8

                maxNoMatched = 0               
                minMedianMagDifference = 0.0
                minRmsResidual = 1e6
                minXShift = 0.0
                minYShift = 0.0  
                mag = real(12 + l)
              
                magu = mag + 0.5

                nmag = count(sourcemagnitude >= magl .and. sourcemagnitude <= magu )
                if ( nmag  .lt. minNostars) cycle
             

               

                do k = 1, size(magOffset)

              
                    
                    sourceIndex = 0
                    catIndex = 0

               
                            
                    call FindBestOffsets(noIntervals(j), iStart, iFinish, jStart, jFinish,   &
                         noStars, raSource, decSource, posErr, sourceMagnitude,              & 
                         nCat, catData%RA, catData%dec, catData%magnitude+magoffset(k),      & 
                         magl, magu, 0.5,                                                    &
                         nSigma(i),  2.0, corrData%noMatched,                                &
                         corrData%xShift, corrData%yShift, corrData%rmsResidual,             &
                         corrData%maxLikelihood, medianPositionDifference,                   &
                         medianMagDifference, sourceIndex, catIndex)







                    if ( corrData%noMatched .ge. maxNoMatched .and. &
                         corrData%noMatched .ge. minNoStars .and. &

                             corrData%rmsResidual .lt. minRmsResidual) then
                        minMedianMagDifference = medianMagDifference + magoffset(k)
                        maxNoMatched = corrData%noMatched
                        minRmsResidual = corrData%rmsResidual
                  
                        minXShift = corrData%xShift
                        minYShift = corrData%yShift

                        tempSourceIndex = sourceIndex
                        tempCatIndex = catIndex


                    end if
 
                end do

                if ( maxNoMatched .ge. max(4, nmag / 2) ) exit

            end do

            medianMagDifference = minMedianMagDifference
            corrData%noMatched = maxNoMatched
            corrData%xShift = minXShift
            corrData%yShift = minYShift


            deallocate( sourceIndex, catIndex )
            nullify( sourceIndex, catIndex )

            sourceIndex => tempSourceIndex
            catIndex => tempCatIndex
       


        else

            sourceIndex = 0
            catIndex = 0   
   
           
            call FindBestOffsets(noIntervals(j), iStart, iFinish, jStart, jFinish,   &
                         noStars, raSource, decSource, posErr, sourceMagnitude,  & 
                         nCat, catData%RA, catData%dec, catData%magnitude,       & 
                         minMagLimit(i), maxmagLimit(i), maxMagDifference(j),    &
                         nSigma(i),  tolerance, corrData%noMatched,              &
                         corrData%xShift, corrData%yShift, corrData%rmsResidual, &
                         corrData%maxLikelihood, medianPositionDifference,       &
                         medianMagDifference, sourceIndex, catIndex)






  
        end if
        write(errstr, '(a, i6, a, f7.2, a, f7.2, a)') "No matched=",  corrData%noMatched, &
               " Median magnitude difference (XMM-Catalogue)=", medianMagDifference      
        call message(AppMsg, Verbose)
        if(corrData%noMatched .ge. nStars) exit
    
        if (i .eq. 1 .and. corrData%noMatched .lt. minNoStars) then
 
              
            cycle
        end if 
      


        if (i .eq. 2 .and. corrData%noMatched .lt. minNoStars) then

           
            maxMagDifference = 9999.0
            maxMagLimit = 25.0
              
            cycle
 
                      
         end if
   

         if (i .eq. 3) then
            if(corrData%noMatched .lt. minNoStars &
                      .and. maxMagDifference(i) < 10.0) then 
               
                 
                maxMagDifference = 9999.0
               
                cycle
            
            else 
                if (corrData%noMatched .ge. minNoStars) then
                  
                     status = .true.
                     exit

                end if 
                write(errstr, '(a, i3, a)') "Too few sources matched for astrometry correction (", &
                          corrData%noMatched, ")"   
                call warning(trim(errstr), trim(minNoStarsText) )
   
                
                status = .false.

                if (associated(catData)) then
                    deallocate(catData)
                    nullify(catData)

                end if 
                
             


                return
            end if
        
        end if 


        if (i .eq. 4) then
            if(corrData%noMatched .lt. minNoStars) then
                write(errstr, '(a, i3, a)') "Too few sources matched for astrometry correction (", &
                          corrData%noMatched, ")"   
                 call warning(trim(errstr), trim(minNoStarsText))
   
                
                status = .false.



                if (associated(catData)) then
                    deallocate(catData)
                    nullify(catData)
                end if 

        


                return
             
            else
                status = .true.
                exit
            end if    
        end if  
      

        !***************************
        ! The new search box
        !***************************
        corrData%xShift = corrData%xShift * radiansToDegrees * 3600.0
        corrData%yShift = corrData%yShift * radiansToDegrees * 3600.0
         
        iStart = int(corrData%xShift)  - 3 
        iFinish = int(corrData%xShift) + 3
        jStart = int(corrData%yShift)  - 3
        jFinish = int(corrData%yShift) + 3

        !**********************************************************
        ! Adjust the catalogue magnitudes for the magnitude offset
        !**********************************************************
        catData%magnitude = catData%magnitude + medianMagDifference
   
        j = j + 1 
     
       write(*, *) "Line 725"

    end do




    corrData%xShift = corrData%xShift * radiansToDegrees * 3600.0
    corrData%yShift = corrData%yShift * radiansToDegrees * 3600.0
   
!write(*, *) "Line 584 corrdata%noMatched=", corrdata%noMatched, " ncat=", ncat
     forall (i = 1 : corrData%noMatched)
!   do i = 1,      corrData%noMatched
!      write(*, *) i, ") catindex=", catIndex(i), " sourceindex=", sourceIndex(i)
        xye(1, i) = xiCat(catIndex(i))    
        xye(2, i) = etaCat(catIndex(i))

        xym(1, i) = xiSource(sourceIndex(i))
        xym(2, i) = etaSource(sourceIndex(i))
       

    end forall 
!end do      
!write(*, *) "Line 595"
!    call Fit(corrData%noMatched, xyM(1, :), xye(1, :), xyM(2, :), xye(2, :), decPnt, deltaX, deltaY)
!write(*, *) "Line 714"      
    
    write(errstr, FMT='(a, f8.2, a, f8.2, a)') "Apparent image offset=", &
              corrData%xShift, ", ", corrData%yShift, " (arc secs)"
  
    call message(AppMsg, Verbose)




    write(errstr, '(a, i5)') "Number of catalogue stars in window: ", nCatInWindow
    call message(AppMsg, Verbose)
    write(errstr, '(a, i5)') "Number of objects in source list: ", noStars
    call message(AppMsg, Verbose)

    corrData%percentageMatched = 100.0 * real(corrData%noMatched) / real(noStars)
    write(errstr,'(a, f5.1, a1)') "Fraction of sources matched: ", &
                     corrData%percentageMatched, "%"

    call message(AppMsg, Verbose)

   

    !***********************************************************
    ! Find the plate solution and associated errors (determined
    ! by Monte Carlo simulations)
    !***********************************************************
    write(errstr, FMT='(a)') "Computing plate solution"
    call message(AppMsg, Verbose)             
 
    rmsResidual = 1e10
    do i = 1, 5
 
        write(errstr, FMT='(a, i1)') "Iteration no ", i
        call message(AppMsg, Verbose)             
 !write(*, *) "Line 749"

        call Solution(corrData%noMatched, xye, xym, coeffs, status, &
                       corrData%xShift, corrData%sigxShift, corrData%yShift, corrData%sigYshift, &
                       corrData%xScale, corrData%yScale, corrData%rotation, corrData%sigRotation, &
                       solutionString, decPnt)
!write(*, *) "Line 754"
        if (.not. status) then
            corrData%noMatched = 0
            corrData%rmsResidual = 0.0_double 
            exit
        end if
        !********************************************
        ! Find the goodness of  fit
        !********************************************

        call GoodnessOfFit(corrData%noMatched, xye, xym, coeffs, raPnt, decPnt, okay, corrData%rmsResidual)

 
       if (.not. okay) exit
       if (corrData%rmsResidual .ge. rmsResidual) exit
       if (abs(corrData%rmsResidual - rmsResidual) .le. 0.01_double) exit

       rmsResidual = corrData%rmsResidual  
    end do

!write(*, *) "Line 770"


   
    if (status .and. correctPositions ) then


     
        !***********************************
        ! Correct the source positions
        !***********************************

        
        call CorrectSourcePositions(nStars, raSource, decSource, &
                                    coeffs, raPnt, decPnt)

        !*******************************
        ! Show the fit
        !*******************************
        
!        do  i = 1, corrData%noMatched
                      
            !**********************************************
            ! Correct these coordinates using the fitted
            ! parameters
            !*********************************************
 
!            call sla_xy2xy(xym(1, i), xym(2, i), coeffs, xiSource(i), etaSource(i))
       
!            xiCat(i)  = xye(1, i)
!            etaCat(i) = xye(2, i)
            
!        end do


    else

        call warning("noFit", "No astrometric corrections applied")

    end if
   
 !   write(*, *) "Line 811"


    !**********************************
    ! Tidy up
    !**********************************
    deallocate(xiSource, etaSource, stat = istat)
    deallocate(catData, stat = istat)
    deallocate(separation, stat = istat)
    deallocate(xiCat, etaCat, stat = istat)
    deallocate(useCat, stat = istat)
    deallocate(xye, xym, stat = istat)
    deallocate(sourceIndex, catIndex, stat = istat)
    call CheckDeallocation(istat, "xye")
   
 !        write(*, *) "Line 827"

  
end subroutine CorrectAstrometry


!************************************************
! subroutine Solution
! 
! 
!***********************************************
subroutine Solution(npts, xye, xym, coeffs, fitOkay,        &
                         xOffset, stdxOffset, yOffset,      &
                         stdyOffset, xScale, yScale,        &
                         orient, stdOrient, solutionString, &
                         decPoint)
    integer, intent(in) :: npts
    real(double), intent(in), dimension(1 : 2, 1 : npts) :: xye, xym
    real(double), intent(out), dimension(1 : 6) :: coeffs
    real(single), intent(out) :: xOffset, yOffset, xScale, yScale, orient
    real(single), intent(out) :: stdxOffset, stdyOffset, stdOrient
    real(single) :: stdxScale, stdyScale 
    real(double)              :: xOffset1, yOffset1, xScale1, yScale1, orient1
    real(double)              :: stdxOffset1, stdyOffset1, stdxScale1, stdyScale1, stdorient1
      
    real(double), pointer :: xyeSim(:, :), xymSim(:, :) => null()
    real(double), pointer :: simCoeffs(:, :) => null()
    real(double), pointer :: data(:) => null()
    real(double), pointer :: error(:) => null()
    real(double) :: perp
    real(double) :: number
    integer i, j, k, index, nData, nSim
    integer :: istat
    logical, intent(out) :: fitOkay
    character(*), intent(out) :: solutionString
    real(double), intent(in) :: decPoint


    !*************************************************
    ! Set the number of simulations to 10 x data size,
    ! or a maximum of 1000
    !*************************************************
    nSim = min(1000, 10 * npts)

    xyeSim => Allocate2DDouble(1, 2, 1, npts, "xyeSim")
    xymSim => Allocate2DDouble(1, 2, 1, npts, "xymSim")
    simCoeffs => Allocate2DDouble(1, 5, 1, nSim, "simCoeffs")
    nData = 0


    
    !********************************************************
    ! Find the plate constants for the whole data
    !*******************************************************     
    call PlateSolution(npts, xye, xym, coeffs, fitOkay)
    call sla_dcmpf(coeffs, xOffset1,  yOffset1,  xScale1, & 
                   yScale1, perp, orient1)
    xOffset =  xOffset1
    yOffset =  yOffset1
    xScale = xScale1
    yScale = yScale1
    orient = orient1
   


    do i = 1, nSim
        do j = 1, 2
            do k = 1, npts
                call random_number(number)
                index = int(number * real(npts - 1) + 0.0001) + 1
                xyeSim(j, k) = xye(j, index)
                xymSim(j, k) = xym(j, index)
            end do
        end do
        !********************************************************
        ! Find the plate constants for this simulation
        !*******************************************************     
        call PlateSolution(npts, xyeSim, xymSim, coeffs, fitOkay)
        if (fitOkay) then
            nData = nData + 1

            !*******************************************************
            ! decompose the fit to offsets, scales and orientations
            !*******************************************************
            call sla_dcmpf(coeffs, simCoeffs(1, nData),      & 
                   simCoeffs(2, nData), simCoeffs(3, nData), & 
                   simCoeffs(4, nData), perp, &
                   simCoeffs(5, nData))
               
        end if
    end do                       

  

    if (.not. fitOkay) return
 
    !********************************************************
    ! Find the mean and standard deviation of the simulated
    ! fitted coefficients
    !********************************************************
    error => AllocateDouble(1, nData, "error")
    error = 0.0_double 
    data => AllocateDouble(1, nData, "data")
     
    forall(i = 1 : nData) data(i) = simCoeffs(1, i) 
    call ComputeAverage(data, error, nData, xOffset1, stdxOffset1)

   
    forall(i = 1 : nData) data(i) = simCoeffs(2, i) 
    call ComputeAverage(data, error, nData, yOffset1, stdyOffset1)

    forall(i = 1 : nData) data(i) = simCoeffs(3, i) 
    call ComputeAverage(data, error, nData, xScale1, stdxScale1)

    forall(i = 1 : nData) data(i) = simCoeffs(4, i) 
    call ComputeAverage(data, error, nData, yScale1, stdyScale1)

    forall(i = 1 : nData) data(i) = simCoeffs(5, i) 
    call ComputeAverage(data, error, nData, orient1, stdOrient1)

     
   


    stdxScale = real(stdxScale1)


    stdyScale = real(stdyScale1)


    stdOrient = real(stdOrient1)
 
       
    write(errstr, '(a, i5, a)') 'Plate solution (no=', npts, '): '
   
    solutionString = trim(errstr) 
    call message(AppMsg, Verbose)


    ! Correct for declination and convert to arc secs
    xOffset = ConvertToRa(xOffset1, yOffset1, decPoint)

    stdxOffset = ConvertToRa(stdxOffset1, stdyOffset1, decPoint)

    yOffset = ConvertToDec(xOffset1, yOffset1, decPoint)

   
    stdyOffset = ConvertToDec(stdxOffset1, stdyOffset1, decPoint)
    

    !*******************************************************************
    ! Uncorrect Ra offset for declination SPR 3547 November 24 2005
    !*******************************************************************
    xOffset = xoffset * cos(decPoint)
    stdxOffset = stdxoffset * cos(decPoint)
   


    write(errstr, '(a, f6.2, a, f6.2, a)') '"RA" offset=', &
               xOffset, "+/-", stdxOffset, " (arcsecs)"
    solutionString = trim(solutionString) // " " // trim(errstr)   
    call message(AppMsg, Verbose)

 
    solutionString = trim(solutionString) // " " // trim(errstr)   
  
    write(errstr, '(a, f6.2, a, f6.2, a)') 'Dec offset=', &
               yOffset, "+/-", stdyOffset, " (arcsecs)"
    solutionString = trim(solutionString) // " " // trim(errstr)   
    call message(AppMsg, Verbose)
       
   
    write(errstr, '(a, f7.4, a, f7.4 )') 'xscale=', xScale, "+/-", stdxScale
  
    call message(AppMsg, Verbose)
    write(errstr, '(a, f7.4, a, f7.4 )') 'yscale=', yScale, "+/-", stdyScale
   
    call message(AppMsg, Verbose)
    orient = orient * radiansToDegrees
    stdOrient = stdOrient * radiansToDegrees
     
    write(errstr, '(a, f6.3, a, f6.3, a)') 'rotation=', orient, "+/-", stdOrient, " (degrees)"
    solutionString = trim(solutionString) // " " // trim(errstr)   

      
    call message(AppMsg, Verbose)

    deallocate(xyeSim, xymSim, simCoeffs, data, error, stat = istat)

    call CheckDeallocation(istat, "xye, etc")


  
end subroutine Solution 
!************************************************
! subroutine PlateSolution
!***********************************************
subroutine PlateSolution(npts, xye, xym, coeffs, fitOkay)
    integer, intent(in) :: npts
    real(double), intent(in), dimension(1 : 2, 1 : npts) :: xye, xym
    real(double), intent(out), dimension(1 : 6) :: coeffs
 
    integer j
    logical, intent(out) :: fitOkay


    if ( npts .lt. 10) then ! Just do a simple least-squares fit
      
        call SimpleFit( npts, xym(1, :), xym(2, :), xye(1, :), xye(2, :), coeffs)
        fitOkay = .true.
        return
    end if



    !*******************************************************************
    ! call linear fit, assume that x,y are perpendicular (i.e. no shear)
    !******************************************************************* 
    call sla_fitxy(4, npts, xye, xym, coeffs, j)

    coeffs(5) = 0.0 ! No shear
    !********************************
    ! check that the fit was ok
    !********************************
    select case (j)
        case(-2)
            call warning("insuffData",&
                  "insufficient guide star data for astrometric fit")

            fitOkay = .false.
    
            return
        case(-3)
            call warning("singularSol","Singular solution to astrometric fit")
     
            fitOkay = .false.
            return
        case default

            fitOkay = .true.
    end select
 
     
end subroutine PlateSolution 

!******************************************
! Subroutine CorrectSourcePositions
! Corrects the standard coordinates
! NB RAS and decs are in radians
!******************************************
subroutine CorrectSourcePositions(nStars, raSource, decSource, &
                                  coeffs, &
                                  raPnt, decPnt)
    integer, intent(in) :: nStars
    real(double)  :: raSource(:)
    real(double)  :: decSource(:)
 
    real(double), dimension(1 : 6) :: coeffs
    real(double), intent(in) :: raPnt, decPnt
    integer i, j
    real(double) :: x1, x2, y1, y2
   
    write(errstr, '(a, i5)') "Correcting source positions, no of sources=", nStars
    call message(AppMsg, Verbose)

    !***************************************************
    ! correct the source positions using the fitted 
    ! parameters
    !***************************************************
    do i = 1, nStars

        !*********************************************
        ! Calaculate the standard coordinates for this 
        ! object
        !*********************************************
        call sla_ds2tp(raSource(i), decSource(i), raPnt, &
                  decPnt, x1, y1, j)

        !*********************************
        ! Check to see okay
        !********************************
        call SlaErrorCheck(j)
 
        !**********************************************
        ! Correct these coordinates using the fitted
        ! parameters
        !********************************************* 
        call sla_xy2xy(x1, y1, coeffs, x2, y2)
       
        !***********************************************
        ! Convert the corrected standard coordinates
        ! to RA and DEC
        !***********************************************
        call sla_dtp2s(x2, y2, raPnt, decPnt, &
                       raSource(i), decSource(i))       
        
    end do
end subroutine CorrectSourcePositions
!**********************************
! subroutine SlaErrorCheck
!**********************************
subroutine SlaErrorCheck(j)
    integer, intent(in) :: j

    select case (j)
        case (1) 
            call error("toofar", "Guide star too far from axis")
        case (2)
            call error("antistar", "Anti-star on tangent plane")
        case (3)
            call error("toofar", "Anti-star too far from axis")
    end select
end subroutine SlaErrorCheck
!************************************************************
! Subroutine SortOffsets
! Sorts  a i-d array of objects of type SHIFT into increasing
! radius order
!************************************************************
subroutine SortOffsets(array, size)
    integer(int32), intent(in) :: size
    type(SHIFT), intent(inout), dimension(1 : size) :: array
    integer(int32) :: i, ir, j, l
    type(SHIFT) :: rra
  
    if (size .lt. 2) return

    l = size / 2 + 1
    ir = size   

    do 
        if (l .gt. 1) then
            l = l - 1
            rra = array(l)    
        else
            rra = array(ir)
            array(ir) = array(1)
            ir = ir - 1
            if (ir .le. 1) then
                array(1) = rra
                exit
            end if
        end if

        i = l
        j = l + l
        do while(j .le. ir)
       
            if (j .lt. ir) then
                if(array(j)%radius .lt. array(j + 1)%radius) j = j + 1
            end if
            if (rra%radius .lt. array(j)%radius) then
                array(i) = array(j)
                i = j
                j = j + j
            else
                exit
            end if
            
        end do   
        array(i) = rra    
    end do
end subroutine SortOffsets

!**********************************************************
! Subroutine DisplayFit
!
!*********************************************************
subroutine DisplayFit(size, raPoint, decPoint, raSource, decSource, &
                     raCat, decCat,  &
                     object, &
                     corrData, solutionString)
    integer, intent(in) :: size
    real(double), intent(in)  :: raPoint, decPoint 
    real(double), pointer :: raSource(:)
    real(double), pointer :: decSource(:)
    real(double), pointer :: raCat(:)
    real(double), pointer :: decCat(:)
    type(CorrectionData), intent(in) :: corrData
              
    character(*), intent(in) :: object, solutionString
    real(double), pointer :: separation(:) => null()
    real(double), pointer :: error(:) => null()
    real(single), pointer :: residual(:) => null()
    real(single), pointer :: raOffset(:) => null()
    real(single), pointer :: decOffset(:) => null()
    real(single), pointer :: error1(:) => null()
    real(single), pointer :: distance(:) => null()
    real(double) :: mean, standardDeviation
    integer :: allocErr
    character(100) :: xlabel, title
    integer :: noBIns
    character(200) :: comment, comment1
    character(50) :: fileName
    real(single) :: binWidth
    integer :: i    
    real(single) xmin, xmax, ymin, ymax
    real(double) :: x2, y2

    !******************************************
    ! Plot out the residuals 
    !******************************************
 
    fileName = corrData%expID(1 : 10) // "_" // corrData%expID(11 : 13) // ".ps/VC"
 
    call pgbegin(0, trim(fileName), 1, 1)
    
    call pgsubp(1, 3)
    call pgpage
  
    separation => AllocateDouble(1, size, "separation")
    residual => AllocateSingle(1, size, "residual")
    error => AllocateDouble(1, size, "error")

    call ComputeResiduals(size, raSource, raCat, residual, mean, standardDeviation)
    

    write(comment, '(a, f8.4, a, f8.4, a)') "Mean=", mean, &
                  " (arc secs) Standard deviation=", standardDeviation, " (arc secs)"

    noBins = size / 10
    noBins = min(noBins, 20)    
    xLabel = "RA CORRECTED - RA CATALOGUE STAR (ARC SECS)"
   
    title = "OM ASTROMETRY CORRECTION"
    binWidth = 0.5
    call Histogram(size, residual, binWidth, &
                 xLabel, "", 0.1, 0.9, 0.1, 0.8)

!    call DrawNormalDistribution(size, dble(binWidth), mean, standardDeviation)
  
    call pgsch(1.5)
    call pgmtxt("t", 3.0, 0.5, 0.5, trim(title))
  

    call pgsch(1.2)
    comment1 = "OBS-ID=" // corrData%expId(1 : 10) // " OBJECT=" &
                // trim(object) //  " EXPOSURE " // corrData%expId(11 : 13) 
    call pgmtxt("t", 2.0, 0.5, 0.5, trim(comment1))
       
  

    call pgsch(1.2)
   
    call pgmtxt("t", 1.0, 0.5, 0.5, trim(solutionString))

    write(comment, '(a, f8.4, a, f8.4, a)') "Mean=", mean, &
                  " (arc secs) Standard deviation=", standardDeviation, " (arc secs)"
    call pgmtxt("t", -2.0, 0.5, 0.5, trim(comment))
  

    call pgpanl(1, 2)
    xLabel = "DEC CORRECTED - DEC CATALOGUE STAR (ARC SECS)"
   
    call ComputeResiduals(size, decSource, decCat, residual, mean, standardDeviation)
    
    binWidth = 0.5
    call Histogram(size, residual, binWidth, &
                 xLabel, "", 0.1, 0.9, 0.1, 0.8)

!    call DrawNormalDistribution(size, dble(binWidth), mean, standardDeviation)
      
    call pgsch(1.2)
   
    write(comment, '(a, f8.4, a, f8.4, a)') "Mean=", mean, &
                  " (arc secs) Standard deviation=", standardDeviation, " (arc secs)"
    call pgmtxt("t", -2.0, 0.5, 0.5, trim(comment))
  

    call pgpanl(1, 3)

    !*************************************
    ! A histogram of the separations
    !*************************************
    do i = 1, size
        separation(i) = ArcLength(raSource(i), decSource(i), & 
                        raCat(i), decCat(i))
        residual(i) = real(separation(i))
    end do 

        
    error = 1.0_double
    call ComputeAverage(separation, error, size, mean, standardDeviation)

    xLabel = "CORRECTED POSITION DEVIATIONS (ARC SECS)"
    binWidth = 0.5
    call Histogram(size, residual, binWidth, &
                 xLabel, "", 0.1, 0.9, 0.1, 0.8)

         
 
    call pgsch(1.2)
   

    write(comment, '(a, f8.4, a, f8.4, a)') "Mean=", mean, &
                  " (arc secs) Standard deviation=", standardDeviation, " (arc secs)"
    call pgmtxt("t", -2.0, 0.5, 0.5, trim(comment))
    call pgpage 

    raOffset => AllocateSingle(1, size, "raOffset")
    decOffset => AllocateSingle(1, size, "raOffset")
    error1 => AllocateSingle(1, size, "error1")
    distance => AllocateSingle(1, size, "distance")
   
    
    do  i = 1,  size
        raOffset(i) = (raSource(i) - raCat(i)) *  radiansToDegrees * 3600.0_double
        decOffset(i) = (decSource(i) - decCat(i)) * radiansToDegrees * 3600.0_double
        error1(i) = 0.0

        ! The distance from the centre in minutes of arc

        !***********************************************
        ! Convert the standard coordinates
        ! to RA and DEC
        !***********************************************
        call sla_dtp2s(raSource(i), decSource(i), raPoint, decPoint, &
                       x2, y2)       

        distance(i) = ArcLength(raPoint, decPoint, x2, y2) / 60.0_double
      
    end do
    
    call pgsubp(1, 1)
    call pgpage
  
    call DrawGraph(size, raOffset, error1, decOffset, error1, &
         title, "RA SOURCE - RA CAT (arc secs)", "DEC SOURCE - DEC CAT (arc secs)", 0.1, 0.9, 0.1, 0.9)

    call pgsch(1.0)
 


    call pgmtxt("t", 1.0, 0.5, 0.5, trim(comment1))
    call pgqwin(xmin, xmax, ymin, ymax)
    call pgmove(0.0, ymin)
    call pgdraw(0.0, ymax)
    call pgmove(xmin, 0.0)
    call pgdraw(xmax, 0.0)      
    call pgpage

    !**********************************************************************************
    !
    !*********************************************************************************
    call DrawGraph(size, raOffset, error1, distance, error1, &
         title, "RA SOURCE - RA CAT (arc secs)", "Distance from centre (arc mins)", 0.1, 0.9, 0.1, 0.9)

    call pgmove(0.0, 0.0)
    call pgdraw(0.0, 100.0)
 


    call pgpage
    !**********************************************************************************
    !
    !*********************************************************************************
    call DrawGraph(size, decoffset, error1, distance, error1, &
         title, "DEC SOURCE - DEC CAT (arc secs)", "Distance from centre (arc mins)", 0.1, 0.9, 0.1, 0.9)

    call pgmove(0.0, 0.0)
    call pgdraw(0.0, 100.0)
 

    call pgclos    

    !*********************************************
    ! save the data
    !
    !*********************************************
!    call SaveData(corrData, size,  raOffset, decOffset, &
!                  distance)

    deallocate(separation, residual, error, stat = allocerr)
    deallocate(raOffset, decOffset, error1, distance, stat = allocerr)
    
    
 
          
end subroutine DisplayFit
!**************************************
! Subroutine ComputeResiduals
!*************************************
subroutine ComputeResiduals(size, obsData, fitData, residual, &
                           mean, standardDeviation)
    integer, intent(in) :: size
    real(double), pointer :: obsData(:), fitData(:)
    real(single), intent(out), dimension(1 : size) :: residual
    real(double), intent(out) :: mean, standardDeviation
    real(double) :: diff, sum1
    integer :: i

    mean = 0.0_double
    
    do i = 1, size
         diff = obsData(i) - fitData(i)
         residual(i) = diff * radiansToDegrees * 3600.0_double
         mean = mean + diff   
    end do
    mean = (mean / dble(size))  * radiansToDegrees * 3600.0_double

    sum1 = 0.0_double
    

    do i = 1, size
        diff = mean - residual(i)
        sum1 = sum1 + diff * diff 
    end do
  
    standardDeviation = dsqrt(sum1 / dble(size - 1))  

end subroutine ComputeResiduals

!*******************************************
! Subroutine SortCatalogue
!*******************************************
subroutine SortCatalogue(size, ra, dec, mag)
    integer, intent(in) :: size
    real(double) :: ra(:)
    real(double) :: dec(:)
    real(single) :: mag(:)
    real(double), pointer :: temp(:) => null()
    integer(int32), pointer :: index(:) =>  null()
        
    integer :: i, status

    temp => AllocateDouble(1, size,  "temp/SortCatalogue")


    index => AllocateInt32(1, size,  "index/SortCatalogue")


    temp(1 : size) = ra(1 : size)
  
    call IndexSort2(temp, index, size)
    
    
    forall(i = 1 : size) ra(i) = temp(index(i))

    temp(1 : size) = dec(1 : size)
    forall(i = 1 : size) dec(i) = temp(index(i))

    temp(1 : size) = mag(1 : size)
    forall(i = 1 : size) mag(i) = real(temp(index(i)))

    deAllocate(temp, index, stat = status)

    call CheckdeAllocation(status, "size/SortCatalogue")
  

end subroutine SortCatalogue

!*****************************************
! Subroutine SetUpOffsets
! Sets up an array of ra-dec offsets to be
! used when matching sources with catalogue
! stars
!*****************************************
subroutine SetUpOffsets(noIntervals, iStart, iFinish, jStart, &
                         jFinish, noOffsets, offset)
    integer, intent(in) :: noIntervals
    integer, intent(in) :: iStart, iFinish, jStart, jFinish
    integer, intent(out) :: noOffsets
    type(SHIFT), intent(inout), dimension(:) :: offset 
    integer i, j
    integer :: iMax, jMax
    real :: x, y
    real :: radius
    real(single) :: constant
        
    constant = degreesToRadians / 3600.0
   
    noOffsets = 0
    iMax = (abs(iStart) + abs(iFinish)) * noIntervals + 1
    jMax = (abs(jStart) + abs(jFinish)) * noIntervals + 1
  
    do i = 1, iMax
        x = real(iStart) + real(i - 1) / real(noIntervals)

        do j = 1, jMax
            y = real(jStart) + real(j - 1) / real(noIntervals)
    
            radius  = sqrt(x * x + y * y) 
            noOffsets = noOffsets + 1
            offSet(noOffsets)%x = x * constant
            offSet(noOffsets)%y = y * constant
            offSet(noOffsets)%radius = radius
            
        end do
    end do
 

    !*************************************************
    ! Sort these offsets into increasing radius order
    !************************************************
    call SortOffsets(offset, noOffsets)
  
 
end subroutine SetUpOffSets
!*************************************
! Function GetCatalogue data
! Reads in the RA, dec and magnitudes
! from the fits catalogue file
!*************************************
function GetCatalogueData(fileName, obsEpoch, nCat) result(catData)
    character(*), intent(in) :: fileName
    real(double), intent(in) :: obsEpoch
    integer(int32), intent(out) :: nCat   
    type(CatalogueData), dimension(:), pointer  :: catData
    type(CatalogueData), dimension(:), pointer  :: tempcatData => null()
    
 
    type(DataSetT) :: catDataSetH
    type(TableT) :: inTableH
    integer :: i, n1, n2
    real(double)    :: ep1
   

    ep1 = 2000.0_double
    
  
    n1 = 0 
    n2 = 0
    catDataSetH = dataSet(fileName, read) 

     ! 
    if (hasBlock(catDataSetH, "SRCRES")) then
        inTableH = table(catDataSetH, "SRCRES")
    else if(hasBlock(catDataSetH, "RAWRES")) then
        inTableH = table(catDataSetH, "RAWRES")
    else
        call fatal("Catalogue file", "No source-list table")
    end if

    nCat = numberOfRows(inTableH)


    allocate(catData(1 : nCat))

    call Read1Real64Column(inTableH, "CAT_RA", nCat, catData%RA)


    call Read1Real64Column(inTableH, "CAT_DEC", nCat, catData%dec)



    ! Load the catalogue magnitudes
    catData%magnitude = 0.0
    if (HasColumn(inTableH, "Bmag")) then
        call Read1Real32Column(inTableH, "Bmag", nCat, catData%magnitude)
    else if (HasColumn(inTableH, "BMAG")) then
        call Read1Real32Column(inTableH, "BMAG", nCat, catData%magnitude)
    else if (HasColumn(inTableH, "Rmag")) then       
        call Read1Real32Column(inTableH, "Rmag", nCat, catData%magnitude)
    else
        call Read1Real32Column(inTableH, "RMAG", nCat, catData%magnitude)
    end if

    call release(inTableH)
    call release(catDataSetH)


  
    catData%error = 1.0  ! Set the position error to 1 arcsec    
 

    !*********************************************
    ! Sort the catalogue into RA increasing order
    !********************************************
    call SortCatalogue(nCat, catData%RA, catData%dec, catData%magnitude)


    !*********************************************************
    ! Need to check to see if there are RAS near zero degrees
    !*********************************************************
    
    if (catData(1)%RA .lt. 0.006) then
        allocate(tempCatData(1 : nCat))
        tempCatData = catData
        deallocate(catData)
        n1 = 10 
        nCat = nCat + n1
        allocate(catData(1 : nCat))
      
        ! Move them to the right by 10
      
        do i = n1 + 1,  nCat
            catData(i) = tempcatData(i - n1)
        end do
 
  
        deallocate(tempCatData)

        do i = 1, n1
            catData(i) = catData(nCat - n1 + i)
            catData(i)%RA =  catData(i)%RA - 360.0
     
        end do
       

    end if  
    

    !*******************************************************
    ! Need to check to see if there are RAS near 360 degrees
    !*******************************************************
    if (catData(nCat)%RA .gt. 359.994) then
        allocate(tempCatData(1 : nCat))
        tempCatData = catData
        deallocate(catData)
        n2 = 10
        nCat = nCat + n2
        allocate(catData(1 : nCat))  
        catData(1 : nCat - n2) = tempCatData
  
        do i = 1, n2
            catData(nCat - n2 + i) = catData(i)
             catData(nCat - n2 + i)%RA =  catData(i)%RA + 360.0
        end do
        deallocate(tempCatData)
    
    end if  

    

    !******************************************
    ! Precess the catalogue coordinates to the 
    ! observation date
    !*********************************************
    catData%RA  = catData%RA * degreesToRadians
    catData%dec = catData%dec * degreesToRadians

    if (ep1 .ne. obsEpoch) then
        call warning("Precessing the catalogue stars") 
   
        do i = 1, nCat

            !***************************************
            ! precess the gs coords to observation epoch
            !***************************************
            call sla_preces("FK5", ep1, obsEpoch, catData(i)%RA, catData(i)%dec)

        end do
    end if


end function GetCatalogueData
!******************************************
! Subroutine FindBestOffsets 
!******************************************
subroutine FindBestOffsets(noIntervals, iStart, iFinish, jStart, jFinish,         &
                           noStars, raSource, decSource, posErr, sourceMagnitude, & 
                           nCat, raCat, decCat, magCat, minMagLimit, maxmagLimit, &
                           maxMagDifference, nSigma, tolerance, maxNoMatched,     &
                           xShift, yShift, rmsResidual, maxLikelihood,            &
                           medianPositionDifference, medianMagDifference,         &
                           sourceIndex, catIndex)

    integer, intent(in) :: noIntervals
    integer, intent(in) :: iStart
    integer, intent(in) :: iFinish
    integer, intent(in) :: jStart
    integer, intent(in) :: jFinish
    integer, intent(in) :: noStars
    real(double)     :: raSource(:)
    real(double)     :: decSource(:) 
    real(single)     :: posErr(:) 
    real(single)     :: sourceMagnitude(:)  
    integer(int32), intent(in) :: nCat 
    real(double)  :: raCat(:)     ! RAs of the catalogue stars
    real(double)  :: decCat(:)    ! decs of the catalogue stars
    real(single)  :: magCat(:)    ! mags of the catalogue stars 
    real(single), intent(in) ::  minMagLimit, maxMagLimit
    real(single), intent(in) ::  maxMagDifference
    integer, intent(in) :: nSigma
    real(single), intent(in) :: tolerance          ! Tolerance for catalog star search (arcsec) was 2.

    integer(int32), intent(out) :: maxNoMatched
    real(single), intent(out) :: xShift, yShift 
    real(double), intent(out) :: rmsResidual
    real(double), intent(out) :: maxLikelihood

    real(single), intent(out) :: medianPositionDifference
    real(single), intent(out) :: medianMagDifference
    integer, intent(out), dimension(1 : noStars) :: sourceIndex
    integer, intent(out), dimension(1 : noStars) :: catIndex

    !*********************************
    ! Local variables
    !********************************* 
    real(single), pointer :: magnitudeOffset(:) => null()
    real(single), pointer :: separation(:) => null()
    real(single) :: fractionFound
    real(single) :: tol, mindist                  ! distances to GS
    real(single) :: spacing
    real(single) :: xiShift
    real(single) :: etaShift   
    real(double) :: sum
   
    real(double) :: weight
    real(double) :: sumWeight
    real(double) :: likelihood
    real(double) :: sigmaSquared    
 
    integer :: i
    integer :: k, l
    integer :: noOffsets
    integer :: iStat, noMatched
    logical, pointer :: catMatched(:) => null()
    Type(SHIFT), dimension(:), allocatable  :: offSet 
    
    spacing = 1.0 / real(noIntervals)
    !******************************************************
    ! Set up the ra and dec offsets to be used in matching
    ! the sources with the catalogue stars
    !******************************************************
    noOffsets = ((abs(iStart) + abs(iFinish) + 1) * noIntervals) * &
               ((abs(jStart) + abs(jFinish) + 1) * noIntervals)  
    

    allocate(offSet(1 : noOffsets), stat = istat)

    call CheckAllocation(istat, "offSet")
    allocate(catMatched(1 : nCat), stat = istat)
    call CheckAllocation(istat, "catMatched")
 

    call SetUpOffsets(noIntervals, iStart, iFinish, jStart, jFinish, noOffsets, offset)
    magnitudeOffset => AllocateSingle(1, noStars, "magnitudeOffset")
    separation => AllocateSingle(1, noStars, "position")

    medianPositionDifference = 0.0
    medianMagDifference = 0.0
      
    rmsResidual = 999.9
    fractionFound = 0.0
    maxLikelihood = 0.0_double   
    maxNoMatched = 0
    !********************************************************
    ! Loop through all the offsets to find the best match
    ! between the sources and catalogue objects
    !********************************************************

    do i = 1, noOffsets 
      
        xiShift = offSet(i)%x
        etaShift = offset(i)%y
       

        noMatched = 0
        sum = 0.0_double
        sumWeight = 0.0_double        
        likelihood = 0.0_double
        catMatched = .false.
 

        ! Loop through the stars to try to match them with catalogue stars 
        do k = 1, noStars       
 
            if (sourceMagnitude(k) &
                .gt. maxmagLimit .or. sourceMagnitude(k) .lt. minMagLimit) cycle

            !**************************************************
            ! Find the nearest catalogue object to the source,
            ! checking that the magnitude difference is within
            ! the allowed range
            !**************************************************

            tol = max(real(nSigma) * posErr(k), tolerance)
        
            if (SourceMatch(raSource(k) + dble(xiShift), decSource(k) + dble(etaShift), &
                          sourceMagnitude(k), &
                          nCat, raCat, decCat, magCat, catMatched, tol, &
                          maxMagDifference, minDist, l)) then
                
                noMatched = noMatched + 1            ! The number of sources matched       
                magnitudeOffset(noMatched) = sourceMagnitude(k) - magCat(l) 
                sigmaSquared = max(posErr(k) * posErr(k), 1.0)
        
                weight = 1.0_double / sigmaSquared
                sum = sum + weight * minDist * minDist
                sumWeight = sumWeight + weight
!                likelihood = likelihood + likeLi 

            end if

        end do
  
        if (noMatched .gt. maxNoMatched) then
!        if (likelihood .gt. maxLikelihood) then

            sum = sqrt(sum / sumWeight)

            if (sum .lt. rmsResidual) then

                rmsResidual = sum
                maxNoMatched = noMatched
                xShift = xiShift
                yShift = etaShift          
                maxLikelihood = likelihood
            end if        
        end if
    end do ! End of looping through offsets


    ! Match them again to get various parameters

    noMatched = 0

    catMatched = .false.
    do k = 1, noStars       
   
       if (sourceMagnitude(k) .gt. maxmagLimit &
                            .or. sourcemagnitude(k) .lt. minMagLimit) cycle

        !**************************************************
        ! Find the nearest catalogue object to the source,
        ! checking that the magnitude difference is within
        ! the allowed range
        !**************************************************

        tol = max(real(nSigma) * posErr(k), tolerance)

     
        if (SourceMatch(raSource(k) + dble(xShift), decSource(k) + dble(yShift), &
             sourceMagnitude(k), nCat, raCat, decCat, magCat, catMatched, tol, &
             maxMagDifference, minDist, l)) then
                
            noMatched = noMatched + 1            ! The number of sources matched       
            magnitudeOffset(noMatched) = sourceMagnitude(k) - magCat(l)
            separation(noMatched) = minDist
            sourceIndex(noMatched) = k
            catIndex(noMatched) = l 
        
        end if

    end do
       
    medianPositionDifference = MedianValue(separation(1 : noMatched), noMatched)
    medianMagDifference = MedianValue(magnitudeOffset(1 : noMatched), noMatched)
              
    fractionFound = real(noMatched) / real(noStars)
              
    if (noMatched .ne. maxNoMatched) call fatal("no found ne maxnoFound")
            
    deallocate(magnitudeOffset, separation, offset, catMatched)

end subroutine FindBestOffsets
!************************************************
! subroutine GoodnessOfFit
! xye(1, i) - xi catalogue
! xye(2, i) - eta catalogue
! xym(1, i) - xi source
! xym(2, i) - eta source
! coeffs(1-6) - the fitted coefficients
!***********************************************
subroutine GoodnessOfFit(npts, xye, xym, coeffs, raPnt, &
              decPnt, okay, rmsResidual)
    integer, intent(inout) :: npts
!    real(single) :: weight(:)
    real(double), intent(inout), dimension(1 : 2, 1 : npts) :: xye, xym
    real(double), intent(in), dimension(1 : 6) :: coeffs
    real(double), intent(in) :: raPnt, decPnt
    logical, intent(out) :: okay
    real(double), intent(out) :: rmsResidual
    real(single), pointer :: raSource(:)
    real(single), pointer :: decSource(:)
    real(single), pointer :: raCat(:)
    real(single), pointer :: decCat(:)
    real(double), pointer :: distance(:)
    real(double), pointer :: error(:)

    real(double) :: x, y, x1, y1
    real(double) :: mean, standardDeviation
    real(double) :: rmsResidual1, dist, tolerance
    real(double) :: sumWeight
    integer :: i, k, count
    integer :: nSigma

    
    raSource => AllocateSingle(1, npts,  "raSource")
    decSource => AllocateSingle(1, npts,  "decSource")

    raCat => AllocateSingle(1, npts,  "raCat")
    decCat => AllocateSingle(1, npts,  "decCat")
    distance => AllocateDouble(1, npts,  "distance")
   

    error => AllocateDouble(1, npts,  "error")



    rmsResidual = 0.0_double
    rmsResidual1 = 0.0_double
    sumWeight = 0.0_double
    !***************************************************
    ! correct the source positions using the fitted 
    ! parameters
    !***************************************************
    do i = 1, npts


        !**************************************************
        ! Convert the standard coordinates of the catalogue
        ! objects to RA and DEC
        !**************************************************
        call sla_dtp2s(xye(1, i), xye(2, i), raPnt, decPnt, x, y)

        raCat(i) = real(x)
        decCat(i) = real(y)       
     
          
        !***********************************************
        ! Convert the source standard coordinates
        ! to RA and DEC
        !***********************************************
        call sla_dtp2s(xym(1, i), xym(2, i), raPnt, decPnt, x1, y1)
              

        dist = ArcLength(x1, y1, dble(raCat(i)), dble(decCat(i)))
!        sumWeight = sumWeight + weight(i) 
!        rmsResidual1 = rmsResidual1 + dble(weight(i)) * dist * dist
        rmsResidual1 = rmsResidual1 + dist * dist
 

        !**********************************************
        ! Correct the source coordinates
        ! 
        !********************************************* 
        call sla_xy2xy(xym(1, i), xym(2, i), coeffs, x, y)
   
     

   
        !***********************************************
        ! Convert the corrected standard coordinates
        ! to RA and DEC
        !***********************************************
        call sla_dtp2s(x, y, raPnt, decPnt, x1, y1)
        raSource(i) = real(x1) 
        decSource(i) = real(y1)       
        distance(i) = ArcLength(dble(raSource(i)), dble(decSource(i)), dble(raCat(i)), dble(decCat(i)))
!        rmsResidual = rmsResidual + dble(weight(i)) * distance(i) * distance(i)
       rmsResidual = rmsResidual + distance(i) * distance(i)
    
    end do
!    rmsResidual1 = rmsResidual1 / sumWeight
!    rmsResidual = rmsResidual / sumWeight


    error = 1.0_double
    

    call ComputeAverage(distance, error, npts, mean, standardDeviation)
!    write(*, *) "Goodnessoffit- mean=", mean, " standard deviation=", standardDeviation


    rmsResidual = sqrt(rmsResidual / dble(npts))
    rmsResidual1 = sqrt(rmsResidual1 / dble(npts))
    write(errstr, '(a, f5.1, a)') "RMS residual before correction=", rmsResidual1, " arc secs"
    call message(AppMsg, Verbose)
  

    write(errstr, '(a, f5.1, a)') "RMS residual after correction=", rmsResidual, " arc secs" 
    call message(AppMsg, Verbose)
        
    if (rmsResidual .gt. 4.0) then
        nsigma = 1
    else if(rmsResidual .gt. 2.0) then
        nsigma = 2
    else
        nsigma = 3
    end if
    okay = .false.

    if (npts .gt. 20) then
        
        count = 0
        
        tolerance = min(3.0_double, dble(nSigma) * rmsResidual)

        do i = 1, npts
            if (distance(i) .gt. rmsResidual) count = count + 1
        end do
        write(errstr, '(a, i5, a, i1, a)') "No of data rejected = ", count, " (", nSigma, " x rmsresidual)"
        call message(AppMsg, Verbose)
              
        !****************************************************
        ! Exclude data with residuals > nSigma * rmsResidual
        !****************************************************

        if (real(npts  - count) / real(npts) .gt. 0.5) then
            k = 0
            okay = .true.
              
            do i = 1, npts
                if (distance(i) .le. tolerance) then
                    k = k + 1
                    xye(1, k) = xye(1, i)
                    xye(2, k) = xye(2, i)
                    xym(1, k) = xym(1, i)
                    xym(2, k) = xym(2, i)
!                    weight(k) = weight(i)
                end if 


            end do

            if (npts .eq. k) okay = .false.

            npts = k   ! The new number of points
        end if
    end if

    deallocate(raSource, decSource, raCat, decCat)
    deallocate(distance, error)
end subroutine GoodnessOfFit


!*****************************************************
! Subroutine AddCorrectionColumns
!*****************************************************
subroutine AddCorrectionColumns(fileName, size, raSource, decSource, &
                                corrData)

    character(*), intent(in) :: fileName
    integer(int32), intent(in) :: size
    real(double), pointer :: raSource(:)
    real(double), pointer :: decSource(:) 
    type(CorrectionData), intent(in) :: corrData
   
    real(single) :: nullValue
    integer(int32) :: position
    type(DataSetT) :: set
    type(TableT)   :: sourceTable

    nullvalue = -real(99999)
    set = dataSet(fileName, MODIFY)
    sourceTable = table(set, "SRCLIST", MODIFY)



!    call setAttribute(sourceTable, "RAOFFSET", corrData%xShift, "arcsecs", "RA correction applied")
!    call setAttribute(sourceTable, "RAOFFERR", corrData%sigXShift, "arcsecs", "Error in RA correction")
!    call setAttribute(sourceTable, "DEOFFSET", corrData%yShift, "arcsecs", "Dec correction applied")
!    call setAttribute(sourceTable, "DEOFFERR", corrData%sigYShift, "arcsecs", "Error in dec correction")



    call setAttribute(sourceTable, "RAOFFSET", corrData%xShift, "arcsec", "X offset at pointing declination")
    call setAttribute(sourceTable, "RAOFFERR", corrData%sigXShift, "arcsec", "Error in X offset")
    call setAttribute(sourceTable, "DEOFFSET", corrData%yShift, "arcsec", "Y offset at pointing declination")
    call setAttribute(sourceTable, "DEOFFERR", corrData%sigYShift, "arcsec", "Error in Y offset")

 




    call setAttribute(sourceTable, "ROT_CORR", corrData%rotation, "degrees", "Rotation correction applied")
    call setAttribute(sourceTable, "ROT_ERR",  corrData%sigRotation, "degrees", "Error in rotation correction")
    call setAttribute(sourceTable, "NMATCHES", corrData%noMatched, "", "Number of sources matched")
    call setAttribute(sourceTable, "RMSRESID", corrData%rmsResidual, "arcsecs", "Root-mean-square residual")



       
    position = columnNumber(sourceTable, "BII") + 1

    call AddReal64Column(sourceTable, decSource, size, "DEC_CORR", "degrees", &
            "Corrected for boresight offset", nullValue, position)
  
    call AddReal64Column(sourceTable, raSource, size, "RA_CORR", "degrees", &
            "Corrected for boresight offset", nullValue, position)
  
    call release(sourceTable)
    call release(set)

end subroutine AddCorrectionColumns

!*****************************************************
! Subroutine AddCorrectionColumns
!*****************************************************
subroutine AddCorrectionColumns2(sourceTable, size, raSource, decSource, &
                                corrData)
    type(TableT), intent(inout)   :: sourceTable


    integer(int32), intent(in) :: size
    real(double), pointer :: raSource(:)
    real(double), pointer :: decSource(:) 
    type(CorrectionData), intent(in) :: corrData
   
    real(single) :: nullValue
    integer(int32) :: position
   
    nullvalue = -real(99999)
    
    call setAttribute(sourceTable, "RAOFFSET", corrData%xShift, "arcsec", "X offset at pointing declination")
    call setAttribute(sourceTable, "RAOFFERR", corrData%sigXShift, "arcsec", "Error in X offset")
    call setAttribute(sourceTable, "DEOFFSET", corrData%yShift, "arcsec", "Y offset at pointing declination")
    call setAttribute(sourceTable, "DEOFFERR", corrData%sigYShift, "arcsec", "Error in Y offset")

    call addcomment(sourceTable, "NB to obtain the actual RA/DEC corrections applied divide the offsets by cos(dec)")

    call setAttribute(sourceTable, "ROT_CORR", corrData%rotation, "degrees", "Rotation correction applied")
    call setAttribute(sourceTable, "ROT_ERR",  corrData%sigRotation, "degrees", "Error in rotation correction")
    call setAttribute(sourceTable, "NMATCHES", corrData%noMatched, "number", "Number of sources matched")
    call setAttribute(sourceTable, "RMSRESID", corrData%rmsResidual, "arcsecs", "Root-mean-square residual")

    position = columnNumber(sourceTable, "DEC") + 1

    call AddReal64Column(sourceTable, decSource, size, "DEC_CORR", "deg", &
            "Corrected for boresight offset", nullValue, position)
  
    call AddReal64Column(sourceTable, raSource, size, "RA_CORR", "deg", &
            "Corrected for boresight offset", nullValue, position)
  
   

end subroutine AddCorrectionColumns2

!***************************************************
! Function SourceMatch
! Tries to match a catalogue star to a given source
!
!***************************************************
logical function SourceMatch(raXMM, decXMM, magXMM, &
                           size, raCat, decCat, magCat, catMatched, tolerance, &
                           maxMagDifference, minRadius, index)
 
    implicit none 
    real(double), intent(in) :: raXMM, decXMM
    real(single), intent(in) ::  magXMM      ! XMM ra, dec & magnitude
    
    integer(int32), intent(in) :: size

    real(double) :: raCat(:), decCat(:)
    real(single) :: magCat(:)
    logical  :: catMatched(:)

!    type(CatalogueData), intent(in), dimension(1 : size) :: catData
    real(single), intent(in) :: tolerance
  
   
    real(single), intent(in) :: maxMagDifference 
    integer, intent(out) :: index     
!    real(double), intent(out) :: likelihood
    real(single), intent(out) :: minRadius
    real(double) :: distance
    real(double) :: delta
    real(double) :: sigmaSquared
    
 
    real(double), parameter         :: degreesToRadians = PI / 180.0_double

    integer i, kl, ku, jStart

    jStart = 1   

    delta = (tolerance / 3600.0_double) * degreesToRadians 
    sigmaSquared = tolerance * tolerance
 
    ! Bracket the catalogue stars that might match the star
    kl = Locate(size, raCat, raXMM - delta, jStart)
    ku = Locate(size, raCat, raXMM + delta, jStart)
                   
    minRadius = 10000.0
    SourceMatch = .false.
!    likelihood = 0.0_double


    do i = kl, ku
        
        if (abs(decCat(i) - decXMM) .gt. delta .or. catMatched(i)) cycle
        

        !*******************************************
        ! Check the magnitudes
        !******************************************* 
        if(abs(magCat(i) - magXMM) .le. maxMagDifference) then
            
            distance = ArcLength(raXMM, decXMM, raCat(i), decCat(i))
            if (distance .le. tolerance) then  
!                likelihood = likelihood + exp(-0.5_double * distance * distance / sigmaSquared) 
 
           

                if(distance .lt. minRadius) then
                    minRadius = distance
                    SourceMatch = .true.
                    index = i
                end if

                
            end if
        end if
    end do
    if (SourceMatch) catMatched(index) = .true.
end function SourceMatch   


!*******************************************************
! ArcLength - returns the distance (arc secs) between 2
! sources with coords (ra1, dec1) & (ra2, dec2)
! Returns the distance in arc seconds.
!*******************************************************
real(double) function ArcLength(ra1, dec1, ra2, dec2)
    real(double), intent(in) ::  ra1, ra2
    real(double), intent(in) ::  dec1, dec2
   

    ArcLength = sla_DSEP(ra1, dec1, ra2, dec2) * radiansToArcSeconds
end function ArcLength

!**************************************************
! Function to convert dx into dra
! xo & y0 is in standard coordinates in radians
!**************************************************
real(single) Function ConvertToRa(x, y, delta0)
    real(double), intent(in) :: x, y, delta0
    real(double) :: secDelta
    
    if (delta0 ==  PI) then
        ConvertToRa = 100.0
    else 
        secDelta = 1.0_double / dcos(delta0)
 

        ConvertToRa = real((x * secDelta - &
                  x * x * x * secDelta * (secDelta * secDelta - 3.0_double) / 6.0_double - &
                  0.5_double * x * y * y * secDelta) * radiansToDegrees * 3600.0_double)
    end if
end function ConvertToRa

!**************************************************
! Function to convert dy into ddec
! x0 & yo is in standard coordinates in radians
!**************************************************
real(single) Function ConvertToDec(x, y, delta0)
    real(double), intent(in) :: x, y, delta0
    
  
    ConvertToDec = real((y - 0.5_double * x * x * dtan(delta0) - &
               y * y * y / 3.0_double) * radiansToDegrees * 3600.0_double)

end function ConvertToDec

!*************************************************************
! Computes the RA and dec offsets assuming no rotation, etc
!
!*************************************************************
subroutine SimpleFit(nStars, raOm, decOm, raCat, decCat, coeffs)
    integer, intent(in) :: nStars
    real(double), intent(in), dimension(:) :: raOm, raCat, decOm, decCat
   
    real(double), intent(out) :: coeffs(1 : 6)
    real(double) :: sum1, sum2
 
    sum1 = sum(raOm(1 : nStars))
    sum2 = sum(raCat(1 : nStars))

    coeffs(1) = (sum2 - sum1) / dble(nStars)

    coeffs(2) = 1.0_double
    coeffs(3) = 0.0_double

    sum1 = sum(decOm(1 : nStars))
    sum2 = sum(decCat(1 : nStars))

    coeffs(4) = (sum2 - sum1) / dble(nStars)
    coeffs(5) = 0.0_double
    coeffs(6) = 1.0_double

end subroutine SimpleFit
!*****************************************
! Function FitOkay
!***************************************
logical Function FitOkay(corrData, maxRADecErr, &
           maxRMSResidual)  
    type(CorrectionData), intent(in) :: corrData
    real(single), intent(in) :: maxRADecErr, maxRMSResidual 
  
    if (corrData%sigXShift .le.  maxRADecErr .and. &
        corrData%sigYShift .le.  maxRADecErr .and. &
        corrData%rmsResidual .le. maxRMSResidual) then
        FitOkay = .true.
    else
       FitOkay = .false.
    end if
end Function FitOkay           

end module omastromcorrect_mod






