!**************************************************************
! module omdetectcoinccorr_mod.f90 (15th August 2002)
!
! Contains routines for correcting count rates
! for coincidence losses
!
!**************************************************************
module omcoinccorr_mod
    
    use param
    use cal
    use dal
    use omsourceparam_mod




!*************************************************
!
!************************************************
real(single), private :: significanceRadius = 6.0



contains               

!***************************************************
! SetSignificanceRadius
! Set the radius for which the source-significance
! is computed (signal-to-noise)
! The radii are hard-wired so that for "V" the
! radius is 6 unbinned pixels, and for the other
! OM filters the radius is such that the fraction
! of the total counts contained within that radius
! is the same as the V-band value (95.4%)
! 
!
!***************************************************
subroutine setSignificanceRadius(filter)
    character(*), intent(in) :: filter


    select case(trim(filter))
    
        case ("U")
            significanceRadius = 6.48
        case ("B")
            significanceRadius = 6.66

        case ("V")
            significanceRadius = 6.00

        case ("UVW2")
            significanceRadius = 8.38

        case ("UVM2")
            significanceRadius = 8.41
        case ("UVW1")
            significanceRadius = 8.72

        case default
            significanceRadius = 6.0

    end select



end subroutine setSignificanceRadius

!*********************************************************
! Function CoincidenceLossCorrection
!   
!*********************************************************
real(double) function CoincidenceLossCorrection(rawCountRate, & 
                  countRateError, frameTime)
    real(double), intent(in)     :: rawCountRate
    real(single), intent(in)     :: countRateError
    real(double), intent(in)     :: frameTime
    real(single) :: tempRawRate
    
    type(ValueAsymmetricErrorT) :: rawRate
    type(ValueAsymmetricErrorT) :: corrRate
 
    if (rawCountRate .le. 0.0) then
        CoincidenceLossCorrection = 0.0_double
        return
    end if

    !*************************************************************
    ! NB frame time is in milli-seconds
    !*************************************************************    

    tempRawRate = min(0.9999_double / frameTime, dble(rawCountRate))

    if (rawCountRate .gt. 1.0_double / frameTime) then
        CoincidenceLossCorrection = 1.0_double
        return
    end if
    
    rawRate = ValueAsymmetricErrorT(temprawRate, dble(countRateError), dble(countRateError))





    !******************************************************************************************
    ! NB the final boolean parameter switches on the empirical correction
    !*****************************************************************************************
    corrRate = CAL_omPhotoNatural(rawRate, frameTime, 0.0_double, 0.0_double, .FALSE., .TRUE.)
 

    !****************************************************************
    ! If the count rate is greater than the frame rate then the above
    ! routine cannot do the correction- instead it sets the corrected 
    ! count rate to the largest number that can be stored on the computer. 
    !*****************************************************************


    if (CorrRate%value .lt. 1e10) then
        CoincidenceLossCorrection = corrRate%value / rawRate%value
       
    else
        CoincidenceLossCorrection = 1.0_double
    end if
   
    

end function CoincidenceLossCorrection





!****************************************************
! subroutine PointSourceCoincLossCorr
! Corrects a point-source count-rate for coincidence-losses
! 
!****************************************************
subroutine PointSourceCoincLossCorr(source, coincidenceLossRadius, frameTime)     
    type(SourceParameters), intent(inout) :: source

    real(single), intent(in) :: coincidenceLossRadius

    real(double), intent(in) :: frameTime
    real(single) :: countFrameRatio, countsRatio
    real(double) :: totalRate
    real(single) :: error, totRate
    real(double) :: corrRate, corrRateErr
    real(double) :: corrRateErrMinus1, corrRateErrPlus1
    real(double) :: corrRateErrMinus2, corrRateErrPlus2
    real(double) :: oldCorrRate  
    real(single) :: sourceRate, backgroundRate

!integer, save :: count= 0 

!logical check

!check = .false.
!if (abs(source%xpos - 704.7) .lt. 2.0 .and. abs(source%ypos - 313.0) .lt. 2.0) check = .true.

!if (check) then


!write(*, *) "InPointsourceCoincLossCorr line 106 RAW RATES"
!write(*, *) "PointsourceCoincLossRadius- coincidence-loss radius=",   coincidenceLossRadius
!write(*, *) "x=", source%xpos, " y=", source%ypos
!write(*, *) "Aperture radius=", source%apertureRadius
!write(*, *) "raw count rate=", source%rawCountRate
!write(*, *) "raw background count rate=", source%rawbackgroundCountRate
!write(*, *) "raw background count ratev err=", source%rawbackgroundCountRateerr
!end if
    


    !*********************************************************************************************
    ! 21st May 2006 compute the significance for a 3 arc sec aperture at V
    ! 
    !*********************************************************************************************
    if ( abs(source%apertureRadius - significanceRadius) > 0.1 ) then

        countFrameRatio = min(real(source%rawCountRate * frameTime), 1.0)       
        countFrameRatio = max(0.0, countFrameRatio)
      

        sourceRate = source%rawCountRate * PSFCountsRatio( 6.0, source%apertureRadius, &
                             source%xPos, source%yPos, countFrameRatio)

        backgroundRate = source%rawBackgroundCountRate * 36.0 / ( source%apertureRadius * source%apertureRadius)

        source%significance = computeSourceSignificance(sourceRate, backgroundRate, source%exposure)


    else
        source%significance = computeSourceSignificance(source%rawCountRate, source%rawBackgroundCountRate, source%exposure)
    end if



    !**********************************************************************
    ! Check to see if the aperture radius was less than the
    ! coincidence-loss radius- apply a correction if so
    !*********************************************************************
    if (source%apertureRadius .lt. coincidenceLossRadius) then

        !*********************************************************
        ! Extrapolate counts from the omdetect aperture radius
        ! to 12-pixel radius and also the background counts
        !**********************************************************

        countFrameRatio = min(real(source%rawCountRate * frameTime), 1.0)       
        countFrameRatio = max(0.0, countFrameRatio)
        countsRatio = PSFCountsRatio(coincidenceLossRadius, source%apertureRadius, &
                             source%xPos, source%yPos, countFrameRatio)


        source%corrCountRate = source%rawCountRate * countsRatio
!  write(*, *) "Line 110 source%corrCountRateerr=",    source%corrCountRateErr


        source%corrCountRateErr = source%rawCountRateErr * countsRatio

!    write(*, *) "Line 125 source%corrCountRateerr=",    source%corrCountRateErr, " countsratio=", countsRatio 

!write(*, *) "Line 118 rawbackgroundrate=", source%rawbackgroundcountRate, "+/-", source%rawbackgroundcountRateErr





        !***************************************************************************
        ! Extrapolate the background to the coincidence-loss radius
        !**************************************************************************
        source%corrBackGroundCountRate = source%rawBackGroundCountRate * &
                (coincidenceLossRadius / source%apertureRadius) ** 2 

        totRate = source%corrCountRate + source%corrBackGroundCountRate  
        
        if (totRate .gt. 0.0) then         

            source%corrCountRateErr = sqrt(totRate + source%corrBackgroundCountRate) * sqrt(source%exposure) / source%exposure


            source%corrBackgroundCountRateErr = sqrt(source%corrBackgroundCountRate * source%exposure) / source%exposure

      

!write(*, *) "omcoincorr line 153 x=", source%xpos, " y=", source%ypos, " significance=",   source%significance
        else

            source%corrCountRateErr = source%rawCountRateErr
            source%corrBackgroundCountRateErr = source%rawBackgroundCountRateErr

!write(*, *) "omcoincorr line 161 x=", source%xpos, " y=", source%ypos, " significance=",   source%significance


        end if


!write(*, *) "Line 118 corrbackgroundrate=", source%corrbackgroundcountRate, "+/-", source%corrbackgroundcountRateErr


        !***********************************************************************************
        ! NEED TO CHANGE THIS DEFINITION TO A RATIO FOR CONSISTENCY WITH OTHER CORRECTIONS
        !***********************************************************************************
        source%psf1Correction = (countsRatio - 1.0) * 100.0



!write(*, *) "aperture radius=", source%apertureRadius, " coincloss radius=",  coincidenceLossRadius, &
!        " counts ratio=", countsRatio 
    else
        source%corrCountRate = source%rawCountRate
        source%corrCountRateErr = source%rawCountRateErr
        source%corrBackGroundCountRate = source%rawBackGroundCountRate
        source%corrBackGroundCountRateerr = source%rawBackGroundCountRateerr

        source%psf1Correction = 0.0


    end if


    

!write(*, *) "PointsourceCoincLossRadius line 195 - coincidence-loss radius=",   coincidenceLossRadius
!write(*, *) "x=", source%xpos, " y=", source%ypos

!write(*, *) "Line 196 corrcountrate =", source%corrCountRate, " background=", &
!     source%corrBackGroundCountRate, " significance=", source%significance




!if (check) then
!write(*, *) "Line 229 before correction for coincidence-loss"
!write(*, *) "rawcountrate=", source%rawCountRate, "+/=", source%rawCountRateErr
!write(*, *) "corrected count rate=", source%corrCountRate
!write(*, *) "raw background count rate=", source%rawbackgroundCountRate
!write(*, *) "corrected background count rate=", source%corrbackgroundCountRate
!write(*, *) "corrected background count rate error=", source%corrbackgroundCountRateErr  
!end if

    !************************************************************
    ! The total count-rate (source+background)
    !************************************************************
    totalRate = source%corrCountRate + source%corrBackGroundCountRate


    !***************************************************************
    ! The error in the total rate
    !***************************************************************
    error  = sqrt(totalRate) * sqrt(source%exposure) / source%exposure 
               

    !****************************************************************
    ! Correct the total (source + background) for coincidence-loss
    ! NB if the source is "extended" then the total source counts
    ! will have been corrected for coincidence-loss correction by
    ! omdetect, but not the background
    !****************************************************************    
    corrRate = CorrectForCoincidenceLoss(totalRate, error, frameTime, &
                  corrRateErrMinus1, corrRateErrPlus1) 


    !******************************************************************
    ! Temporarily store correction factors in RA and Dec struct fields
    ! since no entries for them yet
    !******************************************************************

    if (totalRate .gt. 0.0) then
        source%RightAscension= CorrRate / totalRate 
        source%coincidenceLossCorr = corrRate / totalRate 
    else
        source%RightAscension = 0.0
        source%coincidenceLossCorr = 0.0 
    end if


    !**********************************************************************
    ! The error in the corrate
    !**********************************************************************
    corrRateErr = (abs(corrRateErrMinus1) + abs(corrRateErrPlus1)) / 2.0    


    !************************************************************
    ! Correct the background for coincidence-loss correction
    !************************************************************
    oldCorrrate = source%corrBackgroundCountRate
    source%corrBackgroundCountRate = CorrectForCoincidenceLoss(dble(source%corrBackGroundCountRate), &
                     error, frameTime, corrRateErrMinus2, corrRateErrPlus2) 


    if (oldcorrrate .gt. 0.0) then
        source%declination = source%corrBackgroundCountRate / oldCorrRate
    else
        source%declination = 0.0
    end if



    !**********************************************************
    ! The error in the background count-rate
    !**********************************************************
    source%corrBackgroundCountRateErr = (abs(corrRateErrMinus2) + abs(corrRateErrPlus2)) / 2.0    


    !*****************************************
    ! Subtract the corrected background
    !***************************************** 
    source%corrCountRate = corrRate - source%corrBackgroundCountRate   


    !******************************************************
    ! The error in the background-subtracted count rate
    !****************************************************** 
    source%corrCountRateErr = sqrt( corrRateErr * corrRateErr + &
            source%corrBackgroundCountRateErr * source%corrBackgroundCountRateErr)









!if (check) then
!write(*, *) "Line 149"

!write(*, *) "omcoinccorr_mod.f90 line 327 corrected total rate=", corrRate

!write(*, *) "corrected count rate=", source%corrCountRate

!write(*, *) "corrected background count rate=", source%corrbackgroundCountRate
!write(*, *) "total rate coincidence-loss correction=", source%rightascension, " background=", source%declination    


!write(*, *) "Omcoinccorr_mod.f90 line 255 significance=", source%significance

!write(*, 100)    source%xpos, source%ypos,   " raw countrate=", source%rawCountRate, "+/-",  &
!           source%rawCountRateErr, & 
!          " corrected source rate=", source%corrCountRate, "+/-", source%corrCountRateErr
!              source%rawBackgroundCountRate, source%corrBackgroundCountRate    
!100 format(2f10.2, a, f12.4, a, f12.2, a, f8.4, a, f8.2, 2f12.4)            

!end if
end subroutine PointSourceCoincLossCorr
!************************************************************
! Function PSFCountsRatio
! Computes the ratio of the total counts between
! the radii radius1 and radius2 
!************************************************************
real(double) Function PSFCountsRatio(radius1, radius2, x0, y0, countFrameRatio)
    real(single), intent(in) :: radius1    ! The aperture radius that the input counts refer to
    real(single), intent(in) :: radius2    ! The radius that the counts have to be scaled to 
    real(single), intent(in) :: x0         ! The x coordinate of the source 
    real(single), intent(in) :: y0         ! The y coordinate of the source 

    real(single), intent(in) :: countFrameRatio

    PSFCountsRatio = PSFFractionOfTotalCounts(x0, y0, radius1, countFrameRatio) / &
                     PSFFractionOfTotalCounts(x0, y0, radius2, countFrameRatio)


end function PSFCountsRatio  

!*********************************************************
! Function CorrectForCoincidenceLoss
! Corrects the raw count rate for coincidence losses  
!*********************************************************
real(double) function CorrectForCoincidenceLoss(rawCountRate, & 
                  countRateError, frameTime,  &
                  countRateErrorMinus, countRateErrorPlus)
    real(double), intent(in)     :: rawCountRate
    real(single), intent(in)     :: countRateError
    real(double), intent(out)    :: countRateErrorMinus
    real(double), intent(out)    :: countRateErrorPlus
 
    real(double), intent(in) :: frameTime
    real(single) :: rawTempRate
    type(ValueAsymmetricErrorT) :: rawRate
    type(ValueAsymmetricErrorT) :: corrRate

    if (rawCountRate .le. 0.0_double) then
        CorrectForCoincidenceLoss = rawCountRate
        countRateErrorMinus = countRateError
        countRateErrorPlus = countRateError
        return
    end if


    !*************************************************************
    ! NB frame time is in seconds
    !*************************************************************
    rawTempRate = min(0.99999_double / frameTime, dble(rawCountRate)) 
   
!   if (dble(rawCountRate) .gt. 0.99999_double / frameTime) then

        
!        CorrectForCoincidenceLoss = rawCountRate
!        countRateErrorMinus = countRateError
!        countRateErrorPlus = countRateError
!        return
!    end if

!write(*, *) "Correct for coincidence-loss raw counte rate=", rawCountRate, " countrateerror=", countrateerror
           
    rawRate = ValueAsymmetricErrorT(rawTempRate, dble(countRateError), dble(countRateError))




    !******************************************************************************************
    ! NB the final boolean parameter switches on the empirical correction
    !*****************************************************************************************


!write(*, *) "Without empirical correction"
!   corrRate = CAL_omPhotoNatural(rawRate, frameTime, 0.0_double, 0.0_double, .FALSE., .FALSE.)

!write(*, *) "Line 381 corrrate=", corrrate%value, " countrateerror minus=",  &
!corrrate%ErrorMinus, " plus=",    corrrate%Errorplus




!write(*, *) "With empirical correction"


    corrRate = CAL_omPhotoNatural(rawRate, frameTime, 0.0_double, 0.0_double, .FALSE., .TRUE.)



    
!write(*, *) "Correct for coincidence-loss corrected counte rate=", corrRate, " frametime=", frametime, &
!" rawtemprate=",   rawTempRate
!write(*, *) "framerate=", frametime / rawCountRate

    !****************************************************************
    ! If the count rate is greater than the frame rate then the above
    ! routine cannot do the correction- instead it sets the corrected 
    ! count rate to the largest number that can be stored on the computer. 
    !*****************************************************************

    if (corrRate%value .lt. 1e10) then
        CorrectForCoincidenceLoss = real(CorrRate%value)
        
        countRateErrorMinus = corrRate%errorMinus
     
        countRateErrorPlus  = corrRate%errorPlus


!write(*, *) "Line 381 countrateerror minus=",  countRateErrorMinus, " plus=",    countRateErrorplus



       
        
    else
        CorrectForCoincidenceLoss = rawCountRate


    end if
!write(*, *) "countrateerror minus=",  countRateErrorMinus, " plus=",    countRateErrorplus

end function CorrectForCoincidenceLoss

!******************************************
! PSFFractionOfTotalCounts
! Returns the fraction of the total counts
! contained within the given radius
!******************************************
real(double) function & 
    PSFFractionOfTotalCounts(x0, y0, radius, countFrameRatio)
    real(single), intent(in) :: x0
    real(single), intent(in) :: y0
    real(single), intent(in) :: radius
    real(single), intent(in) :: countFrameRatio
    type(OmPsfT) :: psf
 
    psf = CAL_omGetPSF(dble(x0), dble(y0), dble(countFrameRatio))

    PSFFractionOfTotalCounts = CAL_omPsfEncircledEnergy(psf, dble(radius))


end function PSFFractionOfTotalCounts 

end module omcoinccorr_mod


























