!**************************************************************
! module omphotometry_mod (10th July 2001)
!
! Module containing routines, etc,
! for doing aperture photometry on each source in the list. It has been added
! because the counts that initially computed for each source (at the same
! time that the centroided positions for each source are obtained) is not
! consistent with the CAL calibration of source counts to magnitudes.
! It also sets the source confusion flag and the source
! quality flag. 
!
!**************************************************************
module omPhotometry_mod
    
    use param
    use cal
    use dal

    use OmAllocation_mod      ! For memory allocation
    use omconstants_mod
    use OmPhotometryDefs_mod
    use OmNumerics_mod
    use OmSkyBackground_mod   ! To compute sky background
    use OmSourceparam_mod
    use OmCoinccorr_mod       ! For coincidence-loss corrections   
    use OmSmoothImage_mod
    use omellipse_mod    
interface GetExtendedSourceCounts
    module procedure GetExtendedSourceCounts1, &
                     GetExtendedSourceCounts2
end interface
interface GetBackgroundMode
    module procedure GetBackgroundMode1, &
                     GetBackgroundMode2
end interface
interface GetCounts
    module procedure GetCounts1, GetCounts2
end interface

interface GetBackgroundCounts
    module procedure GetBackgroundCounts1
end interface


interface AperturePhotometry
module procedure AperturePhotometry1, &
                 AperturePhotometry2, &
                 AperturePhotometry3

end interface AperturePhotometry
   
interface ComputeSourceRateError
    module procedure ComputeSourceRateError1

end interface ComputeSourceRateError

type(SourceParameters), pointer, private :: currentSource => null()       
real(single), private, parameter :: backgroundSigma = 2.0
 
    contains               
!****************************************************************************************
! subroutine AperturePhotometry
! Performs aperture photometry on a set of sources!
! using a circular aperture of a given pixel radius
! Parameters-
!     nx and ny - the dimensions of the images
!     image - the image
!     sourceQualityImage - describes the quality of each image pixel
!     photometryRadius - The aperture radius for doing the photometry
!     noSources - number of sources
!     xSource - array of source x-positions
!     ySource - array of source y-positions
!     psfXwidth - array holding the computed x extents of the sources (pixels).
!     psfYwidth - array holding the computed y extents of the sources (pixels).y
!     fSource - array to hold the counts for each source.
!               Contains the Jon Mittaz counts on entry
!               and where possible these will be replaced 
!               by the aperture photometry values.!
!     fBackGround - array to hold the background 
!               counts for each source. 
!               Contains the Jon Mittaz counts on entry
!               and where possible these will be replaced 
!               by the aperture photometry values.
!     sourceExtendedFlag - array containg either 0 (not extended source)
!                or 1 (extended source)
!     sourceConfusionFlag - array containing either 1 - no source within
!                           2, source within aperture but outside minimum aperture
!                           radius, or 3 - source within minimum apertuire radius.
!     sourceQualityFlag - Set to 1 if source region has a bad-quality pixel- 0 otherwise
!     sourceApertureRadius - the aperture radius that was used for     
!                      each source 
!****************************************************************************************

subroutine AperturePhotometry1(   &
    nx1,                     &  ! The x-extent of the images 
    ny1,                     &  ! The y-extent of the images 
    image,                   &  ! The image (including background) 
    backgroundImageIn,       &
    levelImage,              &  ! The source number image
    sourceQualityImage,      &  ! The pixel-quality image array 
    photometryRadius,        &  ! The aperture radius (pixels) to be used for doing photometry
    noSources,               &  ! The number of sources
    source)                     ! Array of type SourceParamters structures

    ! Input and unmodified variables
    integer(int16), intent(in) :: nx1, ny1                      
    real(single), pointer              :: image(:, :)   
    real(single), pointer    :: backGroundImageIn(:, :)   
    integer(int16), pointer           :: levelImage(:, :) 
    logical(bool), pointer :: sourceQualityImage(:, :)
    real(single), intent(in) :: photometryRadius       
    integer (int16), intent(in)                      :: noSources

    ! The following structure is defined in  "omdetectsource_mod.f90"
    ! It stores all the parameters associated with a source 
   
    type(SourceParameters), pointer :: source(:)       

    backGroundImage => backgroundImageIn 
 

    call AperturePhotometry2(    &
        nx1,                     &  ! The x-extent of the images 
        ny1,                     &  ! The y-extent of the images 
        image,                   &  ! The image (including background) 
        levelImage,              &  ! The source number image
        sourceQualityImage,      &  ! The pixel-quality image array 
        photometryRadius,        &  ! The aperture radius (pixels) to be used for doing photometry
        noSources,               &  ! The number of sources
        source)                     ! Array of type SourceParamters structures


end subroutine AperturePhotometry1
!****************************************************
!
!
!****************************************************
subroutine AperturePhotometry2(   &
    nx1,                     &  ! The x-extent of the images 
    ny1,                     &  ! The y-extent of the images 
    image,                   &  ! The image (including background) 
    levelImage,              &  ! The source number image
    sourceQualityImage,      &  ! The pixel-quality image array 
    photometryRadius,        &  ! The aperture radius (pixels) to be used for doing photometry
    noSources,               &  ! The number of sources
    source)                     ! Array of type SourceParamters structures

    ! Input and unmodified variables
    integer(int16), intent(in) :: nx1, ny1                      
    real(single), pointer              :: image(:, :)      
    integer(int16), pointer           :: levelImage(:, :) 
    logical(bool), pointer :: sourceQualityImage(:, :)
    real(single), intent(in) :: photometryRadius       
    integer (int16), intent(in)                      :: noSources

    ! The following structure is defined in  "omdetectsource_mod.f90"
    ! It stores all the parameters associated with a source 
   
    type(SourceParameters), pointer :: source(:)       
 
    ! Local variables

    integer(int32)           :: i, k, status, type    
    real(single) :: maxRadius         
    real(single) :: backGroundCountsError
    real(single) :: sigma    

    ! NB Set minSignificance to zero if no counts extrapolation
    ! to be used ! NB until the CAL calibration data is in place
    ! the full aperture will be used on all sources, unless
    ! another source lies within the aperture. If the source 
    ! significance is less than this value then
    ! the photometric counts will be extrapolated using
    ! the point-spread-function
    real(single), parameter :: minSignificance = 0.0  
    integer (int32) :: index

!integer, save :: count = 0       
 
    if (noSources .lt. 1) return ! Return if no sources


!    write(errstr,'(a)') "Performing aperture photometry on the sources"
!    call message(AppMsg, Verbose)
 

    ! Initialise arrays, etc
    apertureRadius = photometryRadius
    
    call Initialise(noSources, nx1, ny1)


    !******************************************************************
    !
    !
    !
    !******************************************************************

    CCDImage => image  ! Assign to input image

    qualityImage => sourceQualityImage ! The quality image (false=good pixels, true=bad pixels)
   
    
    sourceNumImage => levelImage ! The source number image

    noPointSources = maxval(levelImage)

    ! Set the minimum aperture radius to half the required aperture
    minimumApertureRadius = photometryRadius * 0.5
    fullApertureArea = PI * photometryRadius * photometryRadius  

   
 
    ! Obtain an index of the source positions in
    ! increasing x-pixel position. NB xSource is not
    ! sorted. sortIndex(i) will give the index of the 
    ! array element that gives the positions in increasing
    ! x-position. This is used for obtaining the distance
    ! from one source to its nearest neighbour. 
    call IndexSort(source%xPos, sortIndex, int(noSources, int32))  


    ! Compute the mode of the image, ignoring source regions

    backgroundCountsMode = ComputeBackgroundMode(image, &
                     sourceNumImage, nx, ny, backgroundSigma, sigma)




    backgroundCountsModeError = sigma


    ! sourceCoords will contain the position of the sources
    ! in increasing x-position- this array is used to obtain
    ! the maximum-sized aperture for doing photometry.
   
    do i = 1, noSources
        sourceCoords(i)%x = source(sortIndex(i))%xPos - 0.5   
        sourceCoords(i)%y = source(sortIndex(i))%yPos - 0.5  
        sourceCoords(i)%extension = source(sortIndex(i))%extensionFlag
       
        ! objectWidth will store the maximum of the x and y extents of the sources
        ! * 3, or if this is less than the aperture radius, to the aperture radius.                       
        objectWidth(i) = photometryRadius   

    end do
  
    ! Loop through all the sources doing aperture photometry on each one (if possible)

  
    k = 0


    do i = 1,  noSources ! Remember photometry on sources is conducted in increasing x-order
      
        index = sortIndex(i)
        sourceNumber = source(index)%number

       
        currentSource => source(index)         
        currentSource%number = sourceNumber
        if (currentSource%extensionFlag .eq. 0 ) then  

            ! Point-source photometry using circular aperture
            
            apertureRadius = photometryRadius    
            !********************************************
            ! Find the aperture radius that can be used
            ! for the photometry. If type is 1 then
            ! can use the required aperture, 2 - have
            ! to use a smaller one, 3 can't do photometry
            !********************************************



        



            if (currentSource%apertureRadius > coincLossRadius) then
                type = 5

            else if (currentSource%apertureRadius .gt. 0.0) then


                maxRadius = GetMaximumAperture(noSources, i, type)          



                
                maxRadius = currentSource%apertureRadius
                type = 1

               
            else
                !***********************************************************
                ! Basically the default in imaging chain
                !***********************************************************
                maxRadius = GetMaximumAperture(noSources, i, type)
                if (type .eq. 4 .or. btest(currentSource%qualityFlag, 4)) then 
                    maxRadius = coincLossRadius          
                end if 

            end if 
                         

            select case (type)    
                case (1, 2, 4)     ! 1 - No other objects within aperture, 
                                   ! 2 - Another object within aperture, but outside 
                                   ! minimum aperture radius- use aperture that 
                                   ! excludes this source
                                                     

                    maxradius = max(maxRadius, coincLossRadius / 2.0)
                    ! Point-source photometry



                    call PointSourcePhotometry1(currentSource, maxRadius)

  

                    if (type .eq. 4) then ! Source lies near an image edge

                  


                        currentSource%rawCountRate = currentSource%rawCountRate * fullApertureArea / apertureArea 
                        currentSource%rawCountRateErr = currentSource%rawCountRateErr * fullApertureArea / apertureArea 
 
                     
                        currentSource%rawBackGroundCountRate = currentSource%rawBackGroundCountRate &
                            * fullApertureArea / apertureArea 
                        backGroundCountsError = backGroundCountsError * fullApertureArea / apertureArea 
                        currentSource%rawbackgroundCountRateErr =  backGroundCountsError / currentSource%exposure
                
                        currentSource%apertureRadius = maxRadius



                    end if
                    
                    currentSource%ApertureRadius = maxRadius * real(binX)
                                
                    ! Correct counts for coincidence losses
                    ! NB The corrected count rates are stored in source
                    ! members correctedCountRate, correctedCountRateErr,
                    ! and correctedBackgroundRateErr   


                  
                    call PointSourceCoincLossCorr( currentSource, 12.0, frameTime )     




                    ! Compute the signal-to-noise (source counts / sqrt(background counts)   
   
!                    currentSource%Significance = & 
!                         ComputeSourceSignificance(currentSource%rawCountRate, &
!                             currentSource%rawBackGroundCountRate, currentSource%exposure)
                case (3)   

                    
                    


                    maxRadius = coinclossRadius / 2.0                    

                    call PointSourcePhotometry1(currentSource, maxRadius)
                    currentSource%ApertureRadius = maxRadius * real(binX)

                    call PointSourceCoincLossCorr(currentSource, 12.0, frameTime)     


!                    currentSource%Significance = & 
!                         ComputeSourceSignificance(currentSource%rawCountRate, &
!                             currentSource%rawBackGroundCountRate, currentSource%exposure)

                                      
                    type = 3
                     

                case (5)  ! Extended photometry on point-source (ie aperture bigger
                          ! than coincidence-loss aperture
                    call PointSourcePhotometry2(currentSource, currentSource%apertureRadius)

     
            end select
        else

            ! Extended source photometry
            ! The counts (source + background) are corrected
            ! for coincidence-losses  
   

            if (currentsource%apertureRadius .le. 0.0) then
                call GetExtendedSourceCounts(currentSource, coincLossRadius)
            else
                ! For elliptical apertures

                call GetExtendedSourceCounts(currentSource, coincLossRadius, &
                      currentSource%apertureRadius * currentSource%axialRatio, &   ! semi a
                      currentSource%apertureRadius, &                              ! semi b
                      currentSource%alpha)
            end if 

                       
            type = 0
        end if        

!if (count .lt. 20) then
!count = count + 1
!write(*, 100) index, source(index)%rawCountRate, source(index)%rawCountRateerr,  &
!  source(index)%corrCountRate, source(index)%corrCountRateerr 
 
!write(*, *) i, " )Line 403 index=", index, " source x=", source(index)%xpos, " y=", source(index)%ypos, " sig=", &
!source(index)%significance, " type=", type, " eflag=", source(index)%extensionFlag, &
!" ra=", source(index)%rightascension, " dec=",  source(index)%declination 

!end if

        ! Set the source confusion flag - 0 isolated source
        ! 1 Source within aperture radius and greater than minimum aperture radius
        ! 2 Source within minimum aperture radius         
        nullify(currentSource)        





!100 format(i8, 8e12.4, f12.2, i8, f8.1, i6)
    end do



!do i = 1, min(20, size(sortindex) )

!index = sortIndex(i) 
!write(*, *) index, " ) Line 403 index=", index, " source x=", source(index)%xpos, " y=", source(index)%ypos, " sig=", &
!source(index)%significance, " eflag=", source(index)%extensionFlag, &
!" ra=", source(index)%rightascension, " dec=",  source(index)%declination 

!end do





                   
    nullify(ccdImage, backgroundImage, qualityImage)     
 

    deallocate(sortIndex, stat = status)
    nullify(sortIndex)
    call CheckDeallocation(status, "sortIndex (Subroutine AperturePhotometry)")

    deallocate(sourceCoords, stat = status)
    nullify(sourceCoords)
    call CheckDeallocation(status, "sourceCoords (Subroutine AperturePhotometry)")

    deAllocate(backgroundPixel, stat = status)
    nullify(backgroundPixel)
  
    call CheckDeallocation(status, "backgroundPixel (Subroutine AperturePhotometry)")


    deAllocate(objectWidth, stat = status)
    call CheckDeallocation(status, "objectWidth (Subroutine AperturePhotometry)")
    nullify(objectWidth)
    if (associated(backgroundImage)) nullify(backgroundImage)

end subroutine AperturePhotometry2




!*********************************************
! Function GetApertureCounts
! Computes the counts within a given aperture
! Returns the value via the function name
!*********************************************
real(double) function GetApertureCounts(&
                    x0,  &  ! The x-pixel coordinate of the centre of the aperture
                    y0,  &  ! The y-pixel coordinate of the centre of the aperture
                radius)  ! The aperture radius (pixels) for summing the counts
               
    implicit none

    !*********************************************
    ! Input (unmodfied) variables
    !*********************************************
    real(single), intent(in) :: x0
    real(single), intent(in) :: y0
    real(single), intent(in) :: radius
    
    ! Local variables
    integer(int32) :: i, j
    integer(int32) :: i1, i2, j1, j2
   

    apertureRadius = radius        
           
    ! Determine upper and lower array indices for
    ! summing counts
   
    i1 = int(x0 - radius - 0001)
    i1 = max(il, i1)
  
    i2 = int(x0 + radius + 0.0001)
    i2 = min(iu, i2)  

    j1 = int(y0 - radius - 0.0001)
    j1 = max(jl, j1)

    j2 = int(y0 + radius + 0.0001)
    j2 = min(ju, j2)
 
    GetApertureCounts = 0.0_double 

    apertureArea = 0.0_double

    ! Sum the pixels in the strip round the centre

!write(*, *) "GetAperturecounts x0=", x0, " y0=", y0, " i1=", i1, " i2=", i2, " j1=", j1, " j2=", j2
    
    do i = i1, i2
        do j = j1, j2         
            GetApertureCounts = GetApertureCounts + PixelValue1(x0, y0, i, j)
            
        end do
    end do

  
end function GetApertureCounts       
!*********************************************
! Function GetApertureCounts1
! Computes the counts within a given aperture
! Returns the value via the function name
! This version for extended sources only
!*********************************************
real(double) function GetApertureCounts1(&
                    x0,  &  ! The x-pixel coordinate of the centre of the aperture
                    y0,  &  ! The y-pixel coordinate of the centre of the aperture
                radius  )  ! The aperture radius (pixels) for summing the counts
             
    implicit none

    !*********************************************
    ! Input (unmodfied) variables
    !*********************************************
    real(single), intent(in) :: x0
    real(single), intent(in) :: y0
    real(single), intent(in) :: radius
   
    !********************************************
    ! Local variables
    !*******************************************
    integer(int32) :: i, j
    integer(int32) :: i1, i2, j1, j2
   

    apertureRadius = radius        
        
    ! Determine upper and lower array indices for
    ! summing counts

    i1 = int(x0 - radius - 0001)
    i1 = max(il, i1)
  
    i2 = int(x0 + radius + 0.0001)
    i2 = min(iu, i2)  

    j1 = int(y0 - radius - 0.0001)
    j1 = max(jl, j1)

    j2 = int(y0 + radius + 0.0001)
    j2 = min(ju, j2)
 
    GetApertureCounts1 = 0.0_double 

    apertureArea = 0.0_double


    ! Sum the pixels in the strip round the centre
       
    do i = i1, i2
        do j = j1, j2         
            GetApertureCounts1 = GetApertureCounts1 + PixelValue2(i, j)
            
        end do
    end do

  
end function GetApertureCounts1       
!********************************************
! Function PixelValue1
! Purpose - returns the value of the pixel
! at the given position weighted by the area
! of the aperture contained within the pixel
!*********************************************
real(double) function PixelValue1(x0, y0, iIndex, jIndex)
    implicit none   
    real (single), intent(in) :: x0, y0
    real (single) :: r1, r2, r3, r4
    real (single) :: x, y, x1, x2, y1, y2
    real (single) :: minRadius, maxRadius
    real(double) :: area    
    integer (int32), intent(in) :: iIndex, jIndex


    x = real(iIndex)  
    y = real(jIndex) 
 
    x1 = x - x0
    y1 = y - y0
    x2 = x1 + 1.0
    y2 = y1 + 1.0

    r1 = sqrt(x1 * x1 + y1 * y1)
    minRadius = r1
    maxRadius = r1
 
    r2 = sqrt(x2 * x2 + y1 * y1)
    minRadius = min(minRadius, r2)
    maxRadius = max(maxRadius, r2)

    r3 = sqrt(x1 * x1 + y2 * y2)
    minRadius = min(minRadius, r3)
    maxRadius = max(maxRadius, r3)

    r4 = sqrt(x2 * x2 + y2 * y2)
    minRadius = min(minRadius, r4)
    maxRadius = max(maxRadius, r4)
  

    if (apertureRadius .le. minRadius) then
        PixelValue1 =  0.0_double
        return 
    end if

    ! Check that the pixel is okay to use
    if (associated(qualityImage)) then    
        if (qualityImage(iIndex, jIndex)) then ! Bad pixel- set source quality flag
            call setQualityFlag
      
        end if
    end if
    if (apertureRadius  .ge. maxRadius) then
        apertureArea = apertureArea + 1.0_double      
        PixelValue1 = CCDImage(iIndex, jIndex)
    else
        area = EffectivePixelArea(x1, y1)  ! Pixel area (1 if full pixel)
        pixelValue1 = area * CCDImage(iIndex, jIndex)
        apertureArea = apertureArea + area          
    end if
 
end function PixelValue1
!********************************************
! Function PixelValue2
! Purpose - For extended sources
!*********************************************
real(double) function PixelValue2(iIndex, jIndex)
       
    
    integer(int32), intent(in) :: iIndex, jIndex
  
  
    ! Check that the pixel is okay to use
    if (associated(qualityImage)) then 
        if (qualityImage(iIndex, jIndex)) then ! Bad pixel- set source quality flag
            call setQualityFlag

        end if
    end if
    
    apertureArea = apertureArea + 1.0_double      
    PixelValue2 = CCDImage(iIndex, jIndex)
   
 
end function PixelValue2
!**************************************************
! Function EffectivePixelArea
! Computes the area of a pixel contained
! within the aperture and returns the value
! via the name
! This will shortly be replaced by a better routine
!***************************************************
real(double) function EffectivePixelArea(x1, y1)
  
    real(single), intent(in) :: x1, y1
    real(single) :: radiusSqr
    real(single) :: xPos, yPos
    
    integer(int32) :: nIn
    integer(int32) :: i, j
     

    real(single), save :: deltaY(1 : 99)
    
    forall (j = 1 : 99) deltaY(j) = y1 + 0.01 * real(j)

    nIn = 0

    do i = 2, 100

        xPos = x1 + 0.01 * real(i - 1)
        do j = 1, 99     
            ypos = deltaY(j)        
            radiusSqr = xPos * xPos + yPos * yPos 
            if (sqrt(radiusSqr) .le. apertureRadius) then                         
                nIn = nIn + 1
            end if
        end do
    end do

    EffectivePixelArea = dble(nin) / 9801_double 

end function EffectivePixelArea


!***********************************************
! ComputeSourceRateError
! Computes the source counts error and returns 
! the value via the name
! NB sourceCounts is the background-subtracted counts
! and backgroundcounts the background counts
!***********************************************
real(double) function ComputeSourceRateError1(sourceCounts, backGroundCounts, exposure)
    real(double), intent(in) :: sourceCounts       ! The background-subtracted source counts
    real(double), intent(in) :: backGroundCounts   ! The background counts
    real(single), intent(in) :: exposure
  

    ComputeSourceRateError1 = sqrt(sourceCounts + 2.0_double * backGroundCounts) / dble(exposure)
 
end function ComputeSourceRateError1




!******************************************************************
! Function GetMaximumAperture
! Finds the maximum aperture size that can be used for photomery
! Returns the value via the function name
! Also return "type" -
! type - 1 = no object within aperture radius pixels
!        2 = object within aperture radius, but outside minimum aperture radius
!        3 = object within minimum aperture radius
!*******************************************************************
real(single) function GetMaximumAperture(noSources, index, type)
    integer(int16), intent(in) :: noSources        ! The number of sources    
    integer(int32), intent(in) :: index            ! Index of source 
    integer(int32), intent(out) :: type            ! Determines what kind of photometry can be performed
    TYPE(OMPoint) :: centre                   ! Coords of source
    real(single) :: minimumSeparation  ! Minimum distance between source of interest and other sources
    real(single) :: borderDistance     ! Distance to the edge of the image from the source 
    real(single) :: okayDistance       ! Minimum allowed separation between source and nearest neighbour   
    real(single) :: distance           ! Distance bewteen source and neighbour
    real (single):: d, d1
    integer(int32) :: k    
   
    centre = sourceCoords(index)

    okayDistance = 2.0 * apertureRadius  ! Okay for aperture photometry if nearest star is 2 aperture radii distant
    borderDistance = 1e8
    minimumSeparation = 1e6

    type = 0

    !******************************************
    ! Checks to see if object is near the
    ! edge of the image
    !******************************************
    if (centre%x - xLeft .lt. apertureRadius) then
        borderDistance = centre%x - xLeft
        type = 4
        currentSource%confusionFlag = ibset(currentSource%confusionFlag, 2)
 

    else if  (xRight - centre%x .lt. apertureRadius) then
        type = 4
        borderDistance = xRight - centre%x
        currentSource%confusionFlag = ibset(currentSource%confusionFlag, 2)
 
    end if

    if (centre%y - yBottom .lt. apertureRadius) then
        type = 4
        borderDistance = min(centre%y - yBottom, borderDistance)
        currentSource%confusionFlag = ibset(currentSource%confusionFlag, 2)
 
    else  if (yTop - centre%y .lt. apertureRadius) then
        type = 4
        borderDistance = min(yTop - centre%y, borderDistance)
        currentSource%confusionFlag = ibset(currentSource%confusionFlag, 2)
 
    end if


    !***************************************************
    ! Find minimum distance between current source 
    ! and sources to its left. NB sourceCoords contains
    ! the x-coordinates of the sources in increasing 
    ! order.
    !***************************************************
    if (index .gt. 1) then
       
        k = index
    
        do while (k .gt. 1)
            k = k - 1
            if (sourceCoords(k)%extension .eq. 1) cycle
                
                           
            d = centre%x - sourceCoords(k)%x
            d = max(0.0, d)
            if (d .gt. apertureRadius + objectWidth(k) ) exit
            if (sourceCoords(k)%y .ge. centre%y) then
                d1 = sourceCoords(k)%y - centre%y
                d1 = max(0.0, d1)
                
            else
                d1 = centre%y - sourceCoords(k)%y
                d1 = max(0.0, d1)
            end if


            distance = sqrt(d * d + d1 * d1)
            
            if (distance + apertureRadius - objectWidth(k) .lt. minimumSeparation) then 

                minimumSeparation = distance + apertureRadius - objectWidth(k)


            end if
  
        end do
    end if

    !***************************************************
    ! Find minimum distance between current source
    ! and sources to its right.
    !**************************************************
    if (index .lt. noSources) then
        k = index
        do while (k .lt. noSources)
            k = k + 1
            if (sourceCoords(k)%extension .eq. 1) cycle
        
            d = sourceCoords(k)%x - centre%x
            d = max(0.0, d)

            if (d .gt. apertureRadius + objectWidth(k)) exit

            if (sourceCoords(k)%y .ge. centre%y) then
                d1 = sourceCoords(k)%y - centre%y
                d1 = max(0.0, d1)
            else
                d1 = centre%y - sourceCoords(k)%y
                d1 = max(0.0, d1)
            end if

            distance = sqrt(d * d + d1 * d1)
            if (distance + apertureRadius - objectWidth(k) .lt. minimumSeparation) then
                minimumSeparation = distance + apertureRadius - objectWidth(k)


            end if
            
        end do
    end if

!if (type .eq. 4) then
!write(*, *) "Line 801 type=4, minimum separation=", minimumSeparation, " border distance=", borderDistance
!end if

    if (type .eq. 4) then
      
        if (minimumSeparation .ge. borderDistance) then
            GetMaximumAperture = borderDistance
            return
         end if
    end if    

   

    if (minimumSeparation .ge. okayDistance) then
    
        type = 1
        currentSource%confusionFlag = 0
 
        GetMaximumAperture = apertureRadius

    else if (minimumSeparation .ge. 2.0 * minimumApertureRadius) then  
        type = 2 
        GetMaximumAperture = 0.5 * minimumSeparation   ! - apertureRadius
        currentSource%confusionFlag = ibset(currentSource%confusionFlag, 0 )
 
    else
        type = 3    ! Can't do photometry on
        GetMaximumAperture = 0.0
        currentSource%confusionFlag = ibset(currentSource%confusionFlag, 1 )
 
    end if


end function GetMaximumAperture


!*********************************************
! subroutineGetExtendedSourceCounts
! Computes the counts for an extended source
! Returns the value via the function name
!*********************************************
subroutine GetExtendedSourceCounts1(source, radius)

    ! Input (unmodfied) variables
    real(single), intent(in) :: radius 
    type(SourceParameters), intent(inout) :: source

    ! Local variables
    integer(int32) :: i, j, count
  
    real(single) :: x0, y0
      
    real(double) :: rawCounts, correctedCounts, countRate
    real(double) :: f
    real(double) :: backgroundRate, totalRate
    real(double) :: backgroundCounts, correctedBackgroundCounts 
    real(double) :: backgroundCountsPerPixel, backgroundRatePerPixel
    real(double) :: correction
    logical :: useBackgroundMap = .false.
    real(double) :: backgroundMode 
!    integer, save :: count1 = 0    

    ! Determine the corrected background count-rate per pixel

!if (count1 .lt. 20) then
!write(*, *) "getextendedsourcecounts1"
!end if

    if (source%rawBackgroundCountRate > 0.0) then       ! Use the user-specified value

        backgroundMode = source%rawBackgroundCountRate  ! Use user background value
        backgroundRate = backgroundMode * PI * 144.0 / (dble(binX * binX) * source%exposure)

        ! The coincidence-loss correction for each background pixel
        correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), frameTime)
        backgroundCountsPerPixel = backgroundCountsMode !* correction
        backgroundRatePerPixel = backgroundCountsPerPixel / source%exposure
              

    else if (associated(backGroundImage)) then          ! Using a background image
        useBackgroundMap = .true.
        backgroundCounts = 0.0_double
        correctedBackgroundCounts = 0.0_double

!if (count1 .lt. 20) then
!       write(*, *) "********* Using background image ************"
!count1 = count1 + 1
!end if
    else
        backgroundMode = backgroundCountsMode           ! Use global image mode                 
        backgroundRate = backgroundMode * PI * 144.0 / (dble(binX * binX) * source%exposure)

        ! The coincidence-loss correction for each background pixel
        correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), frameTime)
        backgroundCountsPerPixel = backgroundCountsMode * correction
        backgroundRatePerPixel = backgroundCountsPerPixel / source%exposure
      
    end if

  
    count = 0
    rawCounts = 0.0_double
    correctedCounts = 0.0_double


!if (count1 .lt. 20) then
!write(*, *) "radius=", radius, " source%apertureRadius=", source%apertureRadius
!end if  

    do i = source%iMin, source%iMax
        x0 = real(i) + 0.5
     
        do j = source%jMin, source%jMax
            
            if (sourceNumImage(i, j) .eq. source%number) then
               
                y0 = real(j) + 0.5

                ! Compute the coincidence-loss correction for this pixel
                countRate= GetApertureCounts1(x0, y0, radius) / source%exposure                   
                f =  PixelValue2(i, j)            
                rawCounts = rawCounts + f
                correctedCounts = correctedCounts + f * &
                          CoincidenceLossCorrection(countRate, sqrt(real(countRate)), frameTime)

!                if (count .lt. 20) then
!                    write(*, *) "Pixel(", i, ",", j, ") value=", ccdimage(i, j), &
!                                      " aperture count-rate=", countRate, " cl factor=", &
!                                        CoincidenceLossCorrection(countRate, sqrt(real(countRate)), frameTime), &
!             " corrected counts=", correctedCounts
!                end if



               
                count = count + 1   

                if (useBackgroundMap) &       ! Use the background map           
                    backgroundCounts = backgroundCounts + backgroundImage(i, j)
      
                
                  
            end if
        end do
    end do
    if (useBackgroundMap) then            
         f = backgroundCounts / dble(count) 
         backgroundRate = f * PI * radius * radius / source%exposure

!if (count1 .lt. 20) then 
!         write(*, *) "Background counts=", backgroundCounts, " count=", count
!         write(*, *) "background rate=", backgroundRate    
!end if
       correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), frameTime)

!if (count1 .lt. 20) then 
!write(*, *) "background correction factor=", correction
!end if

        backgroundCountsPerPixel = f * correction
        backgroundRatePerPixel = backgroundCountsPerPixel / source%exposure
   

   

        rawCounts = rawCounts - backgroundCounts


       




        correctedCounts = correctedCounts - backgroundCounts * correction

!if (count1 .lt. 20) then 
!write(*, *) "corrected counts=", correctedCounts
!end if


    else
        rawCounts = rawCounts - dble(count) *  backgroundCountsMode
        correctedCounts = correctedCounts - dble(count) * backgroundCountsPerPixel
        backgroundCounts = backgroundCountsMode * dble(count)
     
 

    end if
    source%rawBackGroundCountRate = backgroundCounts / source%exposure 
    source%corrBackGroundCountRate = backgroundRatePerPixel * dble(count)




    ! The raw count rates
    source%rawCountRate = rawCounts / source%exposure
    source%rawCountRateErr = ComputeSourceRateError(rawCounts, backGroundCounts, source%exposure)
           






!if (count1 .lt. 20) then
!write(*, *) "raw counts=", rawCounts, " exposure=", source%exposure, " num pixels=", count
!write(*, *) "raw count rate=", source%rawCountRate, "+/-",  source%rawCountRateErr
!end if         
    source%corrCountRate = correctedCounts / source%exposure
    source%corrCountRateErr =  ComputeSourceRateError(correctedCounts, backGroundCounts, source%exposure)
                   


    ! The source coincidence-loss correction
    totalRate = source%rawCountRate + source%rawBackgroundCountRate
    if (totalrate .gt. 0.0) then
        source%rightAscension =  (source%corrCountRate + source%corrBackgroundCountRate) / &
           totalrate
    else
        source%rightAscension = 1.0
    end if
    ! The background coincidence-loss correction
    if (source%rawBackgroundCountRate .gt. 0.0) then
        source%declination =  source%corrBackGroundCountRate / source%rawBackGroundCountRate
    else  
        source%declination = 1.0

    end if


!if (count .lt. 20) then
!write(*, *) "corr count rate=", source%corrCountRate, "+/-",  source%corrCountRateErr
!write(*, *) "corr background count rate=", source%corrbackgroundCountRate, "+/-",  source%corrbackgroundCountRateErr
!end if
    !*****************************************
    ! Recalculate the source significance
    !*****************************************   
    source%Significance = ComputeSourceSignificance(real(rawCounts), real(backgroundCounts))


   
    source%apertureRadius = nullValue  
    source%psf1Correction = 0.0
    

    source%coincidenceLossCorr = max(0.0, (source%corrCountRate - source%rawCountRate) * 100.0 / source%rawCountRate) 

 
!if (count1 .lt. 20) then
!    write(*, *) "Extended source number ", sourceNumber, " x=", source%xpos, " y=", source%ypos 
!    write(*, *) " Original count rate=", source%rawCountRate, " new rate=" , source%corrCountRate
!    write(*, *) " Original background rate=", source%rawBackgroundCountRate, &
!                " new rate=" , source%corrBackgroundCountRate, " radius=",  source%apertureRadius

!write(*, *) "source+background coinc  correction=", source%rightascension

!write(*, *) "background coincidence  correction=", source%declination

!write(*, *) "Significance=",  source%Significance
!end if
end subroutine GetExtendedSourceCounts1       
!*********************************************
! subroutine GetExtendedSourceCounts2
! For extended sources with elliptical apertures
! Computes the count-rates/etc for an extended source
! 
!*********************************************
subroutine GetExtendedSourceCounts2(source, coincLossRadius, semiA, semiB, angle)

    ! Input (unmodfied) variables
    real(single), intent(in) :: coincLossRadius, semiA, semiB, angle
    type(SourceParameters), intent(inout) :: source

    ! Local variables
    integer(int32) :: i, j
    real(single) :: x0, y0, theta
      
    real(double) :: f, rawCounts, correctedCounts, countRate
    real(double) :: backgroundRate
    real(double) :: backgroundCounts = 0.0_double, correctedBackgroundCounts
    real(double) :: correction, pixArea, totalArea
    real(single) :: error    
    logical :: useBackgroundMap = .false.
    logical :: useImage = .false.
!integer, save :: count = 0
 
    ! Determine the corrected background count-rate per pixel using 
    if (angle .lt. 0.0) then
        theta = angle + 180.0
    else
        theta = angle
    end if

!write(*, *) "GetextendedSourceCounts2 line 993, exposure=", source%exposure
    ! the elliptical background annulus


    if (source%rawBackgroundCountRate > 0.0) then ! Using a fixed background mode to get background counts
        backgroundCounts = source%rawBackgroundCountRate * PI * semiA * semiB
      
    else if (associated(backGroundImage)) then ! Using a background image
        useBackgroundMap = .true.

    else ! Using the mode from the image
!if (count .lt. 5) then
!write(*, *) "calling getbackgroundmode2"
!end if
        
        backgroundCounts = GetBackGroundMode2( &
                     source%xPos,          &   ! The centre x-pixel coordinate of the aperture
                     source%yPos,          &   ! The centre y-pixel corrdinate of the aperture 
                     source%innerBckRadius, &   ! The aperture inner radius
                     source%outerBckRadius, &   ! The aperture outer radius
                     source%axialRatio,  &   ! The ellipse axial ratio
                     theta,       &   ! The ellipse position angle
                     error)      ! The error
        useImage = .true.

    end if    


! if (count .lt. 5) then
!if (usebackgroundmap) write(*, *) "using background map"
!write(*, *) "Line 1130 backgroundcounts=",   backgroundCounts, " xpos=", source%xpos, " ypos=", source%ypos, &
! " inner bck=", source%innerBckRadius, " outer=",   source%outerBckRadius, " axial ratio=", source%axialratio 
!end if           







  
    rawCounts = 0.0_double
    correctedCounts = 0.0_double
    totalArea = 0.0_double
  
    totalArea = 0.0_double
    do i = max(0, int(floor(source%xPos) - ceiling(semiA)) - 1 ),  &
                min(nx - 1, int(ceiling(source%xPos) + semiA + 0.1) + 1) 
        x0 = real(i) + 0.5
     
        do j = max(0, int(floor(source%yPos) - ceiling(semiA) ) - 1),  &
                     min(ny - 1, int(ceiling(source%yPos) + semiA + 0.1) + 1) 
       
                            
            y0 = real(j) + 0.5 ! 
            pixArea = PixelArea(x0 + 0.5, y0 + 0.5, source%xPos, source%yPos, semiA, semiB, theta)

            if (pixArea <= 0.0) cycle
            totalArea = totalArea + pixArea
            ! Compute the coincidence-loss correction for this pixel
            countRate= GetApertureCounts1(x0, y0, coincLossRadius) / source%exposure 
            if (pixArea .lt. 1.0) then                  
                f =  PixelValue2(i, j) * pixArea            

                if (useBackgroundMap) &                  
                    backgroundCounts = backgroundCounts + backgroundImage(i, j) * pixArea
         
            else
                f =  PixelValue2(i, j)            

                if (useBackgroundMap) &                  
                    backgroundCounts = backgroundCounts + backgroundImage(i, j)
         
            end if

            rawCounts = rawCounts + f
            correctedCounts = correctedCounts + f * &
                        CoincidenceLossCorrection(countRate, sqrt(real(countRate)), frameTime)
                                 
        end do
    end do
!write(*, *) "Line 1191 background counts=", backgroundcounts

    if (useImage) backgroundCounts = backgroundCounts * totalArea

!write(*, *) "Line 1195 background counts=", backgroundcounts












!if (count .lt. 5) then
!count = count + 1
!write(*, *) "rawCounts=", rawCounts, " background counts=", backgroundCounts, " total area=", totalArea, &
!        " semia=", semia, " semib=", semib, " theoretical area=", (PI * semia * semib)
!if (associated(backgroundimage)) write(*, *) "using a background image"
!end if

    backgroundRate = backgroundCounts / source%exposure ! Background count rate
    source%rawBackgroundCountRate = backgroundCounts / source%exposure
      





    !**********************************************************
    ! The coincidence-loss correction for the background
    !*********************************************************
    correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), frameTime)
    source%corrBackGroundCountRate = backgroundRate * correction ! Corrected background rate



!write(*, *) "Line 1229 correction=", correction

    correctedBackgroundCounts = backgroundCounts * correction

   
   

    ! Background-subtracted counts
    rawCounts = rawCounts - backgroundCounts
    source%rawCountRate = rawCounts / source%exposure

    !*****************************************
    ! Recalculate the source significance
    !*****************************************   
    source%Significance = ComputeSourceSignificance(real(rawCounts), real(backgroundCounts))

 


    source%rawCountRateErr = ComputeSourceRateError(rawCounts, backGroundCounts, source%exposure)
               
         
!write(*, *) "background subtracted rawCount rate=", source%rawCountRate, " background rate=", source%rawBackgroundCountRate

    ! Corrected background-subtracted counts
    source%corrCountRate = (correctedCounts - correctedBackgroundCounts) / source%exposure
    source%corrCountRateErr =  ComputeSourceRateError(correctedCounts, correctedBackGroundCounts, source%exposure) 


!write(*, *) "GetExtendedsourcecounts2 line 1254 corrected counts=", correctedcounts, &
!" correctedbackgroundcounts=", correctedbackgroundcounts


           
   
!    source%apertureRadius = nullValue  
    source%psf1Correction = 0.0



    ! The source coincidence-loss correction
    if (source%rawCountRate .gt. 0.0) then
        source%rightAscension =  source%corrCountRate / source%rawCountRate
    end if
    ! The background coincidence-loss correction
    if (source%rawBackgroundCountRate .gt. 0.0) then
        source%declination =  source%corrBackGroundCountRate / source%rawBackGroundCountRate
    end if



    if (source%rawCountRate .gt. 0.0) then
        source%coincidenceLossCorr = source%corrCountRate / source%rawCountRate 
    else
       source%coincidenceLossCorr = 1.0
    end if
  



!    write(*, *) "Extended source number ", sourceNumber, " x=", source%xsource, " y=", source%ysource 
!    write(*, *) " Original count rate=", source%rawCountRate, " new rate=" , source%correctedCountRate
!    write(*, *) " Original background rate=", source%rawBackgroundCountRate, &
!                " new rate=" , source%correctedBackgroundCountRate, " radius=",  source%apertureRadius
!    write(*, *) "Significance=",  source%Significance

end subroutine GetExtendedSourceCounts2      
!*********************************************
! Function GetBackGroundMode
! For use with circular annulii
! Computes the background counts within the given
! annulus, using the median counts/pixels obtained
! from an annulus region. 
! Return the value via the function name, and
! also the error (standard deviation) 
!*********************************************
real (double) function GetBackGroundMode1( &
    x0,          &   ! The centre x-pixel coordinate of the aperture
    y0,          &   ! The centre y-pixel corrdinate of the aperture 
    innerRadius1, &   ! The aperture inner radius
    outerRadius1, &   ! The aperture outer radius

    error)      ! The error

    implicit none
    
    ! Input variables  
    real(single), intent(in) :: x0
    real(single), intent(in) :: y0
    real(single), intent(in) :: innerRadius1
    real(single), intent(in) :: outerRadius1
    real(single), intent(out) :: error
  
    ! Local variables
    integer(int32) :: i, j
    integer(int32)  :: i0, j0, i1, i2, j1, j2
    integer(int32) :: ndata  ! The number of background pixels used
    real(single) :: innerRadius, outerRadius, x, y, dx, dy, r, radius
    integer :: iter
    integer, parameter :: minData = 99 ! was 299
    ! Scan the annulus region and add pixel values to the 
    ! backGroundPixel array

    innerRadius = innerRadius1
    outerRadius = outerRadius1

    i0 = int(x0)
    j0 = int(y0)
    nData = 0  

  
!write(*, *) "Line 1270 Get background mode 1 inner radius=", innerRadius, " outer=", outerRadius    




    r = ceiling(outerRadius)

    i1 = i0 - r - 1
    i1 = max(il, i1)
  
    j1 = j0 - r - 1
    j1 = max(jl, j1)



    do iter = 0, 2        
     
       
        outerRadius = outerRadius1 + real(iter * 2)
   
!write(*, *) "Line 1282 iteration ", iter, " innerradiusr=", innerradius, " outerradius=", outerradius, &
!      " ndata=", ndata

        ! Determine upper and lower array indices for
        ! summing counts
        r = ceiling(outerRadius)

        
        i2 = i0 + r + 1
        i2 = min(iu, i2)  

  
        j2 = j0 + r + 1
        j2 = min(ju, j2)
 
    
iLoop : do i = i1, i2

            x = real(i) + 0.5  
            dx = x - x0
            do j = j1, j2          
               
                if (associated(qualityImage)) then
                    if (qualityImage(i, j)) cycle
                end if

                if (associated(sourceNumImage)) then

                    if (sourceNumImage(i, j) .gt. 0) cycle

                end if
           
                y = real(j) + 0.5
                dy = y - y0
                radius = sqrt(dx * dx + dy * dy)
                if (radius .ge. innerRadius .and. radius .le. outerRadius) then         
              
                    nData = nData + 1 
                    backGroundPixel(nData) = CCDImage(i, j)  
                    if (nData == maxNumPixels) then

!                        write(*, *) "Line 940 omphotometry nData=", nData, &
!                          " innerRadius=", innerRadius, " outer=", outerRadius
                        exit iLoop
                    end if  
                end if
            
                              
            end do
        end do iLoop
!write(*, *) "Line 1337 Get background mode 1 inner radius=", innerRadius, " outer=", outerRadius    


        if (nData .gt. minData) exit 
    
        innerRadius = outerRadius
    end do

!write(*, *) "Line 1334 Get background mode 1 inner radius=", innerRadius, " outer=", outerRadius    



    ! Compute the median background pixel value
!write(*, *) "GetBackgroundmode1 - ndata=", ndata, " mindata=", minData    
    if (nData .gt. minData) then  ! w

        GetBackgroundMode1 = BackgroundMode(nData, backGroundPixel, backgroundSigma, error)

!  GetBackgroundMode1 = BackgroundMode2(nData, backGroundPixel, error)



!write(*, *) "Line 1354 mode=", getBackgroundmode1, " ndata=", ndata
    else if (associated(backgroundImage)) then  ! Use the background image

        nData = 0
iLoop1 :    do i = i1, i2

            x = real(i)  
            dx = x - x0
            do j = j1, j2          
              
            
                y = real(j)
                dy = y - y0
                r = sqrt(dx * dx + dy * dy)
                if (r .ge. innerRadius1 .and. r .le. outerRadius1) then         
              
                    nData = nData + 1 
                    backGroundPixel(nData) = backgroundImage(i, j)  
                    if (nData == maxNumPixels) then

                        exit iLoop1
                    end if  
                end if                                        
            end do
        end do iLoop1


        GetBackgroundMode1 = BackgroundMode(nData, backGroundPixel, backgroundSigma, error)
  
!   GetBackgroundMode1 = BackgroundMode2(nData, backGroundPixel, error)

  








!        write(*, *) "Line 1361 Using background image- getbackgroundmode1=", getbackgroundmode1, " ndata=", ndata
!        write(*,  *) "x0=", x0, " y0=", y0, " i1=", i1, " i2=",  i2, " j1=", j1, " j2=", j2
!       write(*, *) "Inner radius=", innerRadius, " outer=", outerRadius    
    else
        GetBackgroundMode1 = backgroundCountsMode
        error = backgroundCountsModeError
!        write(*, *) "Line 1226 using global background  backgroundcountsmode=", backgroundCountsMode 
    end if
!write(*, *) "Omphotometry GetBackgroundMode1 nData=", ndata, " mode=",  &
!     GetBackgroundMode1, " X=", CurrentSource%XPOS, " y=",  CurrentSource%yPos  

!write(*, *) "GetBackgroundMode1 line 1385 innerRadius=", innerRadius, " outerradius=", outerRadius, " outerradius1=", outerradius1
 
end function GetBackgroundMode1       


!*********************************************
! Function GetBackGroundMode2
! For use with elliptical annulii
! Computes the background counts within the given
! annulus, using the median counts/pixels obtained
! from an annulus region. 
! Return the value via the function name, and
! also the error (standard deviation) 
!*********************************************
real (double) function GetBackGroundMode2( &
    x0,          &   ! The centre x-pixel coordinate of the aperture
    y0,          &   ! The centre y-pixel corrdinate of the aperture 
    innerRadius, &   ! The aperture inner radius
    outerRadius, &   ! The aperture outer radius
    axialRatio,  &   ! The ellipse axial ratio
    theta1,       &   ! The ellipse position angle
    error)      ! The error

    implicit none
    
    ! Input variables  
    real(single), intent(in) :: x0
    real(single), intent(in) :: y0
    real(single), intent(in) :: innerRadius
    real(single), intent(in) :: outerRadius
    real(single), intent(in) :: axialRatio
    real(single), intent(in) :: theta1
    real(single), intent(out) :: error
  
    real(single) r, semiA1, semiA2, semiB1, semiB2, theta 

  
    ! Local variables
    integer(int32) :: i, j
    integer(int32)  :: i0, j0, i1, i2, j1, j2
    integer(int32) :: ndata  ! The number of background pixels used
    real(single) :: x, y

!integer, save :: count = 0

    semiA1 = innerRadius * axialRatio
    semiA2 = outerRadius * axialRatio
    semiB1 = innerRadius
    semiB2 = outerRadius

    theta = theta1 * degreesToRadians
    ! Scan the annulus region and add pixel values to the 
    ! backGroundPixel array

    i0 = int(x0)
    j0 = int(y0)
    nData = 0  
        
    i0 = int(x0)
    j0 = int(y0)
       
    r = ceiling(outerRadius * axialRatio)
   
    ! Determine upper and lower array indices for
    ! summing counts

    i1 = i0 - r - 1
    i1 = max(il, i1)
  
    i2 = i0 + r + 1
    i2 = min(iu, i2)  

    j1 = j0 - r - 1
    j1 = max(jl, j1)

    j2 = j0 + r + 1
    j2 = min(ju, j2)
 
    
iLoop :   do i = i1, i2

        x = real(i)  
   
        do j = j1, j2          
                    
            if (associated(qualityImage)) then
                if (qualityImage(i, j)) cycle
            end if

            if (associated(sourceNumImage)) then
                if (sourceNumImage(i, j) .gt. 0) cycle
            end if

           
            y = real(j)
              
            !********************************************************************
            ! Point has to be within the inner and outer ellipse
            !********************************************************************
            if (.not. PointInEllipse(x, y, x0, y0, semiA1, semiB1, theta) .and. &
                      PointInEllipse(x, y, x0, y0, semiA2, semiB2, theta)) then
          
              
                nData = nData + 1 
                backGroundPixel(nData) = CCDImage(i, j)  
                if (nData == maxNumPixels) then

!                  write(*, *) "Line 940 omphotometry nData=", nData, &
!                      " innerRadius=", innerRadius, " outer=", outerRadius
                  exit iLoop
                end if  
            end if
                    
            
        end do
    end do iLoop

!if (count .lt. 5) then
!count = count + 1
!write(*, *) "getbackgroundmode2  semia1=", semia1, " semia2=", semia2, " semib1=", semib1, " semib2=", semib2, &
!" maxnumpixels=", maxnumpixels
!end if
    ! Compute the median background pixel value
    
    if (nData .gt. 100) then   

        GetBackgroundMode2 = BackgroundMode(nData, backGroundPixel, backgroundSigma, error)


!if (count .lt. 5) then
!write(*, *) "ndata=", ndata, " mode=",   GetBackgroundMode2 
!end if




    else





        GetBackgroundMode2 = backgroundCountsMode
        error = backgroundCountsModeError



!if (count .lt. 5) then
!write(*, *) "Line 1596 ndata=", ndata, " mode=",   GetBackgroundMode2 
!end if


        
    end if

 
end function GetBackgroundMode2       


!*******************************************************
! Subroutine Initialise
!******************************************************

subroutine Initialise(noSources, nx1, ny1)
    integer(int16), intent(in) :: noSources, nx1, ny1

 
    real(single) :: calRadius
    integer(int32) :: status
    
    nx = nx1
    ny = ny1
  
    coincLossRadius = real(12 / binX) 
    ! Set up the background annulus
    
    calRadius =  CAL_omGetApertureRadius() / real(binx) 
!   write(*, *) "cal radius=" , calradius

    apertureRadius = real(12 / binx)
    innerRadius = calRadius + 2.0
    
    outerRadius = sqrt(3.0 * apertureRadius * apertureRadius + innerRadius * innerRadius)

!write(*, *) "Line 1553 apertureRadius=", apertureRadius, " innerRadius=", innerRadius, " outerRadius=", outerRadius, &
!     " calradius=", calRadius
    call InitialiseArrays
 

    sortIndex => AllocateInt32(1, int(noSources, int32), "sortIndex (AperturePhotometry)")
     
    allocate(sourceCoords(1 : int(noSources, int32)), stat = status)
    call CheckAllocation(status, "sourceCoords (AperturePhotometry)")
 
    objectWidth => AllocateSingle(1, int(noSources, int32), "objectWidth (AperturePhotometry)")

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
      
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1     

end subroutine Initialise
!********************************************************
!
!*******************************************************
subroutine InitialiseArrays
  

    maxNumPixels = 30000
    if (associated(backgroundPixel)) return
  
    
    backgroundPixel => AllocateSingle(1, maxNumPixels, "backgroundPixel (AperturePhotometry)")


 
end subroutine InitialiseArrays
!********************************************************
!
!*******************************************************
subroutine deleteArrays

    if (associated(backgroundPixel)) then
        deallocate(backgroundPixel)
        nullify(backgroundPixel)
    end if


end subroutine DeleteArrays


!*************************************************************
! Point-source photometry
!*************************************************************
subroutine PointSourcePhotometry1(source, radius) 
    type(SourceParameters), intent(inout) :: source
    real(single), intent(in) ::radius
    real(double) :: sourceCounts, backgroundCounts
    real(single) :: backgroundCountsError
    
    real(double) :: backGroundCountsPerPixel
    real(single) r1, r2, area

    apertureRadius = radius

!write(*, *) "PointSourcePhotometry1 radius=", radius, " x=", source%xpos, " y=", source%ypos

    area = PI * radius * radius ! The required aperture area

    if (source%innerBckRadius > 0.0 .and. source%outerBckRadius > 0.0) then
        r1 = source%innerBckRadius
        r2 = source%outerBckRadius
    else 
        r1 =  innerRadius
        r2 =  outerRadius
    end if

!write(*, *) "Line 1623 source%innerBckradius=", source%innerBckRadius
!write(*, *) "Line 1623 source%outerBckradius=", source%outerBckRadius
 
!write(*, *) "PointSourcePhotometry1 line 1623 annulii - innerradius=", innerRadius, " outer=", outerRadius

    ! Get the total counts within the aperture   


    sourceCounts = GetApertureCounts(source%xPos - 0.5, source%yPos - 0.5, radius)
 
!write(*, *) "Line 1614 source counts=", sourceCounts


   ! Get the background   

    if (source%rawBackgroundCountRate .gt. -0.00001) then
        backgroundCountsPerPixel = source%rawBackgroundCountRate
    else

        backGroundCountsPerPixel = GetBackGroundMode(source%xPos - 0.5, source%yPos - 0.5, &
                              r1, r2, backGroundCountsError)




    end if

!   write(*, *) "Background counts per pixel=",  backGroundCountsPerPixel, " aperturearea=", apertureArea


    backGroundCounts = backGroundCountsPerPixel * apertureArea
    backGroundCountsError = backGroundCountsError * apertureArea 
  
!write(*, *) "pointsourcephotometry1 aperture radius=", apertureRadius
!write(*, *) "inner background radius=", r1, " outer=", r2
!write(*, *) "source counts=", sourceCounts, " background counts=", backgroundcounts, " radius=", radius
!write(*, *) "exposure=", source%exposure 
       
  
    source%rawBackGroundCountRate = backGroundCounts / source%exposure
    source%rawBackGroundCountRateErr = sqrt(backgroundcounts) / source%exposure

  
!write(*, *) "omphotometry line 1658 raw background count rate=",   source%rawBackGroundCountRate, &
!" error=", source%rawBackGroundCountRateErr  

    ! Subtract the background and convert to count-rate
    sourceCounts = sourceCounts - backgroundcounts
    source%rawCountRate = sourceCounts / source%exposure
    source%rawCountRateErr = ComputeSourceRateError(sourceCounts, backgroundCounts, source%exposure) 


!    source%Significance = ComputeSourceSignificance(source%rawCountRate, &
!                      source%rawBackGroundCountRate, source%exposure)
 


!    if (apertureArea < area) then  ! source near an edge
!write(*, *) "Aperturearea =", apertureArea, " area=", area 


!write(*, *) "pointsourcephotometry1- x=", source%xpos, " y=", source%ypos
!write(*, *) "Computed raw count rates"

!write(*, *) "rawsource rate=",  source%rawCountRate, " background=", &
!     source%rawBackgroundCountRate, " apertureArea=", apertureArea
!write(*, *) "rawcountrateerr=",   source%rawCountRateErr, " rawbackgroundcountrateerr=", source%rawBackGroundCountRateErr  

!write(*, *) "PointSourcePhotometry1 line 1674 annulii - innerradius=", innerRadius, " outer=", outerRadius


end subroutine PointSourcePhotometry1

!*********************************************
! Pont2SourcePhotometry
! Computes the counts for a point-source
! having an aperture radius > coincidence-loss
! radius
!*********************************************
subroutine PointSourcePhotometry2(source, radius)

    ! Input (unmodfied) variables
    real(single), intent(in) :: radius 
    type(SourceParameters), intent(inout) :: source
      
    real(double) :: rawCounts, correctedCounts, countRate
    real(double) :: f
    real(double) :: backgroundRate, backgroundCounts, backgroundMode
    real(single) :: backGroundCountsError
    real(double) :: area, pixelArea
    real(double) :: correction
 
    ! Local variables
    integer (int32) :: i, j
    integer (int32) :: i1, i2, j1, j2
    real(single) :: xPos, yPos, x0, y0 !   

!write(*, *) "mssllib  PointSourcePhotometry2 line 1433 radius=", radius, " x=", source%xpos, " y=", source%ypos

    xPos = source%xPos - 0.5
    yPos = source%yPos - 0.5
    apertureRadius = radius        
           
    ! Determine upper and lower array indices for
    ! summing counts
   
    i1 = int(xPos - radius - 0001)
    i1 = max(il, i1)
  
    i2 = int(xPos + radius + 0.0001)
    i2 = min(iu, i2)  

    j1 = int(yPos - radius - 0.0001)
    j1 = max(jl, j1)

    j2 = int(yPos + radius + 0.0001)
    j2 = min(ju, j2)
 

    apertureArea = 0.0_double
    rawCounts = 0.0_double
    correctedCounts = 0.0_double
    area = 0.0_double
 
    do i = i1, i2
        x0 = real(i) + 0.5

        do j = j1, j2
            y0 = real(j) + 0.5 

           
            ! Compute the coincidence-loss correction for this pixel
            countRate = GetApertureCounts(x0, y0, coincLossRadius) / source%exposure
            apertureRadius = radius                   

!            write(*, *) "Line 1116 i=", i, " j=", j, " x0=", x0, " y0=", y0             
            pixelArea = EffectivePixelArea(x0 - xPos, y0 - yPos)

            f = pixelArea * ccdImage(i, j)             
            rawCounts = rawCounts + f
            correctedCounts = correctedCounts + f * &
                      CoincidenceLossCorrection(countRate, sqrt(real(countRate)), frameTime)
                   
            area = area + pixelArea
          
        end do
    end do
!write(*, *) "radius=", radius
!write(*, *) "Line 1128 area=", area    
!write(*, *) "raw counts=", rawCounts, " corrected=", correctedCounts

    ! Get the background   
    if (source%rawBackgroundCountRate .gt. -0.0001) then
        backgroundMode = source%rawBackgroundCountRate
    else
        backGroundMode = GetBackGroundMode(source%xPos - 0.5, source%yPos - 0.5, &
                               source%innerBckRadius, source%outerBckRadius, &
                               backGroundCountsError)
    end if

    backgroundCounts = backGroundMode * area
    source%rawBackGroundCountRate = backGroundCounts / source%exposure
    source%rawBackGroundCountRateErr = backGroundCountsError * area / source%exposure
  
    ! Subtract the background
    rawCounts = rawCounts - backGroundCounts 

  
    source%rawCountRateErr = ComputeSourceRateError(rawCounts, backGroundCounts, source%exposure)

  


    ! Compute the corrected rates
    backGroundRate = PI * coincLossRadius * coincLossRadius * backGroundMode / source%exposure

    ! The coincidence-loss correction for each background pixel
    correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), frameTime)

    source%corrBackGroundCountRate = source%rawBackGroundCountRate * correction

    source%corrBackGroundCountRateErr = source%rawBackGroundCountRateErr * correction

    source%corrCountRate =  source%rawCountRate * correction

    source%corrCountRateErr = ComputeSourceRateError(correctedCounts, backgroundCounts, source%exposure)
           
    
    ! Compute the signal-to-noise (source counts / sqrt(source counts + background counts)   
   
    source%Significance = ComputeSourceSignificance(source%rawCountRate, &
                      source%rawBackGroundCountRate, source%exposure)
                  
    source%apertureRadius = radius * float(binX)  
    
    source%coincidenceLossCorr = (source%corrCountRate - source%rawCountRate) * 100.0 &
             / source%rawCountRate 
    source%rawCountRate = source%rawCountRate / source%exposure
    source%corrCountRate = source%corrCountRate / source%exposure

end subroutine PointSourcePhotometry2       
!*************************************************************
! Point-source photometry
!*************************************************************
subroutine PointSourcePhotometry3(source, radius, backgroundMode) 
    type(SourceParameters), intent(inout) :: source
    real(single), intent(in) ::radius, backGroundMode
    real(double) :: sourceCounts, backgroundCounts
    real(single) :: backgroundCountsError
    
    real(double) :: backGroundCountsPerPixel
    real(single) r1, r2

    apertureRadius = radius

    if (source%innerBckRadius > 0.0 .and. source%outerBckRadius > 0.0) then
        r1 = source%innerBckRadius
        r2 = source%outerBckRadius
    else 
        r1 =  innerRadius
        r2 =  outerRadius
    end if
  
    ! Get the total counts within the aperture   


    sourceCounts = GetApertureCounts(source%xPos - 0.5, source%yPos - 0.5, radius)
 

    ! Get the background   

    backGroundCountsPerPixel = backgroundMode

    backGroundCounts = backGroundCountsPerPixel * apertureArea

    source%rawBackGroundCountRate = backGroundCounts / source%exposure
    backGroundCountsError = backGroundCountsError * apertureArea 
  
    ! Subtract the background and convert to count-rate
    sourceCounts = sourceCounts - backgroundcounts
    source%rawCountRate = sourceCounts / source%exposure
    source%rawCountRateErr = ComputeSourceRateError(sourceCounts, backgroundCounts, source%exposure) 


end subroutine PointSourcePhotometry3


!*****************************************
! setQualityFlag
!***************************************
subroutine setQualityFlag

    currentSource%qualityFlag = ibset(currentSource%qualityFlag, 0)
    
end subroutine setQualityFlag
!**********************************
! subroutine GetCounts
!********************************** 
subroutine GetCounts1(image, source)
    real(single), pointer :: image(:, :)
    type(SourceParameters), intent(inout) :: source
!    real(single) :: trueCountRate
  
    CCDImage => image

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
      
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1     

    apertureRadius = source%apertureRadius
     
    source%rawBackgroundCountRate = &
             GetApertureCounts(source%xPos, source%yPos, apertureRadius )
               

    source%rawBackGroundCountRateErr = sqrt( source%rawBackgroundCountRate) 
 
    
    
   
    
end subroutine GetCounts1
!***************************************************
! subroutine GetCounts2
! Computes the total counts within a circle 
! of radius radius centred on xpos, ypos
!*************************************************** 
subroutine GetCounts2(image, xPos, yPos, radius, counts, area)
    real(single), pointer :: image(:, :)
    real(single), intent(in) :: xPos, yPos, radius
    real(single), intent(out) :: counts
    real(single), optional, intent(out) :: area
    CCDImage => image


    nx = size(image, 1)
    ny = size(image, 2)

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
      
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1     

    apertureRadius = radius
     
    counts = GetApertureCounts(xPos, yPos, radius )
    nullify(CCDImage)           
    if (present(area)) area = apertureArea   ! The actual aperture area used

end subroutine GetCounts2

!**********************************
! subroutine GetBackgroundCounts1
!********************************** 
subroutine GetBackgroundCounts1(image, xPos, yPos, &
                                radius, radius1, radius2, counts)
    real(single), pointer :: image(:, :)
    real(single), intent(in) :: xPos, yPos, radius, radius1
    real(single), intent(inout) :: radius2
    real(single), intent(out) :: counts
    real(single) :: error
  
    CCDImage => image

    nx = size(image, 1)
    ny = size(image, 2)

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
      
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1     

    apertureRadius = radius

    maxNumPixels = 30000
    if (.not. associated(backgroundPixel)) &
        backgroundPixel => AllocateSingle(1, maxNumPixels, "backgroundPixel (AperturePhotometry)")
     

    ! Compute the background counts per pixel from the annulus    
    counts = GetBackGroundMode1(   &
              xPos,    &   ! The centre x-pixel coordinate of the aperture
              yPos,    &   ! The centre y-pixel corrdinate of the aperture 
              radius1, &   ! The aperture inner radius
              radius2, &   ! The aperture outer radius
              error)       ! The error
               

!  write(*, *) "Line 1931 Getbackgroundcounts1 mode=", counts, " radius=", radius

    counts = counts * PI * radius * radius


    deallocate(backgroundPixel)
    nullify(backgroundPixel)
    nullify(CCDImage)    

end subroutine GetBackgroundCounts1

!******************************************************
!
!*****************************************************

subroutine AperturePhotometry3(source, n, image, levelImage, &
                               sourceQualityImage, &
                               radius, minSignificance)

    type(SourceParameters), pointer :: source(:)
    integer, intent(in) :: n       
    real(single), pointer :: image(:, :) 
    integer(int16), pointer :: levelImage(:, :) 
    logical(bool), pointer :: sourceQualityImage(:, :)
 
    real(single), intent(in) :: radius, minSignificance
  
    integer :: i
    type(SourceParameters), target :: tempSource


    qualityImage => sourceQualityImage ! The quality image (false=good pixels, true=bad pixels)
 
    call Initialise(int(size(source), int16), int(size(image, 1), int16), int(size(image, 2), int16))
    CCDImage => image 

    sourceNumImage => levelImage
    
!write(*, *) "**********************************************************************"
!write(*, *) "aperturephotometry3"
!write(*, *) "**********************************************************************"

 
    do i = 1, n
        
        if (source(i)%extensionFlag .gt. 0) cycle
!        if (source(i)%confusionFlag .eq. 4) cycle
        if (btest(source(i)%confusionFlag, 2)) cycle


        if (source(i)%significance .gt. minSignificance) cycle


        tempSource = source(i) 
        tempSource%innerBckRadius = 0.0
        tempSource%outerBckRadius = 0.0

        currentSource => tempsource 
!write(*, *) "Recomputing photometry on source ", i, " old raw count-rate=", source(i)%rawCountRate, &
! " old raw background count-rate=", source(i)%rawBackgroundCountRate

        tempsource%rawBackgroundCountRate = -1.0


        call PointSourcePhotometry1(tempsource, radius)
        tempsource%ApertureRadius = radius * real(binX)

        call PointSourceCoincLossCorr(tempsource, 12.0, frameTime)     



       

!write(*, *) "Source ", i, " x=", source(i)%xpos, " y=", source(i)%ypos
!write(*, *) "old significance=", source(i)%significance, &
!  " new significance=", tempSource%significance
   

  

       if (tempSource%significance .gt. source(i)%significance)  then
           source(i) = tempSource
       end if

!write(*, *) "newsignificance=", source(i)%significance

             
!write(*, *) "New count-rate=", source(i)%rawCountRate   


    end do

    nullify(CCDImage, qualityImage, sourceNumImage, currentSource)
    call deleteArrays


end subroutine AperturePhotometry3
end module omPhotometry_mod

























