! !
! Authors:  Christopher Brindle - MSSL
!           Andrew King - MSSL
!           Tim Harries - MSSL
!          
! Description: Combines OSW source list into an observation source list
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!
! History: 
   
!
!************************************************************************************************
module omsourceflags_mod

  use cal
  use dal
  use Errorhandling
  use omallocation_mod
  use omconstants_mod
  use omutility_mod
  use ominterpolate_mod
  use omnumerics_mod
  use omellipse_mod
  implicit none 


public ::  LookForReadOutStreaks
public ::  LookForStarSpikes
public ::  setSourceFlags 

interface ConvertCoordsToFullFrame
    module procedure ConvertCoordsToFullFrame1,  &
                     ConvertCoordsToFullFrame2, &
                     ConvertCoordsToFullFrame3

end interface
interface ReadOutStreaks
    module procedure ReadOutStreaks1, ReadOutStreaks2

end interface

interface SmokeRings
    module procedure SmokeRings1, SmokeRings2

end interface
    
  
   contains

!****************************************************************************
! QFLAG source quality flags convention
! BIT NO             FLAGGING                         Integer
!
!   0                BAD PIXEL                                 1
!   1                READ-OUT STREAK                           2
!   2                SMOKE-RING                                4
!   3                SOURCE ON STAR-SPIKE                      8
!   4                Mod-8 source                             16
!   5                Source lies near to a bright source      32
!   6                Source within central enhancement        64
!*****************************************************************************
!****************************************************************************
! CFLAG source confusion flags convention
! BIT NO             FLAGGING                                           Integer
!
!   0                one or more source within 0.5-1.0 aperture radius  1 
!   1                one or more source within 0.5 aperture radius      2
!   2                point source lies with 1.0 aperture radius of      4
!                edge or some point of extended source lies on edge      
!   3           Point-source lies partly or entirely within extended    8
!                source region 
!*****************************************************************************






!****************************************************
!
!***************************************************
subroutine SetSourceFlags(n, source, x0, y0, image, backgroundimage, levelImage, qualityimage, plevelimage, elevelimage)
    integer(int16), intent(in) :: n
    type(sourceParameters), pointer :: source(:)
    integer, intent(in) :: x0, y0

    real(single), pointer, dimension(:, :) :: image, backgroundimage
    integer(int16), pointer, dimension(:, :) :: levelimage
    integer(int16), optional, pointer, dimension(:, :) ::  plevelimage, elevelimage
    logical(boolean), pointer :: qualityImage(:, :)
    real(single), pointer :: xCoord(:) => null(), yCoord(:) => null()
    integer :: i
    integer(int16) :: nx, ny
    if (.not. associated(source) .or. n .lt. 1) return


    xCoord => allocateSingle(1_int16, n, "xCoord")
    yCoord => allocateSingle(1_int16, n, "yCoord")

    xCoord(1 : n) = source(1 : n)%xPos
    yCoord(1 : n) = source(1 : n)%yPos

    nx = ubound(image, 1)
    ny = ubound(image, 2)
  
    call ConvertCoordsToFullFrame(xCoord, yCoord, x0, y0, binx, biny)

    call setCornerFlags(xCoord, yCoord, source, n)


    call setCentralRegionFlags(xCoord, ycoord, source, n)

    call starspikes(int(n, int32), source)



    call ReadOutStreaks(int(n, int32),          &
                  source%rawCountRate,          &
                  xCoord,                       &
                  source%extensionFlag,         &
                  source%qualityFlag)

  


    call SmokeRings(int(n, int32),              &
                  source%rawCountRate,          &
                  xCoord, yCoord,               &
                  source%extensionFlag,         &
                  source%qualityFlag)


    call LookForReadOutStreaks(source, image, backgroundimage, levelimage, qualityImage)

!write(*, *) "Line 2189 Number of sources with quality flag=1 or 2 =", count(source%qualityFlag .eq. 1 .or. source%qualityFlag .eq. 2)

!    call LookForStarSpikes(source, image, 20 / binx)


    !**************************************************
    ! Flag sources close to a bright source
    !*************************************************
    call BrightSourceProximityFlags(int(n, int32),      &
                  source%corrCountRate,                 &
                  xCoord, yCoord,                       &
                  source%qualityFlag,                   &
                  source%confusionFlag )


 
    deallocate(xCoord, ycoord)
    nullify(xCoord, ycoord)

    if (present(plevelimage) .and. present(elevelImage)) then 
      
        do i = 1, n

            if (source(i)%extensionFlag .eq. 0) then
                ! Check for any point sources within an extended source region

                if (any(plevelimage .eq. source(i)%number .and. elevelimage .lt. 0)) then
                    source(i)%confusionFlag = ibset(source(i)%confusionFlag, 3)  

                end if
            else

                ! Check for any point of an extended-source lying on image boundary
                if (any(levelimage(0,  : ) .eq. source(i)%number) .or. &
                    any(levelimage(nx,  : ) .eq. source(i)%number) .or. &
                    any(levelimage( :,  0 ) .eq. source(i)%number) .or. &
                    any(levelimage( :,  ny ) .eq. source(i)%number) ) then
                   source(i)%confusionFlag = ibset(source(i)%confusionFlag, 2)  

             
                     
                end if
         
 

            end if

        end do

    end if



end subroutine SetSourceFlags
!***********************************************************************
!
!**********************************************************************
subroutine setCentralRegionFlags(x, y, source, n)
    real(single) :: x(:), y(:)
    type(sourceParameters), pointer :: source(:)
    integer(int16), intent(in) :: n
    real(single), parameter :: xc = 992.0, yc = 1063.0, radius = 160.0
    real(single) :: dx, dy
    integer(int16) :: i


    do i = 1, n
     
        if (x(i) .lt. (xc - radius) .or. x(i) .gt. (xc + radius) .or. &
            y(i) .lt. (yc - radius) .or. y(i) .gt. (yc + radius)) cycle
        dx = x(i) - xc
        dy = y(i) - yc
        if (sqrt(dx * dx + dy * dy) .lt. radius) then
            source(i)%qualityFlag = ibset(source(i)%qualityFlag, 6)  ! BIT 5 set for sources in central enhanced region
        end if
    end do
end subroutine setCentralRegionFlags

!***********************************************************************
!
!**********************************************************************
subroutine setCornerFlags(x, y, source, n)
    real(single) :: x(:), y(:)
    type(sourceParameters), pointer :: source(:)
    integer(int16), intent(in) :: n
    
   
    integer(int16) :: i
  
    
    real(single), parameter :: xcentre = 979.2, ycentre = 1016.2
!    real(single), parameter :: semia = 1335.0, semib = 1335.0

!  real(single), parameter :: semia = 1310.0, semib = 1310.0

  real(single), parameter :: semia = 1320.0, semib = 1320.0

    real(single), parameter :: theta = 90.0 * degreesToRadians
    integer :: noCornerFlagsSet

  

    noCornerFlagsSet = 0
    do i = 1, n
     
      if (.not. PointInEllipse(x(i), y(i), xcentre, ycentre, semia, semib, theta)) then
    
    
                source(i)%confusionFlag = ibset(source(i)%confusionFlag, 2)
                noCornerFlagsSet = noCornerFlagsSet + 1
            end if
  
cycle
  
        if (x(i) .lt. 100.0 .and. y(i) .lt. 100.0) then ! Check near bottom-left hand corner
       
            if (.not. PointInEllipse(x(i), y(i), xcentre, ycentre, semia, semib, theta)) then
    
    
                source(i)%confusionFlag = ibset(source(i)%confusionFlag, 2)
                noCornerFlagsSet = noCornerFlagsSet + 1
            end if
        elseif (x(i) .gt. 1940.0 .and. y(i) .lt. 100.0) then ! Check near bottom-right hand corner
            if (.not. PointInEllipse(x(i), y(i), xcentre, ycentre, semia, semib, theta)) then 
    
                source(i)%confusionFlag = ibset(source(i)%confusionFlag, 2) 
                noCornerFlagsSet = noCornerFlagsSet + 1   
            end if
        elseif (x(i) .gt. 1940.0 .and. y(i) .gt. 1940.0) then ! Check near top-right hand corner
            if (.not. PointInEllipse(x(i), y(i), xcentre, ycentre, semia, semib, theta)) then
    

                source(i)%confusionFlag = ibset(source(i)%confusionFlag, 2)
                noCornerFlagsSet = noCornerFlagsSet + 1   
            end if
         elseif (x(i) .lt. 100.0 .and. y(i) .gt. 1940.0) then ! Check near top-left hand corner
            if (.not. PointInEllipse(x(i), y(i), xcentre, ycentre, semia, semib, theta)) then
      
                source(i)%confusionFlag = ibset(source(i)%confusionFlag, 2)  
                noCornerFlagsSet = noCornerFlagsSet + 1   
            end if
        end if
             

    end do


    write(errstr,'(a, i6)') "No of corner flags set=", noCornerFlagsSet
    call message(AppMsg, Verbose)


end subroutine setCornerFlags



 
!*********************************************
!
!
!********************************************
Subroutine LookForReadOutStreaks(source, image, backgroundimage, levelImage, qualityImage)
    type(sourceParameters), pointer :: source(:)

    real(single), pointer, dimension(:, :) :: image, backgroundImage
    integer(int16), pointer :: levelimage(:, :)
    logical(boolean), pointer :: qualityImage(:, :)

    integer :: i0, i, j, k, il, iu, i0max
    integer :: nx, ny, npix, nStreaks, dx, dxmin, dxmax
    real(single) :: width
    real(double) :: mean1, mean2, mean3

    logical :: streak

    if (.not. associated(source)) return

    write(errstr,'(a)') "Checking for read-out streaks"
    call message(AppMsg, Verbose)

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    dx = 8 / binx
    dxmin = 4 / binx
    dxmax = 16 / binx
    nStreaks = 0
    width = real(14 / binx)

    i0 = dx + 1   ! The central x pixel of a possible streak

    do i = 1, nx - 1
!write(*, *) "Line 197 i=", i, " i0=", i0
    if (i0 .ge. nx) exit 
        streak = .false.   ! Not found  a streak yet

        mean1 = 0.0_double
        npix = 0

        do j = 0, ny
            if (any(levelImage(i0 - 1 : i0 + 1, j) .ne. 0) .or. any(qualityImage(i0 - 1 : i0 + 1, j)) .or. &
                any( image(i0 - 1 : i0 + 1, j) .gt. (backgroundImage(i0 - 1 : i0 + 1, j) + 10.0))) cycle
            npix = npix + 1

            mean1 = mean1 + dble(image(i0 - 1, j)) + dble(image(i0, j) + image(i0 + 1, j))

        end do

        if (npix .gt. ny / 2) then
            mean1 = mean1 / dble(3 * npix)
            if (mean1 .le. 0.0_double) then
                i0 = i0 + 1
                cycle
            end if

        else
           i0 = i0 + 1
           cycle
        end if

!write(*, *) "Line 224 i=", i, " i0=", i0


        do dx = dxmin, dxmax

            il = i0 - dx
            if (il .lt. 0) cycle
            iu = i0 + dx
            if (iu .gt. nx) exit

            npix = 0
            mean2 = 0.0_double
            do j = 0, ny
                if (any(levelImage(il : iu, j) .ne. 0) .or. any(qualityImage(il : iu, j)) .or. &
                    any(image(il : iu, j) .gt. backgroundImage(il : iu, j) + 10.0)) cycle


                npix = npix + 1
                mean2 = mean2 + dble(image(il + 1, j)) +  dble(image(iu - 1, j)) + &
                                dble(image(il + 2, j)) +  dble(image(iu - 2, j))

            end do

            if (npix .gt. ny / 2) then
                mean2 = mean2 / dble(4 * npix)

                if (mean1 .gt. mean2 + max(1.0_double, 0.5 * sqrt(mean2 * mean2 + mean1 * mean1))) then ! Read out streak
                    streak = .true.
                    i0max = i0
                    exit
                end if

            end if
        end do

        if (.not. streak) then
            i0 = i0 + 1
            cycle
        end if

!write(*, *) "Line 263 i=", i, " i0=", i0


        do k = i0 - 3, i0 + 3 ! Refine the centre x position of the streak
            if(k + 1 .gt. nx) exit
            npix = 0

            mean3 = 0.0_double
            do j = 0, ny
                if (any(levelImage(k - 1 : k + 1, j) .ne. 0) .or. any(qualityImage(k - 1 : k + 1, j)) .or. &
                     any(image(k - 1 : k + 1, j) .gt. backgroundImage(k - 1 : k + 1, j) + 10.0)) cycle

                npix = npix + 1

                mean3 = mean3 + dble(dble(image(k - 1, j))) + dble(image(k, j) + dble(image(k + 1, j)))

            end do

            if (npix .gt. ny / 2) then
                mean3 = mean3 / dble(3 * npix)
                if (mean3 .gt. mean1) then
                    i0max = k
                    mean1 = mean3
                end if
            end if

        end do

!write(*, *) "nx=", nx, " ny=", ny, " mean1=", mean1, " mean2=", mean2, " mean3=", mean3
        write(errstr,'(a, i6, a, f7.2, a, f7.2)') &
                      "Possible read-out streak centred on column ", i0max
        call message(AppMsg, Verbose)

        where(abs(source%xpos - real(i0max) + 1.0) .lt. width &
                 .and. source%significance .lt. 50.0) source%qualityFlag = ibset(source%qualityFlag, 1)  ! BIT 1 set for sources on read-out streaks

        nStreaks = nStreaks + 1
        i0 = i0max + 20 / binx
    
    end do
    write(errstr,'(a, i5)') "Number of possible read-out streaks found=", nstreaks

    call message(AppMsg, Verbose)

    write(errstr,'(a, i5)') "Number of source flags set=", count(source%qualityFlag .eq. 2 .or. source%qualityFlag .eq. 3)

    call message(AppMsg, Verbose)

end Subroutine LookForReadOutStreaks

!*********************************************
!
!
!*********************************************
Subroutine LookForStarSpikes(source, image, nsize)
    type(sourceParameters), pointer :: source(:)

    real(single), pointer :: image(:, :)
    integer, intent(in) :: nsize

    integer :: i, j, k, l, n, n1, n2, n3, n4
    integer :: nx, ny, nSpikes
    real(single) :: xl, xu, yl, yu
    real(single) :: xcentre, ycentre, xpos, ypos, theta
    real(single) :: radius, dr, cs, sn, dx, dy


    real(single) :: f1(nsize), f2(nsize), f3(nsize)
    real(single) :: tol
    integer :: nsize1, nsize2, nMin
    integer :: rMin, rMax

    if (.not. associated(source)) return

    n = size(source)

    write(errstr,'(a)') "Checking for stars with diffraction spikes"
    call message(AppMsg, Verbose)

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    nSpikes = 0

    xl = 0.5
    yl = 0.5
    xu = size(image, 1) - 0.5
    yu = size(image, 2) - 0.5


    nsize1 = nsize / 2
    nsize2 = nsize1 + 1

    nMin = nsize

    rmin = 30 / binx
    rMax = 80 / binx

    tol = real(8 / binx)
    l = 0
    dr = 1.0 / real(binx)
iLoop : do i = 1, n
        if (source(i)%rawCountRate .lt. 50.0) cycle


        xcentre = source(i)%xpos - 0.5
        ycentre = source(i)%ypos - 0.5
!     write(*, *) "xcentre=", xcentre, " ycentre=", ycentre, " rmin=", rmin, " rmax=", rmax

        do k = rmin, rmax
!write(*, *) "rmin=", rmin, " rmax=", rmax


            call GetStarSpikeCounts(image, nsize, xcentre, ycentre, 45.0, real(k), dr, f1, f2, f3)
            n1 = count(f2 .gt. f1)
            n2 = count(f3 .gt. f1)


            call GetStarSpikeCounts(image, nsize, xcentre, ycentre, 45.0, -real(k), -dr, f1, f2, f3)


            n3 = count(f2 .gt. f1)
            n4 = count(f3 .gt. f1)

!write(*, *) "Line 2400 n1=", n1, " n2=", n2, " n3=", n3, &
!" n4=", n4, " nmin=", nmin, " nsize=", nsize, " nsize1", nsize1, " nsize2=", nsize2

            if (n1 .ge. nmin) l = l + 1
            if (n2 .ge. nmin) l = l + 1
            if (n3 .ge. nmin) l = l + 1
            if (n4 .ge. nmin) l = l + 1


            call GetStarSpikeCounts(image, nsize1, xcentre, ycentre, 135.0, real(k), dr, f1, f2, f3)

            n1 = count(f2 .gt. f1)
            n2 = count(f3 .gt. f1)



            call GetStarSpikeCounts(image, nsize, xcentre, ycentre, 135.0, -real(k), -dr, f1, f2, f3)



            n3 = count(f2 .gt. f1)
            n4 = count(f3 .gt. f1)

            if (n1 .ge. nmin) l = l + 1
            if (n2 .ge. nmin) l = l + 1
            if (n3 .ge. nmin) l = l + 1
            if (n4 .ge. nmin) l = l + 1


!write(*, *) "Line 2438 n1=", n1, " n2=", n2, " n3=", n3, " n4=", n4, &
!" nmin=", nmin, " nsize=", nsize, " nsize", nsize1, " nsize2=", nsize2

            if (l .gt. 7) then

                nspikes = nspikes + 1

                write(errstr,'(a, i5, a, f6.2, a, f6.2)') &
                        "Possible diffraction spikes found associated with star ", &
                  i, " x=", source(i)%xpos, " y=", source(i)%ypos

                call message(AppMsg, Verbose)

                !*******************************************
                ! Flag sources near spikes
                !*******************************************

                theta = 45.0 * degreesToRadians
                sn = sin(theta)
                cs = cos(theta)

                rmin = -180 / binx

                do j = 1, 361

                    radius = real(rmin) + float(j - 1) * dr

                    xpos = xcentre + radius * cs
                    if (xpos .lt. xl .or. xpos .gt. xu) cycle

                    ypos = ycentre + radius * sn
                    if (ypos .lt. yl .or. ypos .gt. yu) cycle

                    do l = 1, n
                        if (l .eq. i) cycle
                        if (source(l)%qualityFlag .ge. 8) cycle
                        if (source(l)%rawCountrate .ge. 20.0) cycle

                        dx = source(l)%xpos - xpos

                        dy = source(l)%ypos - ypos
                        if (sqrt(dx * dx + dy * dy) .le. tol) then

                            source(l)%qualityFlag = ibset(source(l)%qualityFlag, 3)  ! BIT 3 set for sources on star spikes

!                            write(errstr,'(a, i5, a)') "Setting quality flag of source ", l,  " to 8"

!                            call message(AppMsg, Verbose)

                        end if
                    end do
                end do

                theta = 135.0 * degreesToRadians
                sn = sin(theta)
                cs = cos(theta)

                do j = 1, 361

                    radius = real(rmin) + float(j - 1) * dr

                    xpos = xcentre + radius * cs
                    if (xpos .lt. xl .or. xpos .gt. xu) cycle

                    ypos = ycentre + radius * sn
                    if (ypos .lt. yl .or. ypos .gt. yu) cycle

                    do l = 1, n
                        if (l .eq. i) cycle
                        if (source(l)%qualityFlag .ge. 8) cycle
                        if (source(l)%rawCountrate .ge. 20.0) cycle

                        dx = source(l)%xpos - xpos
                        dy = source(l)%ypos - ypos
                        if (sqrt(dx * dx + dy * dy) .le. tol) then

                            source(l)%qualityFlag = ibset(source(l)%qualityFlag, 3)  ! BIT 3 set for sources on star spikes

!                            write(errstr,'(a, i5, a)') "Setting quality flag of source ", l,  " to 8"

!                            call message(AppMsg, Verbose)

                        end if
                    end do
                end do

                cycle iLoop

            end if
        end do
    end do iLoop

    write(errstr,'(a, i5)') "Number of possible diffraction-spikes found=", nspikes

    call message(AppMsg, Verbose)

    write(errstr,'(a, i5)') "Number of source flags set=", count(source%qualityFlag .ge. 8)

    call message(AppMsg, Verbose)

end Subroutine LookForStarSpikes

!**************************************************************************************
!
!
!*************************************************************************************

subroutine GetStarSpikeCounts(image, nsize1, xcentre, ycentre, angle, startRadius, dr, f1, f2, f3)
    real(single), pointer :: image(:, :)
    integer, intent(in) :: nsize1

    real(single), intent(in) :: xcentre, ycentre, startRadius, angle, dr
    real(single), dimension(1 : nsize1) :: f1, f2, f3
    real(single) :: radius, cs, sn, theta, xpos, ypos, xpos1, ypos1
    real(single) :: dx, dy, xl, yl, xu, yu
    integer :: j

    theta = angle * degreesToRadians

    sn = sin(theta)
    cs = cos(theta)

    dx = real(8 / binx)
    dy = real(8 / binx)

    xl = 0.5
    yl = 0.5
    xu = size(image, 1) - 0.5
    yu = size(image, 2) - 0.5

    f1 = 0.0
    f2 = 0.0
    f3 = 0.0

    do j = 1, nsize1
        radius = startRadius + real(j - 1) * dr

        xpos = xcentre + radius * cs
        if (xpos .lt. xl .or. xpos .gt. xu) cycle

        ypos = ycentre + radius * sn
        if (ypos .lt. yl .or. ypos .gt. yu) cycle

        f1(j) = bilinearinterpolate(image, xPos, yPos)
        xpos1 = xpos - dx * sn
        if (xpos1 .lt. xl .or. xpos1 .gt. xu) cycle

        ypos1 = ypos + dy * cs
        if (ypos1 .lt. yl .or. ypos1 .gt. yu) cycle

        f2(j) = bilinearinterpolate(image, xPos1, yPos1)
        xpos1 = xpos + dx * sn
        if (xpos1 .lt. xl .or. xpos1 .gt. xu) cycle

        ypos1 = ypos - dy * cs
        if (ypos .lt. yl .or. ypos .gt. yu) cycle

        f3(j) = bilinearinterpolate(image, xPos1, yPos1)


    end do

end subroutine GetStarSpikeCounts

!************************************************************
!  Converts source-list file coordinates to full-frame
!  coordinates
!
!************************************************************
subroutine ConvertCoordsToFullFrame1(xCoord, yCoord, il, iu, sourceListDataSetH)

    real(single) :: xCoord(:), yCoord(:)
    integer, intent(in) :: il, iu
    type(dataSetT) :: sourceListDataSetH
 
    integer(int32) :: xbin, ybin
    logical(boolean) :: binBpe
 
    character(100) :: windowString
    character(1) :: c
    integer :: windowX0, windowY0, windowX1, windowY1, winDx, winDy

    xbin = int32Attribute(sourceListDataSetH, "BINAX1")      ! binning factors
    ybin = int32Attribute(sourceListDataSetH, "BINAX2")
 
    if (xbin .ne. ybin) then
        call fatal("diffBin", "This version cannot deal with different binning factors in x and y")
    end if

    if (hasAttribute(sourceListDataSetH, "DETSEC")) then
        windowString = stringAttribute(sourceListDataSetH, "DETSEC")
        read (windowString, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
                  c, windowX0, c, windowX1, c,  windowY0, c, windowY1, c
       
        winDX = windowX1 - windowX0 + 1
        winDY = windowY1 - windowY0 + 1
           
      
    end if

    binBPE = booleanAttribute(sourceListDataSetH, "BINBPE")
 
    ! Calculate the image binning.
    xBin = 2 ** xbin
    yBin = 2 ** ybin


    if (binBPE) then
        xBin = xBin * 2
        yBin = yBin * 2
    end if

    windowX0 = windowX0 * xBin
    windowY0 = windowY0 * yBin


    call ConvertCoordsToFullFrame2(xCoord(il : iu), yCoord(il : iu), &
                windowX0, windowY0, xBin, yBin)


end subroutine ConvertCoordsToFullFrame1



!************************************************************
!
!
!************************************************************
subroutine ConvertCoordsToFullFrame2(xCoord, yCoord, x0, y0, xBin, yBin)
    real(single), dimension(:) :: xCoord, yCoord
    integer, intent(in) :: x0, y0, xBin, yBin
    integer :: i

    ! Convert coordinates to full-frame unbinned coordinates  
    do i = 1, size(xCoord)

        xCoord(i) = xCoord(i) * real(xBin) + real(x0)
        yCoord(i) = yCoord(i) * real(yBin) + real(y0)

    end do



end subroutine ConvertCoordsToFullFrame2
!************************************************************
!
!
!************************************************************
subroutine ConvertCoordsToFullFrame3(xCoord, yCoord, x0, y0, xBin, yBin)
    real(single) :: xCoord, yCoord
    integer, intent(in) :: x0, y0, xBin, yBin


    xCoord = xCoord * real(xBin) + real(x0)
    yCoord = yCoord * real(yBin) + real(y0)

end subroutine ConvertCoordsToFullFrame3
!************************************************************
!
!
!************************************************************
subroutine ConvertCoordsFromFullFrame(xCoord, yCoord, x0, y0, xBin, yBin)
    real(single) :: xCoord, yCoord
    integer, intent(in) :: x0, y0, xBin, yBin


    xCoord = (xCoord - real(x0)) / real(xbin)
    yCoord = (yCoord - real(y0)) / real(yBin)

end subroutine ConvertCoordsFromFullFrame



!***************************************************
! Subroutine 
!
!***************************************************
Subroutine ReadOutStreaks1(n, m, jIndex,   &
                           countTable,     &
                           x,              &
                           extensionFlag,  &
                           qualityFlag)
    integer, intent(in) :: n, m, jIndex(:)
    real(single) :: countTable(:, :), x(:)
    integer(int8) :: extensionFlag(:), qualityFlag(:)
    real(single), pointer :: counts(:) => null()
    integer :: i, j ,k

    counts => allocateSingle(1, n, "counts")


    do i = 1, n
        counts(i) = -1.0
        do j = 1, m
            k = jIndex(j)
            if (countTable(i, k) .gt. 0.0 .and. countTable(i, k) .lt. 10000.0) then
                counts(i) = max(counts(i), countTable(i, k))
            end if
        end do
    end do      

    call ReadOutStreaks2(n,                &
                         counts,           &
                         x,                &
                         extensionFlag,    &
                         qualityFlag)
  
    deallocate(counts)
    nullify(counts)

end subroutine ReadOutStreaks1

!***************************************************
! Subroutine 
!
!***************************************************
Subroutine ReadOutStreaks2(n,              &
                  countTable,              &
                  x,                       &
                  extensionFlag,           &
                  qualityFlag)
    integer, intent(in) :: n
    real(single) :: countTable(:), x(:)
    integer(int8) :: extensionFlag(:), qualityFlag(:)
    integer :: i, j, l
  
    real(single), pointer :: xStreak(:) => null(), streakhalfwidth(:) => null()
    integer, parameter :: maxNoStreaks = 100
    integer :: noStreaks, noFlagsSet
      
    real(single), parameter :: halfWidth1 = 16.0  ! IE 12 pixel aperture radius + 7.0 pixel streak width
    real(single), parameter :: halfWidth2 = 19.0  ! IE 12 pixel aperture radius + 7.0 pixel streak width
    real(single), parameter :: constant = 70.0
!   real(single), parameter :: constant = 60.0

    xStreak => allocateSingle(1, maxNoStreaks, "xStreak")
    streakhalfwidth => allocateSingle(1, maxNoStreaks, "streakhalfwidth")
 

    noStreaks = 0
!write(*, *) "n=", n, " no of pointsource flags=", count(extensionFlag .eq. 0), " extended=", count(extensionFlag .eq. 1)  
iLoop :  do i = 1, n  
!write(*, *) "Source ", i, " counttble=", countTable(i), " constant=", constant, " exposure=", exposure(i), &
!" extension flag=", extensionFlag(i)

            if (extensionFlag(i) .eq. 1) cycle
                   
            if ( countTable(i) .gt. 0.0 .and. countTable(i) .lt. 10000.0) then

!                  if ( btest(qualityFlag(i, 4)) .or. countTable(i) .ge. constant) then
  if ( countTable(i) .ge. constant) then


                    do l = 1, noStreaks
                        if (abs(x(i) - xstreak(l)) .lt. 10.0) cycle iLoop
                    end do

!write(*, *) "counttble=", countTable(i), " constant=", constant, " exposure=", exposure
                    noStreaks = noStreaks + 1
                    xStreak(noStreaks) = x(i)
                    write(errstr,'(a, f8.2)') "Possible read-out streak found centred on pixel ", x(i)
                    call message(AppMsg, Verbose)

         
                    if (countTable(i) .lt. 220.0) then
                        streakhalfwidth(noStreaks) = halfWidth1
                    else
                        streakhalfwidth(noStreaks) = halfWidth2
                    end if
       
 
                    if (noStreaks .eq. maxNostreaks) exit iLoop     
       
                end if

            end if

        
    end do iLoop

    write(errstr, '(a, i6)') "Number of read-out streaks found=", noStreaks
    call message(AppMsg, Verbose)
  
   
!write(*, *) "Readoutstreaks line 238 minval qualityflag=", minval(qualityFlag), " max=", maxval(qualityFlag)

!   return
    noFlagsSet = 0
    do i = 1, n
!        if (countTable(i) .lt. 0.0 .or. countTable(i) .gt. 10000.0) cycle

        do j = 1, noStreaks
            if (btest(qualityFlag(i), 1)) cycle
            if (abs(x(i) - xstreak(j)) .lt. streakhalfWidth(j)) then

                qualityFlag(i) = ibset(qualityFlag(i), 1)  ! BIT 1 SET
                noFlagsSet = noFlagsSet + 1
                exit
            end if
        end do
    end do

    write(errstr,'(a, i6)') "Number of sources with quality-flag bit 1 set=", noFlagsSet
    call message(AppMsg, Verbose)


!write(*, *) "Readoutstreaks line 252 minval qualityflag=", minval(qualityFlag), " max=", maxval(qualityFlag)


    deallocate(xStreak, streakhalfwidth)
    nullify(xStreak, streakhalfwidth)    

end subroutine ReadOutStreaks2


!***************************************************
!  subroutine BrightSourceProximityFlags
!
!***************************************************
Subroutine  BrightSourceProximityFlags(n,      &
                      countTable,              &
                      x, y,                    &
                      qualityFlag,             &
                      confusionFlag)
    integer, intent(in) :: n
    real(single) :: countTable(:), x(:), y(:)
    integer(int8), dimension(:) :: qualityFlag, confusionFlag
    integer :: i, j
    
  
    real(single) :: radius, dx, dy, maxrad
  
    !*********************************************************************
    ! The minimum count-rate for the setting of smoke-ring
    ! and diffraction spike flags
    !*********************************************************************
    real(single), parameter :: minimumCountRate = 80.0    

    do i = 1, n
        if (x(i) .lt. -5.0) cycle
        if ( (btest(qualityFlag(i), 4) .and. btest(confusionFlag(i), 2))  &
                  .or. countTable(i) .ge. minimumCountRate ) then
            if ( btest(qualityFlag(i), 4)) then
                maxRad = 35.0
            else 
                maxRad = min(70.0, 35.0 * countTable(i) / minimumCountRate)
            end if


            do j = i - 1, 1, -1
                dx = x(i) - x(j)
                dy = y(i) - y(j)
                radius = sqrt(dx * dx + dy * dy)
                if (radius .le. maxRad) then
                    x(j) = -x(j) - 10.0
                end if
                  
            end do 

            do j = i + 1, n
                dx = x(i) - x(j)
                dy = y(i) - y(j)
                   
                radius = sqrt(dx * dx + dy * dy)
                if (radius .le. maxRad) then
                    x(j) = -x(j) - 10.0
                end if
    
            end do
        end if
    end do
    
    do j = 1, n
        if (x(j) .lt. -5.0) then 
            x(j) = -x(j) + 10.0
            qualityFlag(j) = ibset(qualityFlag(j), 5)

        end if
        
    end do
end subroutine BrightSourceProximityFlags 



!***************************************************
! Subroutine SmokeRings
!
!***************************************************
Subroutine SmokeRings1(n, m, jIndex,           &
                      countTable,              &
                      x, y,                    &
                      extensionFlag,           &
                      qualityFlag)
    integer, intent(in) :: n, m, jIndex(:)
    real(single) :: countTable(:, :), x(:), y(:)
    integer(int8) :: extensionFlag(:), qualityFlag(:)

    real(single), pointer :: counts(:) => null()
    integer :: i, j ,k

    counts => allocateSingle(1, n, "counts")

    do i = 1, n
        counts(i) = -1.0
        do j = 1, m
            k = jIndex(j)
            if (countTable(i, k) .gt. 0.0 .and. countTable(i, k) .lt. 10000) then
                counts(i) = max(counts(i), countTable(i, k))
            end if
        end do
    end do      


    call SmokeRings2(n, counts,                &
                      x, y,                    &
                      extensionFlag,           &
                      qualityFlag)
   

    deallocate(counts)
    nullify(counts)
 
end subroutine SmokeRings1


!***************************************************
! Subroutine SmokeRings
!
!***************************************************
Subroutine SmokeRings2(n,                      &
                      countTable,              &
                      x, y,                    &
                      extensionFlag,           &
                      qualityFlag)
    integer, intent(in) :: n
    real(single) :: countTable(:), x(:), y(:)
    integer(int8) :: extensionFlag(:), qualityFlag(:)
    integer :: i, j
    real(single) :: xSmokeRing, ySmokeRing
    integer, parameter :: maxNoSmokeRings = 100
    integer :: noSmokeRings
!    real(single), parameter :: halfWidth = 38.0 ! IE 26 ring radius + 12 photometry radus)

!  real(single), parameter :: halfWidth = 45.0 ! IE 26 ring radius + 12 photometry radus)


    real(single), parameter :: halfWidth = 40.0 ! IE 26 ring radius + 12 photometry radus)

    integer(int32), pointer :: smokeRingIndex(:) => null()   
  
    real(single) :: rStar, rRing, theta


  
    !*********************************************************************
    ! The minimum count-rate for the setting of smoke-ring
    ! and diffraction spike flags
    !*********************************************************************
    real(single), parameter :: minimumCountRate = 60.0 ! Was 70.0    



    if (n < 1) return

    ! The following has been put in to prevent a compilation warning
    ! Need to remove this parameter from all calls to the subroutine
    if (extensionFlag(1) .eq. 1) then

    endif 
   









    smokeRingIndex => allocateInt32(1, maxNoSmokeRings, "smokeRingIndex")
  


    noSmokeRings = 0

    ! Loop through the sources to find any that may produce a smoke-ring
iLoop :  do i = 1, n  


!write(*, *) "checking source ",  i,  " for smoke rings ", " x=", x(i), " y=", y(i), &
!" countrate=", countTable(i), " quality flag=", qualityFlag(i), " extension=", extensionFlag(i)
!write(*, *) "btest(qualityflag, 4)=",  btest(qualityFlag(i), 4)



!            if (extensionFlag(i) .eq. 1) cycle 
            if (countTable(i) .gt. 0.0 .and. countTable(i) .lt. 10000.0) then
          if ( counttable(i) .ge. minimumCountRate) then ! This star may have produced spikes
   

               
!write(*, *) "checking source ",  i,  " for smoke rings ", " x=", x(i), " y=", y(i), " count rate=", counttable(i)
!write(*, *) "btest(qualityflag, 4)=",  btest(qualityFlag(i), 4)


                    ! The source is predicted to produce a smoke-ring 
                    ! Predict where smoke-ring is
                    ! Radial distance of star from centre of full-frame image
                    ! xCentre and yCentre are defined in omconstants_mod (1024.5. 1024.5)
                     rStar = RadialDistance(real(xCentre), real(yCentre), x(i), y(i))
                     theta = PositionAngle(real(xCentre), real(yCentre), x(i), y(i))
                     rRing = 1.206 * rStar

 !                    if (rRing .lt. halfWidth) cycle 
                     if (rRing .lt. 20) cycle 
            
                     noSmokeRings = noSmokeRings + 1
              
                     xSmokeRing = xCentre + rRing * cos(theta)
                     ySmokeRing = yCentre + rRing * sin(theta)
                        

!write(*, *) "Possible smoke ring associated with source ", i, " x=", x(i), " y=", y(i)
!write(*, *) " x smoke ring=", xSmokeRing, " y=", ysmokering

                     do j = 1, n
                         if (i .eq. j) cycle
                         
                         if ( RadialDistance(x(j), y(j), xSmokeRing, ySmokeRing) .le. halfwidth) then
!write(*, *) "setting flag on source ", j, " x=", x(j), " y=", y(j), " distance=",  RadialDistance(x(j), y(j), xSmokeRing, ySmokeRing)

                             qualityFlag(j) = ibset(qualityFlag(j), 2)   ! BIT 2 SET
                
        
                        end if
!write(*, *) "star ", j, "x=", x(j), " y=", y(j), " radial distance=",  RadialDistance(x(j), y(j), xSmokeRing, ySmokeRing)

                    end do
 

                end if

            end if
   
    end do iLoop

    write(errstr, '(a, i6)') "Number of smoke rings=", noSmokeRings
    call message(AppMsg, Verbose)
  
  

   
end subroutine SmokeRings2
!************************************************
! 
!***********************************************
real(single) Function RadialDistance(x1, y1, x2, y2)
    real(single), intent(in) :: x1, y1, x2, y2
    real(single) ::  dx, dy

    dx = x2 - x1
    dy = y2 - y1
    RadialDistance = sqrt(dx * dx + dy * dy)

end function RadialDistance
!************************************************
! 
!***********************************************
real(single) Function PositionAngle(x1, y1, x2, y2)
    real(single), intent(in) :: x1, y1, x2, y2
    real(single) ::  dx, dy

    dx = x2 - x1
    dy = y2 - y1
    if (dx .eq. 0.0 .and. dy .eq. 0.0) then
        PositionAngle = 0.0
        return
    end if

   PositionAngle = atan2(dy, dx)

end function PositionAngle

!***********************************************
!
!***********************************************
subroutine EdgeSources(n, source)
                      
    integer, intent(in) :: n
    type(sourceparameters), pointer :: source(:)

    integer :: i
    real(single), parameter :: minDist = 12.0
    real(single), parameter :: width = 2049.0
 
    integer :: count

!write(*, *) "edgesources line 684 minval source%qualityflag=", minval(source%qualityFlag), &
!" max=", maxval(source%qualityFlag)

    write(errstr, *) "Checking for sources near the edge of the image"

    call message(AppMsg, Verbose)

    count = 0
    do i = 1, n
!write(*, *) i, " x=", source(i)%xpos, " y=", source(i)%ypos, " flag=", source(i)%qualityFlag

        if (source(i)%xpos .lt. minDist .or. (width - source(i)%xpos).lt. minDist .or. &
            source(i)%ypos .lt. minDist .or. (width - source(i)%ypos).lt. minDist) then

             source(i)%confusionFlag = ibset(source(i)%confusionFlag, 2)  

            count = count + 1
        end if

        !**************************************************************************
        ! Check for sources close to a corner
        !*************************************************************************
       
    end do

    write(errstr, *) "Number of source flags set=", count
    call message(AppMsg, Verbose)
    
!write(*, *) "edgesources line 720 minval source%qualityflag=", minval(source%qualityFlag), &
!" max=", maxval(source%qualityFlag)


end subroutine EdgeSources

!***************************************************
! Subroutine StarSpikes
!
!***************************************************
Subroutine StarSpikes(n, source)
   
                      
    integer, intent(in) :: n
    type(sourceparameters), pointer :: source(:)
    integer(int16) :: i, j, k, l, m
    real(single) :: dx, dy
    type(sourceparameters) :: tempsource
    type(sourceparameters), pointer :: tempsources(:) => null()
  
    integer(int8) :: maxCount, count1, count2
        
    integer(int16), parameter :: angle(4) = (/45_int16, 225_int16, 135_int16, 315_int16/)   
    integer(int16), parameter :: noOffsets = 17


    real(single), parameter :: offset(noOffsets) = (/0.0, -0.5, 0.5, &
                                                    -1.0, 1.0, -1.5, 1.5, -2.0, 2.0, &
                                                    -2.5, 2.5, -3.0, 3.0, -3.5, 3.5, &
                                                    -4.0, 4.0/)


    !*********************************************************************
    ! The minimum count-rate for the setting of smoke-ring
    ! and diffraction spike flags
    !*********************************************************************
    real(single), parameter :: minimumCountRate = 100.0    

 
    integer(int16) :: noSpikes, minRadius, maxRadius, minrad1, maxrad1
    logical, pointer :: useStar(:) => null()  
    integer, pointer :: index(:) => null()
    integer :: il, iu


    write(errstr, *) "Checking for stars with diffraction spikes"
    call message(AppMsg, Verbose)

              

    noSpikes = 0
    allocate(useStar(1 : n))

    allocate(tempsources(1 : n))
    tempSources = source(1 : n)

  
    index => allocateInt32(1, n, "index")
    call indexSort(tempSources%xpos, index)

    do i = 1, n
        tempSources(i) = source(index(i))
        if(tempSources(i)%corrCountRate .gt. 0.25 * minimumCountRate) then
            useStar(i) = .false.
        else
            useStar(i) = .true.
        end if     
    end do

   

iLoop : do i = 1, n  
        


!write(*, *) i, ") count rate=", tempsources(i)%rawCountRate, " exposure=", tempSources(i)%exposure                
        if (tempSources(i)%extensionFlag .eq. 0 .and. &
                 tempsources(i)%rawCountRate .ge. minimumCountRate) then ! This star may have produced spikes
               

            
        
!            write(*, *) "Checking source ", i, " x=", tempSources(i)%xpos, " y=", tempSources(i)%ypos 
            ! For position angles 45 and 135  check how many stars lie along a radial line
            ! away from the centre of the star
            maxCount = 0
            dx = 0.0
            dy = 0.0
            il = 1
            il = locate(n, tempSources%xpos, max(0.0, tempSources(i)%xpos - 120.0), il)
            iu = 1
            iu = locate(n, tempSources%xpos, tempSources(i)%xpos + 120.0, iu)
            tempSource = tempsources(i)

            count1 = 0          
            minrad1 = 1000
            maxrad1 = 0
            do j = 1, noOffsets    

                tempsources(i)%xpos = tempSource%xpos + offset(j)           
              
                do k = 1, noOffsets
    
                    minRadius = 1000
                    maxRadius = 0
        
                    count1 = 0
                    tempsources(i)%ypos = tempSource%ypos + offset(k)           
                    do l = 1, 4
                              
                        !******************************************************************
                        ! Check for stars along position angle 45 and 135 degrees
                        !***************************************************************** 
                        if (CheckForStars(tempsources, angle(l), i, il, iu, useStar, minRadius, maxRadius, count2)) then
                            count1 = count1 + 1
                            if (count1 .gt. maxcount) then     
                                maxCount = count1        
                                dx = offset(j)
                                dy = offset(k)
            
                            end if         
                        end if
!                 write(*, *) "Line 1278 angl=", angle(l), " count2=", count2, " count1=", count1, " maxcount=", maxcount
  
                end do
            end do    
            end do
            if (maxcount .gt. 2) then
                noSpikes = noSpikes + 1                    
                write(errstr,'(a, i5, a, f7.2, a, f7.2, a, f12.6, a, f12.6, a, f8.2)') &
                               "Possible diffraction spikes found associated with star ", &
                                i, " x=", tempsources(i)%xpos, " y=", tempsources(i)%ypos, &
                                " raw-rate=", tempSources(i)%rawCountRate

                call message(AppMsg, Verbose)
              
!write(*, *) "Line 1131 minrad1=", minrad1, " maxrad1=", maxrad1, " dx=", dx, " dy=", dy
                minrad1 = 10.0
                maxrad1 = 100.0
                tempsources(i)%xpos = tempSource%xpos + dx
                tempsources(i)%ypos = tempSource%ypos + dy
            
                ! Set the flags
                do m = 1, 4
                    call SetStarSpikeFlags(tempsources, i, il, iu, angle(m), useStar, minRad1, maxRad1) 
                end do
            end if 
            tempsources(i) = tempSource 
                            
        end if
             
    end do iLoop    
    write(errstr,'(a, i4)') "No of possible sources with diffraction-spikes=", noSpikes
    call message(AppMsg, Verbose)

    do i = 1, n
        source(index(i)) = tempSources(i)
    end do


    deallocate(index, tempSources, usestar)
    nullify(index, tempSources, usestar)
    
    
end subroutine StarSpikes
!********************************************************
!
!
!
!
!*******************************************************
logical Function CheckForStars(source, theta, i, il, iu, useStar, minRad, maxRad, count1)
    type(sourceparameters), pointer :: source(:)

    integer(int16), intent(in) :: theta
    integer(int16), intent(in) :: i
    integer(int32), intent(in) :: il, iu

    logical, pointer :: useStar(:)
    integer(int16), intent(inout) :: minRad, maxRad
    integer(int8), intent(out) :: count1
    integer(int16) :: j, k
    real(single) :: xpos, yPos, theta1, theta2
    real(single) :: radius, costheta, sintheta
    real(single), parameter :: maxdistance = 3.0 ! was 4.0
    real(single), parameter :: maxDeltatheta = 30.0 ! was 30.0
  
    real(single) :: dx, dy
    integer(int16),  parameter :: minRadius = 10, maxRadius = 100  ! Maximum radial distance from star to check
 
    integer(int8) :: count2  
   
    theta1 = real(theta) * degreesToRadians
  
    costheta = cos(theta1)
    sintheta = sin(theta1)

    count1 = 0_int8
    count2 = 0_int8
    theta1 = mod(real(theta), 180.0)
     

    checkForStars = .false.
    minrad = minRadius       
    do j = minRadius, maxRadius             ! The radii from star to be checked
       
        radius = real(j)
        xpos = source(i)%xpos + radius * costheta      
        yPos = source(i)%ypos + radius * sintheta      ! Y-position along possible spike

        do k = il, iu
            if (k .eq. i .or. .not. useStar(k)) cycle  ! Skip star causing spike or other bright ones

            dx = xpos - source(k)%xpos
            if (dx .gt. maxDistance) cycle 
     
            dy = ypos - source(k)%ypos
            if (dy .gt. maxDistance) cycle 
     
            radius = sqrt(dx * dx + dy * dy)
            if (radius .gt. maxDistance) cycle 
                                                         
            !*************************************************** 
            ! Check that the source is elongated along the spike
            !***************************************************
            theta2 = mod(source(k)%theta, 180.0)
            
            if ( abs(theta1 - theta2) .le. maxDeltaTheta) then
 
                minRad = min(minRad, j)
                maxRad = max(maxRad, j)
                count1 = count1 + 1_int8
            else
                count2 = count2 + 1_int8                            
            end if

            
        end do 
        if ( (j - minRad) .gt. 80) exit
    end do

    if (count1 .gt. 1_int8 .or. (count1 .gt. 0_int8 .and. count2 .gt. 0_int8)) then  
        CheckForStars = .true.
    else
        CheckForStars = .false.
    end if

   
end Function CheckForStars

!********************************************************
!
!
!
!
!*******************************************************
subroutine SetStarSpikeFlags(source, index, il, iu, theta, useStar, minRadius, maxRadius) 
                             
    type(sourceparameters), pointer :: source(:)
    integer(int16), intent(in) :: index
    integer(int32), intent(in) :: il, iu
    integer(int16), intent(in) :: theta
    logical, pointer :: useStar(:)
    integer(int16), intent(in) :: minRadius, maxRadius
    integer(int16) :: i, j, k
    real(single) :: xpos, yPos, theta1
    real(single) :: radius, costheta, sintheta
    real(single) :: tolerance
    integer(int16), parameter :: deltaik = 80

!integer :: ncount 
      
    real(single) :: dx, dy
  
!ncount = 0  
      
    theta1 = real(theta) * degreesToRadians
  
    costheta = cos(theta1)
    sintheta = sin(theta1)
   

!write(*, *) "Setstarspikeflags minradius=", minradius, " maxradius=", maxradius
    k = minRadius
!write(*, *) "il=", il, " iu=", iu

    do i = minRadius, maxRadius  ! The radii from star to be checked
        radius = real(i)
        xpos = source(index)%xpos + radius * costheta      
        yPos = source(index)%ypos + radius * sintheta      

        do j = il, iu
     

            if (.not. usestar(j)) cycle

!  source(j)%qualityFlag = ibset(source(j)%qualityFlag, 3)

            if (j .eq. index) cycle
            if (source(j)%extensionFlag .eq. 0) then
                tolerance = 5.0
            else
                tolerance = 3.0
            end if    

            dx = xpos - source(j)%xpos
            dy = ypos - source(j)%ypos
            radius = sqrt(dx * dx + dy * dy)
          
            if (radius .le. tolerance) then
            
!ncount = ncount + 1
                source(j)%qualityFlag = ibset(source(j)%qualityFlag, 3)
                k = i     
                usestar(j) = .true.                 
            end if
            if ( (i - k) .gt. deltaik) exit

        end do 
    end do
!write(*, *) "No of flags set=", ncount

end subroutine SetStarSpikeFlags


end module omsourceflags_mod









