
!*********************************************************************************************************** 
module omSrcTable_mod

    use dal
    use cal
    use param  
    use sasdal
    
    use OmAllocation_mod         ! Functions for allocating memory to arrays.
    use OmSourceparam_mod             ! Source definitions, etc
    use omconstants_mod
    use OmPhotometryDefs_mod
    use omutility_mod   
    use errorHandling

    implicit none 
 
  private
  public CreateSourceListFile
  public CreateSourceListFile1

  public AddMagnitudesToTable
contains 



!*********************************************
! Subroutine CreateSourceList
!*********************************************
subroutine CreateSourceListFile1(outset,         &
                            oswDataSet,     &
                            nSources,       &
                            sourceData,   &
                            plateScale,     &
                            apertureRadius, &
                            modeBackground, &
                            nSigma,         &
                            exposure,       & 
                            windowX0,       &
                            windowY0,       & 
                            windowDX,       &
                            windowDY) 
    character(len=*), intent(in) ::  outset

    type(DataSetT), intent(in)  :: oswDataSet   ! their associated handles
    integer(int32), intent(in) :: nSources 
    type(sourceParameters) :: sourceData(:)

    real(double), intent(in) :: plateScale
    real(single), intent(in) :: apertureRadius
    real(single), intent(in) :: modeBackground
    real(single), intent(in) :: nSigma
    real(single), intent(in) :: exposure
    integer(int32), intent(in)  :: windowX0     ! WindowX0.
    integer(int32), intent(in)  :: windowY0     ! WindowY0.
    integer(int32), intent(in)  :: windowDX     ! WindowDX.
    integer(int32), intent(in)  :: windowDY     ! WindowDY.
      

    type(DataSetT) :: listDataSet 
    type(TableT) :: listTable
    type(ColumnT)        :: pcolumn                 ! pointer to a column

    integer (int32), pointer :: srcnum(:)  ! the source number
   
   
    integer(int32) :: i
    real (single) :: detectionRate
    

    ! Write out the source list to the oswDataSet
    write(errstr,'(a)') "Creating source list"
    call message(AppMsg, Verbose)

    !**********************************
    ! Create the output source list
    !**********************************
    listDataSet = dataSet(outset, create)

    listTable = addTable(listDataSet, "SRCLIST", nSources)

    ! copy the keywords from the image header
    call copyAllAttributes(attributable(oswDataSet), attributable(listDataSet))

    !*****************************
    ! Add in detection count rate
    ! which will be stored in the source list file
    ! and used by ommag to calculate the
    ! background limiting  magnitude
    !************************************

    if (modeBackground .gt. 0.0) then  
        detectionRate = nSigma * sqrt(modeBackground * PI * apertureRadius * apertureRadius ) / exposure
    else
        detectionRate = 0.0
    end if 

   

     write(errstr,'(a, f4.1, a)') "Detection limiting count-rate (nSigma=", nsigma, ")" 
     call setAttribute(listDataSet, "CRLIMIT", detectionRate, trim(errstr))


    write(errstr,'(a, f8.4, a, f4.1)') "Detection limiting Count Rate: ", detectionRate, " computed for nsigma=", nSigma 
    call message(AppMsg, Verbose)
   
  
    call setAttribute(listDataSet, "WINDOWX0", windowX0, "pixels", "Window x-origin")
    call setAttribute(listDataSet, "WINDOWDX", windowDX, "pixels", "Window dx")
    call setAttribute(listDataSet, "WINDOWY0", windowY0, "pixels", "Window y-origin")
    call setAttribute(listDataSet, "WINDOWDY", windowDY, "pixels", "Window dy")
   
    ! this is an osw source list

    call setAttribute(listDataSet, "CONTENT", "OM OSW SOURCE LIST", &
                  "File contents")
    call setAttribute(listDataSet, "HDUCLASS", "OGIP", &
                  "Format conforms to OGIP/GSFC conventions")
    call setAttribute(listDataSet, "HDUCLAS1", "SRCLIST", &
                  "File contains a source list")

 
    call setAttribute(listDataSet, "INNBCKAN", innerRadius, "unbinned pixels", "Inner background annulus")
    call setAttribute(listDataSet, "OUTBCKAN", outerRadius, "unbinned pixels", "Outer background annulus")



    !********************************
    ! now add the columns one by one
    ! starting with the source number
    !*********************************
    pcolumn = addColumn(listTable, "SRCNUM", Integer32, " ")
    srcnum => int32data(pcolumn)

    forall (i = 0 : nSources - 1) srcnum(i) = i + 1
   
    call release(pcolumn)


   
    ! Add the source positions
    call AddReal32Column(listTable, sourceData%xPos,  nSources, &
                         "XPOS", "pixels", "Centroided x-pixel position", nullValue)

    call AddReal32Column(listTable, sourceData%yPos,  nSources, &
                         "YPOS", "pixels", "Centroided y-pixel position", nullValue)

  
    call AddReal32Column(listTable, sourceData%poserr,  nSources, &
                         "POSERR", "pixels", "Centroided pixel position error", nullValue)
     
     
    call AddReal32Column(listTable, sourceData%rawCountRate, nSources, &
                        "RATE", "counts/sec", "Source-background. No corrections have been applied", nullValue)
!write(*, *) "omsrctable raw count rate(nsources)=", sourceData(nsources)%rawCountRate     
!write(*, *) "omsrctable corr count rate(nsources)=", sourceData(nsources)%corrCountRate     
!write(*, *) "omsrctable raw background count rate(nsources)=", sourceData(nsources)%rawbackgroundCountRate     
!write(*, *) "omsrctable corr background count rate(nsources)=", sourceData(nsources)%corrbackgroundCountRate     



    call AddReal32Column(listTable, sourceData%rawCountRateErr, nSources, &
                        "RATE_ERR", "counts/sec", "No corrections have been applied", nullValue)
  

    call AddReal32Column(listTable, sourceData%rawBackGroundCountRate, nSources, &
              "BACKGROUND_RATE", "counts/sec",  "No corrections have been applied", nullValue)
    

    call AddReal32Column(listTable, sourceData%significance,  nSources, &
               "SIGNIFICANCE", "sigma", "(source-background)/sqrt(source+background)", nullValue)
 
    ! Add the coincidence-corrected source-count rates 
 
   call AddReal32Column(listTable,  sourceData%corrCountRate, nSources, &
                "CORR_RATE", "counts/sec", "Source-background. Corrected for coincidence losses, etc", nullValue)

    

    call AddReal32Column(listTable,  sourceData%corrCountRateErr, nSources, &
                "CORR_RATE_ERR", "counts/sec", "Corrected for coincidence losses", nullValue)
    

    
    call AddReal32Column(listTable, sourceData%corrBackGroundCountRate, nSources, &
                         "CORR_BACKGROUND_RATE", "counts/sec", "Corrected for coincidence losses, etc", nullValue)



    ! Add the coincidence-loss correction  
    call AddReal32Column(listTable, sourceData%psf1Correction, nSources, &
              "PSF1_CORR", "%", "PSF correction from mesurement to 12 pixel aperture", nullValue)


!    call AddReal32Column(listTable, sourceData%coincidenceLossCorr, nSources, &
!              "COINCLOSS_CORR", "%", "Coincidence-loss scaling factor", nullValue)


    call AddReal32Column(listTable, real(sourceData%rightAscension), nSources, &
              "SPB_COILOSS", " ", "(S+B coinc-loss corr)/(S+B un-coinc-loss corr)", nullValue)

    call AddReal32Column(listTable, real(sourceData%declination), nSources, &
              "BK_COILOSS", " ", "(B coinc-loss corr)/(B un-coinc-loss corr)", nullValue)






    !***************************************************************************
    ! add the magnitudes
    !*************************************************************************
   
    call AddReal32Column(listTable, sourceData%magnitude, nSources, "MAG", &
            "mag", "Instrumental magnitude", nullValue)
    

    !****************************************************************
    ! Add the magnitude errors
    !**************************************************************** 
    call AddReal32Column(listTable, sourceData%magnitudeErr, nSources, "MAGERR", &
            "mag", "Instrumental magnitude error", nullValue)
        







    ! Add the number of pixels associated with extended sources- for point-sources
    ! value set to NULL
    call AddInt32Column(listTable, sourceData%noSourcePixels, nSources, &
                "NOPIXELS", "pixels", "No of pixels assigned to extended sources")
 

    ! Add the aperture radius 
    call AddReal32Column(listTable, sourceData%apertureRadius, nSources, &
                    "APERTURE", "unbinned pixels", "Aperture radius used in photometry", nullValue)


    ! Add the source shape parameters 
    ! Convert sigmas to arc secs
    where(sourceData%semia .gt. 0.0)
        sourceData%semiA = sourceData%semiA * 2.354 * binx * plateScale
        sourceData%semiAErr = sourceData%semiAErr * 2.354 * binx * plateScale
  
        sourceData%semiB = sourceData%semiB * 2.354 * biny * plateScale
        sourceData%semiBErr = sourceData%semiBErr * 2.354 * binx * plateScale
    end where


    ! NB the full-width-half-maximum = 2.354 * sigma (gaussian sigma)
    call AddReal32Column(listTable,  sourceData%semiA, nSources, &
                 "FWHM_MAJ", "arcsecs", "FWHM of semi-major axis", nullValue)
 
    call AddReal32Column(listTable,  sourceData%semiAErr, nSources, &
                 "FWHM_MAJ_ERR", "arcsecs", "FWHM error of semi-major axis", nullValue)
 
 

    call AddReal32Column(listTable,  sourceData%semiB, nSources, &
                 "FWHM_MIN", "arcsecs", "FWHM of semi-minor axis", nullValue)

    call AddReal32Column(listTable,  sourceData%semiBErr, nSources, &
                 "FWHM_MIN_ERR", "arcsecs", "FWHM error of semi-minor axis", nullValue)
 

 
    call AddReal32Column(listTable,  sourceData%theta, nSources, &
                 "PA", "degrees", "PA of semi-major axis (measured anti-clockwise from x-axis)", nullValue)
    call AddReal32Column(listTable,  sourceData%thetaErr, nSources, &
                 "PA_ERR", "degrees", "PA error of semi-major axis", nullValue)
 

    ! Add the quality flags
    call AddInt8Column(listTable,  sourceData%qualityFlag, nSources, &
                  "QFLAG", "none", "see comment below)")   ! Can be either 0 or 1
    call AddInt8Column(listTable,  sourceData%confusionFlag, nSources, &
                             "CFLAG", "none", "0=isolated")   ! 
    call AddInt8Column(listTable, sourceData%extensionFlag, nSources, &
              "EFLAG", "none", "0=point, 1=extended source")    ! Can be either 0 (point-source) or 1 (extended)
 




    call addcomment(listTable, "QFLAG- bits set (colour refers to region colour)")
    call addcomment(listTable, "if QFLAG=0 then region colour will be either")
    call addcomment(listTable, "green (point-source) or cyan (extended source)")
    call addcomment(listTable, "0=Bad pixel (red)")
    call addcomment(listTable, "1=Source on possible read-out streak (magenta)")
    call addcomment(listTable, "2=Source on possible smoke-ring (yellow)")
    call addcomment(listTable, "3=Source on possible star diffraction-spike (white)")
    call addcomment(listTable, "4=Source likely affected by mod-8 pattern (black)")
    call addcomment(listTable, "5=Source likely near to a very bright source (black)")
    call addcomment(listTable, "6=Source within central enhanced region (red)")

    call addcomment(listTable, "CFLAG- bits set (colour refers to region colour")
    call addcomment(listTable, "0 (point-source only) one or more sources lie within a radius of 6 to 12 unbinned pixels")
    call addcomment(listTable, "1 (point-source only) one or more sources lie within a radius of 6 unbinned pixels")
    call addcomment(listTable, "2 Source lies near an image edge (blue)")
    call addcomment(listTable, "3 point-source lies partly, or entirely, within an extended-source region")
   

    !************************
    ! release the memory
    !************************
    call release(listTable)
    call release(listDataSet)
   

end subroutine CreateSourceListFile1

!*********************************************
! Subroutine CreateSourceList
!*********************************************
subroutine CreateSourceListFile(outset,         &
                            oswDataSet,     &
                            nSources,       &
                            sourceDataIn,   &
                            plateScale,     &
                            apertureRadius, &
                            modeBackground, &
                            nSigma,         &
                            exposure,       & 
                            windowX0,       &
                            windowY0,       & 
                            windowDX,       &
                            windowDY) 
    character(len=*), intent(in) ::  outset

    type(DataSetT), intent(in)  :: oswDataSet   ! their associated handles
    integer(int32), intent(in) :: nSources 
    type(sourceParameters), pointer :: sourceDataIn(:)

    real(double), intent(in) :: plateScale
    real(single), intent(in) :: apertureRadius
    real(single), intent(in) :: modeBackground
    real(single), intent(in) :: nSigma
    real(single), intent(in) :: exposure
    integer(int32), intent(in)  :: windowX0     ! WindowX0.
    integer(int32), intent(in)  :: windowY0     ! WindowY0.
    integer(int32), intent(in)  :: windowDX     ! WindowDX.
    integer(int32), intent(in)  :: windowDY     ! WindowDY.
      

    type(DataSetT) :: listDataSet 
    type(TableT) :: listTable
    type(ColumnT)        :: pcolumn                 ! pointer to a column

    integer (int32), pointer :: srcnum(:)  ! the source number
   
   
    integer(int32) :: i
    real (single) :: detectionRate

    type(sourceParameters), pointer :: sourceData(:) => null()

    if (associated(sourceDataIn)) then
        sourceData => sourceDataIn
    else
       allocate(sourceData(1 : 1))
    end if

    ! Write out the source list to the oswDataSet
    write(errstr,'(a)') "Creating source list"
    call message(AppMsg, Verbose)

    !**********************************
    ! Create the output source list
    !**********************************
    listDataSet = dataSet(outset, create)

    listTable = addTable(listDataSet, "SRCLIST", nSources)

    ! copy the keywords from the image header
    call copyAllAttributes(attributable(oswDataSet), attributable(listDataSet))

    !*****************************
    ! Add in detection count rate
    ! which will be stored in the source list file
    ! and used by ommag to calculate the
    ! background limiting  magnitude
    !************************************

    if (modeBackground .gt. 0.0) then  
        detectionRate = nSigma * sqrt(modeBackground * PI * apertureRadius * apertureRadius ) / exposure
    else
        detectionRate = 0.0
    end if 

   

     write(errstr,'(a, f4.1, a)') "Detection limiting count-rate (nSigma=", nsigma, ")" 
     call setAttribute(listDataSet, "CRLIMIT", detectionRate, trim(errstr))


    write(errstr,'(a, f8.4, a, f4.1)') "Detection limiting Count Rate: ", detectionRate, " computed for nsigma=", nSigma 
    call message(AppMsg, Verbose)
   
  
    call setAttribute(listDataSet, "WINDOWX0", windowX0, "pixels", "Window x-origin")
    call setAttribute(listDataSet, "WINDOWDX", windowDX, "pixels", "Window dx")
    call setAttribute(listDataSet, "WINDOWY0", windowY0, "pixels", "Window y-origin")
    call setAttribute(listDataSet, "WINDOWDY", windowDY, "pixels", "Window dy")
   
    ! this is an osw source list

    call setAttribute(listDataSet, "CONTENT", "OM OSW SOURCE LIST", &
                  "File contents")
    call setAttribute(listDataSet, "HDUCLASS", "OGIP", &
                  "Format conforms to OGIP/GSFC conventions")
    call setAttribute(listDataSet, "HDUCLAS1", "SRCLIST", &
                  "File contains a source list")

 
    call setAttribute(listDataSet, "INNBCKAN", innerRadius, "unbinned pixels", "Inner background annulus")
    call setAttribute(listDataSet, "OUTBCKAN", outerRadius, "unbinned pixels", "Outer background annulus")



    !********************************
    ! now add the columns one by one
    ! starting with the source number
    !*********************************
    pcolumn = addColumn(listTable, "SRCNUM", Integer32, " ")
    srcnum => int32data(pcolumn)

    forall (i = 0 : nSources - 1) srcnum(i) = i + 1
   
    call release(pcolumn)


   
    ! Add the source positions
    call AddReal32Column(listTable, sourceData%xPos,  nSources, &
                         "XPOS", "pixels", "Centroided x-pixel position", nullValue)

    call AddReal32Column(listTable, sourceData%yPos,  nSources, &
                         "YPOS", "pixels", "Centroided y-pixel position", nullValue)

  
    call AddReal32Column(listTable, sourceData%poserr,  nSources, &
                         "POSERR", "pixels", "Centroided pixel position error", nullValue)
     
     
    call AddReal32Column(listTable, sourceData%rawCountRate, nSources, &
                        "RATE", "counts/sec", "Source-background. No corrections have been applied", nullValue)




    call AddReal32Column(listTable, sourceData%rawCountRateErr, nSources, &
                        "RATE_ERR", "counts/sec", "No corrections have been applied", nullValue)
  

    call AddReal32Column(listTable, sourceData%rawBackGroundCountRate, nSources, &
              "BACKGROUND_RATE", "counts/sec",  "No corrections have been applied", nullValue)
    

    call AddReal32Column(listTable, sourceData%significance,  nSources, &
               "SIGNIFICANCE", "sigma", "(source-background)/sqrt(source+background)", nullValue)
 
    ! Add the coincidence-corrected source-count rates 
 
   call AddReal32Column(listTable,  sourceData%corrCountRate, nSources, &
                "CORR_RATE", "counts/sec", "Source-background. Corrected for coincidence losses, etc", nullValue)

    

    call AddReal32Column(listTable,  sourceData%corrCountRateErr, nSources, &
                "CORR_RATE_ERR", "counts/sec", "Corrected for coincidence losses", nullValue)
    

    
    call AddReal32Column(listTable, sourceData%corrBackGroundCountRate, nSources, &
                         "CORR_BACKGROUND_RATE", "counts/sec", "Corrected for coincidence losses, etc", nullValue)



    ! Add the coincidence-loss correction  
    call AddReal32Column(listTable, sourceData%psf1Correction, nSources, &
              "PSF1_CORR", "%", "PSF correction from mesurement to 12 pixel aperture", nullValue)



    call AddReal32Column(listTable, real(sourceData%rightAscension), nSources, &
              "SPB_COILOSS", " ", "(S+B coinc-loss corr)/(S+B un-coinc-loss corr)", nullValue)

    call AddReal32Column(listTable, real(sourceData%declination), nSources, &
              "BK_COILOSS", " ", "(B coinc-loss corr)/(B un-coinc-loss corr)", nullValue)




    ! Add the number of pixels associated with extended sources- for point-sources
    ! value set to NULL
    call AddInt32Column(listTable, sourceData%noSourcePixels, nSources, &
                "NOPIXELS", "pixels", "No of pixels assigned to extended sources")
 

    ! Add the aperture radius 
    call AddReal32Column(listTable, sourceData%apertureRadius, nSources, &
                    "APERTURE", "unbinned pixels", "Aperture radius used in photometry", nullValue)


    ! Add the source shape parameters 
    ! Convert sigmas to arc secs
    where(sourceData%semia .gt. 0.0)
        sourceData%semiA = sourceData%semiA * 2.354 * binx * plateScale
        sourceData%semiAErr = sourceData%semiAErr * 2.354 * binx * plateScale
  
        sourceData%semiB = sourceData%semiB * 2.354 * biny * plateScale
        sourceData%semiBErr = sourceData%semiBErr * 2.354 * binx * plateScale
    end where


    ! NB the full-width-half-maximum = 2.354 * sigma (gaussian sigma)
    call AddReal32Column(listTable,  sourceData%semiA, nSources, &
                 "FWHM_MAJ", "arcsecs", "FWHM of semi-major axis", nullValue)
 
    call AddReal32Column(listTable,  sourceData%semiAErr, nSources, &
                 "FWHM_MAJ_ERR", "arcsecs", "FWHM error of semi-major axis", nullValue)
 
 

    call AddReal32Column(listTable,  sourceData%semiB, nSources, &
                 "FWHM_MIN", "arcsecs", "FWHM of semi-minor axis", nullValue)

    call AddReal32Column(listTable,  sourceData%semiBErr, nSources, &
                 "FWHM_MIN_ERR", "arcsecs", "FWHM error of semi-minor axis", nullValue)
 

 
    call AddReal32Column(listTable,  sourceData%theta, nSources, &
                 "PA", "degrees", "PA of semi-major axis (measured anti-clockwise from x-axis)", nullValue)
    call AddReal32Column(listTable,  sourceData%thetaErr, nSources, &
                 "PA_ERR", "degrees", "PA error of semi-major axis", nullValue)
 

    ! Add the quality flags
    call AddInt8Column(listTable,  sourceData%qualityFlag, nSources, &
                  "QFLAG", "none", "see comment below)")   ! Can be either 0 or 1
    call AddInt8Column(listTable,  sourceData%confusionFlag, nSources, &
                             "CFLAG", "none", "0=isolated")   ! 
    call AddInt8Column(listTable, sourceData%extensionFlag, nSources, &
              "EFLAG", "none", "0=point, 1=extended source")    ! Can be either 0 (point-source) or 1 (extended)
 




    call addcomment(listTable, "QFLAG- bits set (colour refers to region colour)")
    call addcomment(listTable, "if QFLAG=0 then region colour will be either")
    call addcomment(listTable, "green (point-source) or cyan (extended source)")
    call addcomment(listTable, "0=Bad pixel (red)")
    call addcomment(listTable, "1=Source on possible read-out streak (magenta)")
    call addcomment(listTable, "2=Source on possible smoke-ring (yellow)")
    call addcomment(listTable, "3=Source on possible star diffraction-spike (white)")
    call addcomment(listTable, "4=Source likely affected by mod-8 pattern (black)")
    call addcomment(listTable, "5=Source likely near to a very bright source (black)")
    call addcomment(listTable, "6=Source within central enhanced region (red)")

    call addcomment(listTable, "CFLAG- bits set (colour refers to region colour")
    call addcomment(listTable, "0 (point-source only) one or more sources lie within a radius of 6 to 12 unbinned pixels")
    call addcomment(listTable, "1 (point-source only) one or more sources lie within a radius of 6 unbinned pixels")
    call addcomment(listTable, "2 Source lies near an image edge (blue)")
    call addcomment(listTable, "3 point-source lies partly, or entirely, within an extended-source region")
   

    !************************
    ! release the memory
    !************************
    call release(listTable)
    call release(listDataSet) 
    if (associated(sourceData) .and. .not. associated(sourceDataIn)) deallocate(sourceData) 

end subroutine CreateSourceListFile




!****************************************************
! subroutine AddMagnitudesToTable
!
!****************************************************
subroutine AddMagnitudesToTable(outset, sourceData)
                            
    character(len=*), intent(in) ::  outset
    type(sourceParameters), pointer :: sourceData(:)
 

    type(DataSetT) :: dataSetH   ! their associated handles
    type(TableT) :: inTableH
    integer(kind=int32) :: nSources 
    integer :: position

    nSources = size(sourceData)
    
    ! Write out the source list to the oswDataSet
    write(errstr,'(a)') "Adding instrumental magnitudes to  source list"
    call message(AppMsg, Verbose)

    dataSetH = dataSet(outset, modify) 
    inTableH = Table(dataSetH, "SRCLIST")
 

    !**************************************************
    ! Add the columns for the corrected count rates,
    ! after the significance column
    !**************************************************
    position = columnNumber(inTableH, "SIGNIFICANCE")
    call AddReal32Column(inTableH, sourceData%corrCountRate, nSources, "CORR_RATE", &
            "counts/sec",  "Corrected for coincidence losses, deadfraction, etc", nullValue, position)

    call AddReal32Column(inTableH, sourceData%corrCountRateErr, nSources, "CORR_RATE_ERR", &
            "counts/sec", "Corrected for coincidence losses, deadfraction, etc" , nullValue, position + 1)

!  write(*, *) "adding mags corrcountrate(nsources=)", sourceData(nsources)%corrCountRate
!  write(*, *) "corr background countrate(nsources=)", sourceData(nsources)%corrbackgroundCountRate
 

    call AddReal32Column(inTableH, sourceData%psf2Correction, nSources, "PSF2_CORR", &
            "%", "PSF2 correction (12pixels to cal radius)", nullValue, position + 5)
    

    call AddReal32Column(inTableH, sourceData%corrBackgroundCountRate, nSources, "CORR_BACKGROUND_RATE", &
            "counts/sec", "Corrected for coincidence losses, deadfraction, etc", nullValue, position + 5)

    !****************************************************************
    ! Add the magnitudes
    !**************************************************************** 
    position = columnNumber(inTableH, "BK_COILOSS") + 1
    
    call AddReal32Column(inTableH, sourceData%magnitude, nSources, "MAG", &
            "mag", "Instrumental magnitude", nullValue, position)
    

    !****************************************************************
    ! Add the magnitude errors
    !**************************************************************** 
    call AddReal32Column(inTableH, sourceData%magnitudeErr, nSources, "MAGERR", &
            "mag", "Instrumental magnitude error", nullValue, position + 1)
        
    call release(inTableH)


    call release(dataSetH)

end subroutine AddMagnitudesToTable


end module omSrcTable_mod








